/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC series card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility for:
**
**   o ADPe-XRC-5T
**   o ADM-XRC-6TL
**   o ADM-XRC-6T1
**
** (C) Copyright Alpha Data 2009-2011
**
** The Flash address map used on the ADM-XRC-6TL and ADM-XRC-6T1 is:
**
**  0x0000000 - 0x07FFFFF  Alternate ADB3 PCI-E to OCP Bridge bitstream
**  0x0800000 - 0x0FFFFFF  Default ADB3 PCI-E to OCP Bridge bitstream
**  0x1000000 - 0x10FFFFF  Alpha Data VPD region
**  0x1100000 - 0x11FFFFF  Customer VPD region
**  0x1200000 - 0x28FFFFF  Target FPGA bitstream (*)
**  0x2900000 - 0x3FFFFFF  Failsafe target FPGA bitstream (*)
**
** This utility applies only to the regions marked (*), and cannot affect the
** other regions.
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _stprintf_s snprintf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# if defined(__VXWORKS__) || defined(__vxworks)
#   include <ctype.h>
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tmain main
# define _tprintf printf
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define BLANK_CHECK_BUFFER_SIZE (0x10000)
#define READBACK_BUFFER_SIZE (0x10000)
#define VERIFY_BUFFER_SIZE (0x10000)

const uint64_t sizeTMax = (uint64_t)(size_t)-1;

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

static ADMXRC3_STATUS
eraseFlash(
  ADMXRC3_HANDLE hCard,
  uint64_t eraseStart,
  uint64_t eraseLimit)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint64_t address;

  address = eraseStart;
  while (address < eraseLimit) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llx: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    _tprintf(_T("0x%llx\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_EraseFlash(hCard, 0, 0, blockInfo.Address, blockInfo.Length);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to erase Flash at 0x%llx-0x%llx: %s\n"),
        (unsigned long long)blockInfo.Address, (unsigned long long)(blockInfo.Address + blockInfo.Length - 1U), ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    /* Advance to next block */
    address = blockInfo.Address + blockInfo.Length;
    if (address <= eraseStart) {
      /* Catch 'address' wrapping around to 0 */
      break;
    }
  }

  return status;
}

static ADMXRC3_STATUS
programBitstream(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  uint64_t programStart)
{
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  uint8_t* pData;
  uint64_t address, chunk;
  size_t remaining;
  
  address = programStart;
  remaining = pBitstream->Length;
  pData = &pBitstream->Data[0];
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llx: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_WriteFlash(hCard, 0, 0, address, (size_t)chunk, pData); /* Cast is safe because we guarantee chunk <= maximum size_t value */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to write Flash at 0x%llx-0x%llx: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      break;
    }

    address += chunk;
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
    pData += (size_t)chunk; /* Cast is safe because we guarantee chunk <= maximum size_t value */
  }

  return status;
}

static ADMXRC3_STATUS
checkBlank(
  ADMXRC3_HANDLE hCard,
  uint64_t checkStart,
  uint64_t checkLimit,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, nError, i, chunk;

  pBuffer = (uint8_t*)malloc(BLANK_CHECK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate blank check buffer of size 0x%lx bytes\n"),
      (unsigned long)BLANK_CHECK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  address = checkStart;
  remaining = checkLimit - checkStart;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llx: %s\n"), (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > BLANK_CHECK_BUFFER_SIZE) {
      chunk = BLANK_CHECK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llx-0x%llx: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != 0xffU) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llx: expected=0x%02x, actual=0x%02x\n"),
            (unsigned long long)(address + i), 0xffU, pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

static ADMXRC3_STATUS
verify(
  ADMXRC3_HANDLE hCard,
  ADMXRC3_BITSTREAM* pBitstream,
  uint64_t verifyStart,
  uint64_t* pnError)
{
  uint8_t* pBuffer = NULL;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, chunk, nError;
  size_t i, offset, remaining;

  pBuffer = (uint8_t*)malloc(VERIFY_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lx bytes\n"),
      (unsigned long)VERIFY_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  nError = 0;
  offset = 0;
  address = verifyStart;
  remaining = pBitstream->Length;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llx: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > VERIFY_BUFFER_SIZE) {
      chunk = VERIFY_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llx-0x%llx: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    for (i = 0; i < chunk; i++) {
      if (pBuffer[i] != pBitstream->Data[offset + i]) {
        if (nError < 10) {
          _tprintf(_T("*** Error at Flash address 0x%llx: expected=0x%02x, actual=0x%02x\n"),
            (unsigned long long)(address + i), pBitstream->Data[offset + i], pBuffer[i]);
        } else if (nError == 10) {
          _tprintf(_T("*** (more errors)\n"));
        }
        nError++;
      }
    }

    address += chunk;
    offset += (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
    remaining -= (size_t)chunk; /* Cast is safe because we guarantee chunk < size of buffer */
  }

  free(pBuffer);

  *pnError = nError;
  return status;
}

ADMXRC3_STATUS
readback(
  ADMXRC3_HANDLE hCard,
  uint64_t readAddress,
  uint64_t readLength,
  FILE* pFile)
{
  uint8_t* pBuffer;
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  ADMXRC3_FLASHBLOCK_INFO	blockInfo;
  uint64_t address, remaining, chunk;
  size_t n;

  pBuffer = (uint8_t*)malloc(READBACK_BUFFER_SIZE);
  if (NULL == pBuffer) {
    _tprintf(_T("*** Failed to allocate verify buffer of size 0x%lx bytes\n"),
      (unsigned long)READBACK_BUFFER_SIZE);
    return ADMXRC3_NO_MEMORY;
  }

  address = readAddress;
  remaining = readLength;
  while (remaining) {
    status = ADMXRC3_GetFlashBlockInfo(hCard, 0, address, &blockInfo);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to get Flash block info for location 0x%llx: %s\n"),
        (unsigned long long)address, ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    chunk = blockInfo.Length - (address - blockInfo.Address);
    if (chunk > READBACK_BUFFER_SIZE) {
      chunk = READBACK_BUFFER_SIZE;
    }
    if (chunk > remaining) {
      chunk = remaining;
    }
    _tprintf(_T("0x%llx\r"), (unsigned long long)address);
    fflush(stdout);

    status = ADMXRC3_ReadFlash(hCard, 0, 0, address, (size_t)chunk, pBuffer); /* Cast is safe because we guarantee chunk < size of buffer */
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to read back Flash memory at 0x%llx-0x%llx: %s\n"),
        (unsigned long long)address, (unsigned long long)(address + chunk - 1U), ADMXRC3_GetStatusString(status, TRUE));
      free(pBuffer);
      return status;
    }

    n = fwrite(pBuffer, 1, (size_t)chunk, pFile); /* Cast is safe because we guarantee chunk < size of buffer */
    if (n != chunk) {
      _tprintf(_T("*** Failed to write all data to file, %lu(0x%lX) actual vs. %lu(0x%lX) expected.\n"),
        (unsigned long)n, (unsigned long)n, (unsigned long)chunk, (unsigned long)chunk);
      free(pBuffer);
      return ADMXRC3_UNEXPECTED_ERROR;
    }

    address += chunk;
    remaining -= chunk;
  }

  free(pBuffer);

  return status;
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  int ret = 0;
  const double clockFreq = 33.0e6;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO	cardInfo;
  ADMXRC3_FLASH_INFO flashInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_BITSTREAM* pBitstream = NULL;
  TCHAR* pCommand;
  Arguments arguments = g_initArguments;
  Option options[4];
  OptionType paramType[5];
  TCHAR* paramInfo[5];
  TCHAR* paramMeaning[5];
  unsigned int index, serial, targetIndex;
  bool_t bShowInfo = FALSE, bDoCheckBlank = FALSE, bDoErase = FALSE, bDoProgram = FALSE, bDoReadback = FALSE, bDoVerify = FALSE;
  uint64_t bootFlagAddress = 0;
  uint64_t programAddress = 0;
  uint64_t eraseAddress = 0, eraseLimit  = 0;
  uint64_t nError = 0;
  bool_t b32BitBootFlag = FALSE;
  bool_t bBootFlagVerify = TRUE;
  bool_t bBootFlagOk = FALSE;
  bool_t bFailsafeImage;
  bool_t bForce;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
  options[2].key = _T("failsafe");
  options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = _T("program failsafe image");
  options[3].key = _T("force");
  options[3].type = OptionBoolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = _T("ignore device mismatch");
  arguments.nOption = 4;
  arguments.option = options;

  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("One of erase, info, program, readback, verify or chkblank");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("target");
  paramMeaning[1] = _T("Index of target FPGA image");
  paramType[2] = OptionPath;
  paramInfo[2] = _T("file");
  paramMeaning[2] = _T("Bitstream filename (program or verify commands only)");
  paramType[3] = OptionUInt64;
  paramInfo[3] = _T("offset");
  paramMeaning[3] = _T("Offset for readback (readback command only)");
  paramType[4] = OptionUInt64;
  paramInfo[4] = _T("length");
  paramMeaning[4] = _T("Length for readback (readback command only)");
  arguments.nParamType = ARRAY_LENGTH(paramType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  if (sampleParseCommandLine(_T("flash"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }

  index = (unsigned int)options[0].value.uintVal;
  serial = (unsigned int)options[1].value.uintVal;
  bFailsafeImage = options[2].value.booleanVal;
  bForce = options[3].value.booleanVal;
  pCommand = arguments.param[0].pStringVal;
  targetIndex = (unsigned int)arguments.param[1].intVal;

  bShowInfo = _tcsicmp(pCommand, _T("info")) == 0;
  bDoProgram = _tcsicmp(pCommand, _T("program")) == 0;
  bDoVerify = _tcsicmp(pCommand, _T("verify")) == 0;
  bDoErase = _tcsicmp(pCommand, _T("erase")) == 0;
  bDoCheckBlank = _tcsicmp(pCommand, _T("chkblank")) == 0;
  bDoReadback = _tcsicmp(pCommand, _T("readback")) == 0;

  if (!bShowInfo && !bDoProgram && !bDoVerify && !bDoErase && !bDoCheckBlank && !bDoReadback) {
    _tprintf(_T("*** Please specify a valid command. Use flash /? to get help.\n"));
    ret = -1;
    goto done;
  }

  if (bDoProgram || bDoVerify) {
    if (arguments.nParam < 3) {
      _tprintf(_T("*** Programming or verification requires (1) target FPGA index and (2) .BIT filename.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bDoErase) {
    if (arguments.nParam < 2) {
      _tprintf(_T("*** Erase requires target FPGA index.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bDoReadback) {
    if (arguments.nParam < 4) {
      _tprintf(_T("*** Readback requires (1) target FPGA index, (2) byte offset and (3) byte length.\n"));
      ret = -1;
      goto done;
    }
  }

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get FPGA %lu info: %s\n"), (unsigned long)targetIndex, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }
  _tprintf(_T("Target FPGA is device %s\n"), fpgaInfo.Identifier);

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
    if (bFailsafeImage) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
    break;

  default:
    _tprintf(_T("*** This utility must be run on one of the following cards:\n\n"));
    _tprintf(_T("\tADM-XRC-II\n"));
    _tprintf(_T("\tADPe-XRC-5T\n"));
    _tprintf(_T("\tADM-XRC-6TL\n"));
    _tprintf(_T("\tADM-XRC-6T1\n"));
    ret = -1;
    goto done;
  }

  status = ADMXRC3_GetFlashInfo(hCard, 0, &flashInfo);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRC2:
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    bootFlagAddress = flashInfo.UseableStart + 0x8000U;
    programAddress = flashInfo.UseableStart + 0x8001U;
    eraseAddress = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + 0xC00000U;
      programAddress = flashInfo.UseableStart + 0xC20000U;
      eraseAddress = flashInfo.UseableStart + 0xC00000U;
      eraseLimit = flashInfo.UseableStart + 0x1800000U;
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = flashInfo.UseableStart + 0x20000U;
      eraseAddress = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0xC00000U;
    }
    b32BitBootFlag = TRUE;
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
  case ADMXRC3_MODEL_ADMXRC6T1:
    if (0 != targetIndex) {
      _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    if (bFailsafeImage) {
      bootFlagAddress = flashInfo.UseableStart + 0x1700000U;
      programAddress = flashInfo.UseableStart + 0x1720000U;
      eraseAddress = flashInfo.UseableStart + 0x1700000U;
      eraseLimit = flashInfo.UseableStart + 0x2E00000U;
    } else {
      bootFlagAddress = flashInfo.UseableStart;
      programAddress = flashInfo.UseableStart + 0x20000U;
      eraseAddress = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0x1700000U;
    }
    b32BitBootFlag = TRUE;
    break;

  default:
    _tprintf(_T("*** Value %lu for 'target' option is invalid for this model.\n"), (unsigned long)targetIndex);
    ret = -1;
    goto done;
  }

  _tprintf(_T("Flash type is %s, %llu(0x%llx) kiB\n"), flashInfo.Identifier,
    (unsigned long long)(flashInfo.Size / 1024), (unsigned long long)(flashInfo.Size / 1024));
  _tprintf(_T("Useable region is 0x%llx-0x%llx\n"),
    (unsigned long long)flashInfo.UseableStart, (unsigned long long)(flashInfo.UseableStart + flashInfo.UseableLength - 1U));

  if (bShowInfo) {
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    status = ADMXRC3_SetClockFrequency(hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoProgram || bDoVerify) {
    /* Load the .BIT file specified on the command line */
    status = ADMXRC3_LoadBitstream(arguments.param[2].pPathVal, &pBitstream);
    if (status != ADMXRC3_SUCCESS) {
      _tprintf(_T("*** Failed to load bitstream '%s': %s\n"),
        arguments.param[2].pPathVal, ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    _tprintf(_T("Bitstream is for device %s, length %lu(0x%lX) bytes\n"),
      pBitstream->Identifier, (unsigned long)pBitstream->Length, (unsigned long)pBitstream->Length);
    if (_tcsicmp(pBitstream->Identifier, fpgaInfo.Identifier) != 0) {
      if (!bForce || !bDoProgram) {
        _tprintf(_T("*** ERROR: Bitstream device does not match target FPGA device - aborting\n"));
        ret = -1;
        goto done;
      } else {
        _tprintf(_T("WARNING: Device information in bitstream device does not match target FPGA device.\n"));
        if (bDoProgram) {
          _tprintf(_T("         Overriding due to +force option specified on command line.\n"));
        }
      }
    }
  }

  if (bDoErase || bDoProgram) {
    /* For the 'program' or 'erase' commands, a region of the Flash is first erased  */
    _tprintf(_T("Erasing Flash from 0x%llx-0x%llx...\n"), (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = eraseFlash(hCard, eraseAddress, eraseLimit);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Erase failed: %s"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoProgram) {
    /* For the 'program' commands, the bitstream is written into a region of the Flash  */
    _tprintf(_T("Writing bitstream image at 0x%llx-0x%llx...\n"),
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = programBitstream(hCard, pBitstream, programAddress);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Programming failed: %s"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
  }

  if (bDoErase || bDoCheckBlank) {
    /* For the 'erase' commands, a blank check is performed for a region of the Flash */
    _tprintf(_T("Blank checking from 0x%llx-0x%llx...\n"),
      (unsigned long long)eraseAddress, (unsigned long long)(eraseLimit - 1U));
    status = checkBlank(hCard, eraseAddress, eraseLimit, &nError);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Blank check failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Blank check failed with %llu(0x%llx) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      goto done;
    } else {
      _tprintf(_T("Blank check successful.\n"));
    }
  }

  if (bDoProgram || bDoVerify) {
    /* For the 'program' and 'verify' commands, the bitstream data is compared against the contents of a region of the Flash */
    _tprintf(_T("Verifying from 0x%llx-0x%llx...\n"),
      (unsigned long long)programAddress, (unsigned long long)(programAddress + pBitstream->Length - 1U));
    status = verify(hCard, pBitstream, programAddress, &nError);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Verification check failed: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
      ret = -1;
      goto done;
    }
    if (nError) {
      _tprintf(_T("*** Verification failed with %llu(0x%llx) error(s)\n"),
        (unsigned long long)nError, (unsigned long long)nError);
      goto done;
    } else {
      _tprintf(_T("Verification successful.\n"));
    }
  }

  if (bDoProgram) {
    /*
    ** For the 'program' commands, the flag must be set in the Flash to cause the bitstream to be loaded
    ** at reset / power up.
    */
    if (b32BitBootFlag) {
      uint32_t val32;

      /*
      ** Set the flag that causes the FPGA to be loaded from flash,
      ** and bitstream length.
      */
      val32 = (pBitstream->Length - 1U) << 8;
      _tprintf(_T("Setting boot flag at 0x%llx-0x%llx to 0x%08lx...\n"),
        (unsigned long long)bootFlagAddress, (unsigned long long)(bootFlagAddress + sizeof(val32) - 1U), (unsigned long)val32);
      val32 = cpu_to_le32(val32);
      status = ADMXRC3_WriteFlash(hCard, 0, 0, bootFlagAddress, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
      }
    } else {
      uint8_t val8;

      /* Set the flag that causes the FPGA to be loaded from flash */
      val8 = 0x0U;
      _tprintf(_T("Setting boot flag at 0x%llx to 0x%02lx...\n"), (unsigned long long)bootFlagAddress, (unsigned long)val8);
      status = ADMXRC3_WriteFlash(hCard, 0, 0, bootFlagAddress, 0x1U, &val8);
      if (status != ADMXRC3_SUCCESS) {
        _tprintf(_T("*** Failed to set boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        bBootFlagVerify = FALSE;
      }
    }
  }

  if (bDoProgram || bDoVerify) {
    if (bBootFlagVerify) {
      /*
      ** For the 'program' and 'verify' commands, the flag in the Flash that causes the bitstream to be loaded
      ** at reset / power up must be verified.
      */
      if (b32BitBootFlag) {
        uint32_t expected, bootFlag;

        expected = (pBitstream->Length - 1U) << 8;
        status = ADMXRC3_ReadFlash(hCard, 0, 0, bootFlagAddress, sizeof(bootFlag), &bootFlag);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("*** Failed to read boot flag: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
        } else {
          bootFlag = le32_to_cpu(bootFlag);
          if (bootFlag != expected) {
            _tprintf(_T("*** Boot flag verifcation failed; expecting 0x%08lx, found 0x%08lx\n"),
              (unsigned long)bootFlag, (unsigned long)expected);
          } else {
            bBootFlagOk = TRUE;
          }
        }
      } else {
        uint64_t expected, bootFlag;

        expected = 0x0U;
        status = ADMXRC3_ReadFlash(hCard, 0, 0, bootFlagAddress, 0x1U, &bootFlag);
        if (status != ADMXRC3_SUCCESS) {
          _tprintf(_T("*** Failed to read back boot flag: %s\n"),
            ADMXRC3_GetStatusString(status, TRUE));
        } else {
          if (bootFlag != 0x0U) {
            _tprintf(_T("*** Boot flag verifcation failed; expecting 0x%02lx, found 0x%02lx\n"),
              (unsigned long)bootFlag, (unsigned long)expected);
          } else {
            bBootFlagOk = TRUE;
          }
        }
      }
    }

    if (bBootFlagOk) {
      _tprintf(_T("Boot flag OK. The bitstream will be loaded on reset / power up.\n"));
    } else {
      _tprintf(_T("*** The bitstream may not be loaded on reset / power up.\n"));
      ret = -1;
      goto done;
    }
  }

  if (bDoReadback) {
    uint64_t readAddress, readLength;
    TCHAR* pFilename;
    FILE* pFile;

    readAddress = arguments.param[3].uint64Val;
    readLength = arguments.param[4].uint64Val;
    pFilename = arguments.param[2].pPathVal;
    if (_tfopen_s(&pFile, pFilename, _T("wb"))) {
      pFile = NULL;
      ret = -1;
      _tprintf(_T("*** Failed to open file '%s' for writing.\n"), pFilename);
      goto done;
    }
    _tprintf(_T("Reading data from 0x%llx-0x%llx into '%s'...\n"),
      (unsigned long long)readAddress, (unsigned long long)(readAddress + readLength - 1U), pFilename);
    status = readback(hCard, readAddress, readLength, pFile);
    if (ADMXRC3_SUCCESS != status) {
      fclose(pFile);
      ret = -1;
      goto done;
    }
    fclose(pFile);
  }

done:
  if (NULL != pBitstream) {
    ADMXRC3_UnloadBitstream(pBitstream);
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    status = ADMXRC3_SyncFlash(hCard, 0, 0);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("+++ Warning: failed to synchronize Flash on exit: %s\n"), ADMXRC3_GetStatusString(status, TRUE));
    }
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
