--==============================================================================
--
--     File Name: adb3_ocp_ocp2ddr3_nb.vhd
--       Version: 0.3
--          Date: 05 May 2011
--         Model: syn (Synthesisable)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP to DDR3 MIG user interface (non-blocking)
--
--   Description: Component giving OCP port access to DDR3 MIG user interface
--
--   Limitations: Up to 32 MIG BL8 read bursts in flight
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 24 Sep 2010: Initial design
--                0.2, KDC, 07 Mar 2011:
--                       - Must ensure mdata FIFO is not empty during a write
--                         command before advancing cmd FIFO read pointer.
--                0.3, KDC, 05 May 2011:
--                       - Re-written using new sync/async FIFOs and FSM
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

--
-- A synchronous FIFO used to buffer commands, command attributes, data, byte
-- enables etc. in 'adb3_ocp_split_nb'.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- 'EMPTY_VAL' is the value driven on 'q(EMPTY_LO:EMPTY_HI)' when 'valid' is
-- not asserted; map both 'EMPTY_LO' and 'EMPTY_HI' to 0 if this feature is
-- not required.
--
entity adb3_ocp_new_fifo is
generic(
  ORDER       : in    natural range 2 to 6;
  WIDTH       : in    natural;
  EMPTY_VAL   : in    std_logic := '0';
  EMPTY_LO    : in    natural := 0;
  EMPTY_HI    : in    natural := 0);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  clk         : in    std_logic;                             -- Clock
  wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
  d           : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
  adv         : in    std_logic;                             -- '1' => present next item on 'q' to next item
  full        : out   std_logic;                             -- '1' => full
  not_full    : out   std_logic;                             -- '1' => not full
  valid       : out   std_logic;                             -- '1' => not empty
  q           : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture synth of adb3_ocp_new_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr, rptr : std_logic_vector(ORDER - 1 downto 0);
  signal rn : std_logic_vector(ORDER - 1 downto 0);
  signal not_empty : std_logic;
  signal i_full, i_not_full : std_logic;

  signal qvalid : std_logic;
  signal qce : std_logic;
  signal rd : std_logic;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  valid <= qvalid;
  full <= i_full;
  not_full <= i_not_full;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_not_full <= '1';
      i_full <= '0';
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert CONV_INTEGER(wptr - rptr) < MAX_LEVEL severity failure;
        tmp := wptr - rptr + 1;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      else
        tmp := wptr - rptr;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        assert CONV_INTEGER(wptr - rptr) > 0 severity failure;
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : for i in 0 to WIDTH - 1 generate
    with_reset : if EMPTY_LO <= i AND i < EMPTY_HI generate
      U0 : process(rst, clk)
      begin
        if rst = '1' then
          q(i) <= EMPTY_VAL;
        elsif rising_edge(clk) then
          if qce = '1' then
            if not_empty = '1' then
              q(i) <= sel_ram(i);
            else
              q(i) <= EMPTY_VAL;
            end if;
          end if;
        end if;
      end process;
    end generate;

    without_reset : if NOT(EMPTY_LO <= i AND i < EMPTY_HI) generate
      U0 : process(clk)
      begin
        if rising_edge(clk) then
          if qce = '1' then
            q(i) <= sel_ram(i);
          end if;
        end if;
      end process;
    end generate;
  end generate;

end;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.fifo.all;
use work.adb3_ocp.all;

entity adb3_ocp_ocp2ddr3_nb is
generic(
  app_row_width  : natural;
  app_col_width  : natural;
  app_bank_width : natural;
  app_addr_width : natural);  -- 4-byte address width
port(
  -- OCP Interface
  ocp_rst : in    std_logic;
  ocp_clk : in    std_logic;
  ocp_m2s : in    adb3_ocp_m2sT;
  ocp_s2m : out   adb3_ocp_s2mT;
  -- DDR3 MIG user interface
  mig_rst           : in    std_logic;
  mig_clk           : in    std_logic;
  phy_init_done     : in    std_logic;
  app_rdy           : in    std_logic;
  app_wdf_rdy       : in    std_logic;
  app_rd_data       : in    adb3_ocp_data_s;
  app_rd_data_valid : in    std_logic;
  app_en            : out   std_logic;
  app_cmd           : out   std_logic_vector(2 downto 0);
  app_addr          : out   std_logic_vector(app_addr_width-1 downto 0);
  app_sz            : out   std_logic;
  app_wdf_wren      : out   std_logic;
  app_wdf_data      : out   adb3_ocp_data_s;
  app_wdf_mask      : out   adb3_ocp_be_s;
  app_wdf_end       : out   std_logic;
  -- Debug
  debug_info        : out   std_logic_vector(31 downto 0));
end;

architecture synth of adb3_ocp_ocp2ddr3_nb is

  signal ocp_cmd_r  : ocp_CmdT;
  signal ocp_tag_r  : adb3_ocp_tag_s;
  signal ocp_bl_r   : adb3_ocp_burst_s;
  signal ocp_addr_r : adb3_ocp_addr_s;
  signal wr_1st_bl8 : std_logic;
  signal rd_1st_bl8 : std_logic;
  signal wr_1st_bc4 : std_logic;

  -- Slave command FIFO signals
  constant SLV_CMD_DEPTH : natural := 4; -- 3; Cannot be less than 4 using afifo
  constant SLV_CMD_WIDTH : natural := 3 + ADB3_OCP_CMD_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;

  signal slv_cmd_wr         : std_logic;
  signal slv_cmd_adv        : std_logic;
  signal slv_cmd_nfull      : std_logic;
  signal slv_cmd_not_full   : std_logic;
  signal slv_cmd_empty      : std_logic;
  signal slv_cmd_valid      : std_logic;
  signal slv_cmd_d          : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);
  signal slv_cmd_q          : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);

  signal slv_cmd_wr_bl8 : std_logic;
  signal slv_cmd_rd_bl8 : std_logic;
  signal slv_cmd_wr_bc4 : std_logic;

  type slv_cmd_attr_t is record
    Cmd         : ocp_CmdT;
    Addr        : adb3_ocp_addr_s;
    BurstLength : adb3_ocp_burst_s;
    Tag         : adb3_ocp_tag_s;
  end record;

  signal slv_cmd : slv_cmd_attr_t;

  -- Master Command FIFO
  constant MIG_ADDR_WIDTH : natural := app_bank_width + app_row_width + app_col_width; -- (4-byte addressing)
  constant MST_CMD_DEPTH  : natural := 5;  -- Probably could be 4 if we use most of FIFO ******
  constant MST_CMD_WIDTH  : natural := 5 + MIG_ADDR_WIDTH + ADB3_OCP_TAG_WIDTH;

  signal mst_cmd_wr       : std_logic;
  signal mst_cmd_adv      : std_logic;
  signal mst_cmd_not_full : std_logic;
  signal mst_cmd_full     : std_logic;
  signal mst_cmd_valid    : std_logic;
  signal mst_cmd_d        : std_logic_vector(MST_CMD_WIDTH-1 downto 0);
  signal mst_cmd_q        : std_logic_vector(MST_CMD_WIDTH-1 downto 0);

  signal app_cmd_wr   : std_logic;
  signal app_cmd_rd   : std_logic;
  signal app_cmd_bl8  : std_logic;
  signal app_cmd_addr : std_logic_vector(MIG_ADDR_WIDTH-1 downto 0);
  signal app_cmd_tag  : adb3_ocp_tag_s;

  -- Slave write data FIFO signals
  constant SLV_WDATA_DEPTH : natural := SLV_CMD_DEPTH + 1;
  constant SLV_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;

  signal slv_wdata_wr       : std_logic;
  signal slv_wdata_adv      : std_logic;
  signal slv_wdata_nfull    : std_logic;
  signal slv_wdata_not_full : std_logic;
  signal slv_wdata_empty    : std_logic;
  signal slv_wdata_nempty   : std_logic;
  signal slv_wdata_valid    : std_logic;
  signal slv_wdata_nvalid   : std_logic;
  signal slv_wdata_d        : std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);
  signal slv_wdata_q        : std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);

  type slv_wdata_attr_t is record
    DataByteEn : adb3_ocp_be_s;
    Data       : adb3_ocp_data_s;
  end record;

  signal slv_wdata : slv_wdata_attr_t;

  -- Master write data FIFO signals
  constant MST_WDATA_DEPTH  : natural := MST_CMD_DEPTH;
  constant MST_WDATA_WIDTH  : natural := SLV_WDATA_WIDTH;

  signal mst_wdata_wr       : std_logic;
  signal mst_wdata_adv      : std_logic;
  signal mst_wdata_not_full : std_logic;
  signal mst_wdata_valid    : std_logic;
  signal mst_wdata_d        : std_logic_vector(MST_WDATA_WIDTH-1 downto 0);
  signal mst_wdata_q        : std_logic_vector(MST_WDATA_WIDTH-1 downto 0);

  -- Command translation FSM
  type mcx_state_t is (mcx_s_idle, mcx_s_burst);

  signal mcx_state   : mcx_state_t;
  signal mcx_align   : std_logic;
  signal mcx_cmd_wr  : std_logic;
  signal mcx_cmd_rd  : std_logic;
  signal mcx_blen    : adb3_ocp_burst_s;
  signal mcx_tag     : adb3_ocp_tag_s;
  signal mcx_blast   : std_logic;
  signal mcx_mst_wr  : std_logic;
  signal mcx_en      : std_logic;
  signal mcx_sz      : std_logic;
  signal mcx_wr_end  : std_logic;
  signal mcx_addr    : std_logic_vector(MIG_ADDR_WIDTH-1 downto 0);

  signal n_mcx_bstart  : std_logic;
  signal n_mcx_state   : mcx_state_t;
  signal n_mcx_align   : std_logic;
  signal n_mcx_cmd_wr  : std_logic;
  signal n_mcx_cmd_rd  : std_logic;
  signal n_mcx_blen    : adb3_ocp_burst_s;
  signal n_mcx_tag     : adb3_ocp_tag_s;
  signal n_mcx_blast   : std_logic;
  signal n_mcx_mst_wr  : std_logic;
  signal n_mcx_en      : std_logic;
  signal n_mcx_sz      : std_logic;
  signal n_mcx_wr_end  : std_logic;
  signal n_mcx_addr    : std_logic_vector(MIG_ADDR_WIDTH-1 downto 0);

  -- Slave read resp FIFO signals
  -- Depth of 6 => 64 entries => 32 entries in rbt/rft FIFOs
  -- 32 is required so they don't fill up due to latency of read data from MIG core
  constant SLV_RDATA_DEPTH : natural := 6;
  constant SLV_RDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH + ADB3_OCP_TAG_WIDTH;

  signal slv_rdata_wr       : std_logic;
  signal slv_rdata_adv      : std_logic;
  signal slv_rdata_nfull    : std_logic;
--  signal slv_rdata_not_full : std_logic;
  signal slv_rdata_empty    : std_logic;
  signal slv_rdata_valid    : std_logic;
  signal slv_rdata_d        : std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);
  signal slv_rdata_q        : std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);
  signal slv_rdata_wlevel   : std_logic_vector(15 downto 0); -- *****

  -- Read burst tracking FIFO signals
  -- FIFO should be < half the size of the read resp data FIFO
  constant RBT_DEPTH : natural := SLV_RDATA_DEPTH;  -- ***** -1 when afifo and new_fifo sizes are the same *****
  constant RBT_WIDTH : natural := 1 + ADB3_OCP_TAG_WIDTH;

  signal rbt_wr       : std_logic;
  signal rbt_adv      : std_logic;
  signal rbt_not_full : std_logic;
  signal rbt_valid    : std_logic;
  signal rbt_d        : std_logic_vector(RBT_WIDTH-1 downto 0);
  signal rbt_q        : std_logic_vector(RBT_WIDTH-1 downto 0);
  signal rbt_q_bl8    : std_logic;
  signal rbt_q_tag    : adb3_ocp_tag_s;

  -- Read burst tracking counter signals
  signal rbt_bl_cnt  : std_logic;
  signal rbt_bl_comp : std_logic;

  -- Read full tracking FIFO signals
  -- FIFO should be < half the size of the read resp data FIFO
  constant RFT_DEPTH : natural := SLV_RDATA_DEPTH-1;
  constant RFT_WIDTH : natural := 1;

  signal rft_wr       : std_logic;
  signal rft_adv      : std_logic;
  signal rft_nfull    : std_logic;
  signal rft_not_full : std_logic;
  signal rft_empty    : std_logic;
  signal rft_valid    : std_logic;
  signal rft_d        : std_logic_vector(RFT_WIDTH - 1 downto 0);
  signal rft_q        : std_logic_vector(RFT_WIDTH - 1 downto 0);
  signal rft_q_bl8    : std_logic;
  signal rft_wlevel   : std_logic_vector(15 downto 0); -- *****

  -- Read full tracking counter signals
  signal rft_bl_cnt  : std_logic;
  signal rft_bl_comp : std_logic;

  -- Debug
  signal fifo_err : std_logic_vector(31 downto 0) := (others => '0');

  component adb3_ocp_new_fifo
  generic(
    ORDER     : in    natural range 1 to 6;
    WIDTH     : in    natural;
    EMPTY_VAL : in    std_logic := '0';
    EMPTY_LO  : in    natural := 0;
    EMPTY_HI  : in    natural := 0);
  port(
    rst       : in    std_logic;
    clk       : in    std_logic;
    wr        : in    std_logic;
    d         : in    std_logic_vector(WIDTH - 1 downto 0);
    adv       : in    std_logic;
    full      : out   std_logic;
    not_full  : out   std_logic;
    valid     : out   std_logic;
    q         : out   std_logic_vector(WIDTH - 1 downto 0));
  end component;

begin

  ---------------------------------------------------------------------------
  -- Command path
  ---------------------------------------------------------------------------

  ocp_ret_cmd_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      ocp_cmd_r  <= ADB3_OCP_DEFAULT_M2S.Cmd;
    elsif rising_edge(ocp_clk) then
      ocp_cmd_r  <= ocp_m2s.Cmd;
    end if;
  end process;

  ocp_ret_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      ocp_tag_r  <= ocp_m2s.Tag;
      ocp_bl_r   <= ocp_m2s.BurstLength;
      ocp_addr_r <= ocp_m2s.Addr;
    end if;
  end process;

  -- Decode if first MIG write command produced from OCP write command will be BL8 (2 data words required)
  -- This will occur if BurstLength > 1 and address is 32-byte aligned
  wr_1st_bl8 <= '1' when (ocp_cmd_r(0) = '1')and
                         (OR_Reduce(ocp_bl_r(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1')and(ocp_addr_r(4) = '0') else '0';

  -- Decode if first MIG read command produced from OCP read command will be BL8
  -- This will occur if BurstLength > 1 and address is 32-byte aligned
  rd_1st_bl8 <= '1' when (ocp_cmd_r(1) = '1')and
                         (OR_Reduce(ocp_bl_r(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '1')and(ocp_addr_r(4) = '0') else '0';

  -- Decode if first MIG write command produced from OCP write command will be BC4 (1 data word required)
  -- This will occur if BurstLength = 1 or address is not 32-byte aligned
  wr_1st_bc4 <= '1' when (ocp_cmd_r(0) = '1')and
                         ((OR_Reduce(ocp_bl_r(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '0')or(ocp_addr_r(4) = '1')) else '0';

  -- Slave command FIFO
  -- Advance read pointer if:
  --   slave command FIFO is not empty
  --   command translation FSM indicares start of OCP command processing
  slv_cmd_wr  <= '1' when (slv_cmd_not_full = '1')and(ocp_cmd_r /= OCP_CMD_IDLE) else '0';
  slv_cmd_d   <= wr_1st_bl8 & rd_1st_bl8 & wr_1st_bc4 & ocp_cmd_r & ocp_tag_r & ocp_bl_r & ocp_addr_r;
  slv_cmd_adv <= slv_cmd_valid and n_mcx_bstart;

  slv_cmd_fifo_i : afifo
    generic map(
      width  => SLV_CMD_WIDTH,
      order  => SLV_CMD_DEPTH,
      wpfl   => 2,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => ocp_clk,
      wd     => slv_cmd_d,
      wadv   => slv_cmd_wr,
      we     => open,
      wpf    => slv_cmd_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => mig_clk,
      rq     => slv_cmd_q,
      radv   => slv_cmd_adv,
      rflush => '0',
      re     => slv_cmd_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  slv_cmd_not_full  <= not(slv_cmd_nfull);
  ocp_s2m.CmdAccept <= slv_cmd_not_full;
  slv_cmd_valid     <= not(slv_cmd_empty);

  slv_cmd_wr_bl8 <= slv_cmd_q(SLV_CMD_WIDTH-1);
  slv_cmd_rd_bl8 <= slv_cmd_q(SLV_CMD_WIDTH-2);
  slv_cmd_wr_bc4 <= slv_cmd_q(SLV_CMD_WIDTH-3);

  slv_cmd <= (Cmd         => slv_cmd_q(SLV_CMD_WIDTH-4 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH),
              Tag         => slv_cmd_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH),
              BurstLength => slv_cmd_q(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_ADDR_WIDTH),
              Addr        => slv_cmd_q(ADB3_OCP_ADDR_WIDTH-1 downto 0));

  -- Error checking
  slv_cmd_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (slv_cmd_valid = '0')and(n_mcx_bstart = '1') then
        fifo_err(0) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to read from slv_cmd_fifo when empty"
          severity error;
      end if;
    end if;
  end process;


  -- Master command FIFO
  -- Advance read pointer if:
  --   master command FIFO is not empty
  --   rbt/rft FIFOs are not full
  --   MIG is ready to receive command (app_rdy) and data (app_wdf_rdy)
  mst_cmd_wr  <= mcx_mst_wr;
  mst_cmd_d   <= mcx_en & mcx_cmd_wr & mcx_cmd_rd & mcx_sz & mcx_wr_end & mcx_addr & mcx_tag;
  mst_cmd_adv <= mst_cmd_valid and rbt_not_full and rft_not_full and app_wdf_rdy and app_rdy;

  mst_cmd_fifo : adb3_ocp_new_fifo
  generic map(
    ORDER     => MST_CMD_DEPTH,
    WIDTH     => MST_CMD_WIDTH,
    EMPTY_VAL => '0',
    EMPTY_LO  => 0,
    EMPTY_HI  => MST_CMD_WIDTH)
  port map(
    rst      => mig_rst,
    clk      => mig_clk,
    wr       => mst_cmd_wr,
    d        => mst_cmd_d,
    adv      => mst_cmd_adv,
    full     => open,
    not_full => mst_cmd_not_full,
    valid    => mst_cmd_valid,
    q        => mst_cmd_q);

  -- Error checking
  mst_cmd_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      mst_cmd_full <= not(mst_cmd_not_full);
      -- State machine writes for 1 cycle after nearly full active
      if (mst_cmd_wr = '1')and(mst_cmd_full = '1') then
        fifo_err(4) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to write to mst_cmd_fifo when full"
          severity error;
      end if;
    end if;
  end process;

  app_cmd_wr   <= mst_cmd_q(MST_CMD_WIDTH-2);
  app_cmd_rd   <= mst_cmd_q(MST_CMD_WIDTH-3);
  app_cmd_bl8  <= mst_cmd_q(MST_CMD_WIDTH-4) and mst_cmd_adv;
  app_cmd_addr <= mst_cmd_q(MST_CMD_WIDTH-6 downto ADB3_OCP_TAG_WIDTH);
  app_cmd_tag  <= mst_cmd_q(ADB3_OCP_TAG_WIDTH-1 downto 0);

  -- ROW address width will change depending on density of memory (1Gb/2Gb).
  -- ROW address bits need to be the MS bits of OCP 4-byte address so MIG core for 2Gb will cope with both
  -- So OCP 4-byte address is organised as ROW, BANK, COL
--  app_addr_row  <= app_cmd_addr(MIG_ADDR_WIDTH-1 downto app_bank_width+app_col_width);
--  app_addr_bank <= app_cmd_addr(app_bank_width+app_col_width-1 downto app_col_width);
--  app_addr_col  <= app_cmd_addr(app_col_width-1 downto 0);

  -- MIG address app_addr is organised as RANK (unused), BANK, ROW, COL
--  app_addr <= '0' & app_addr_bank & app_addr_row & app_addr_col;
  -- Add 1 bit rank address to mst cmd FIFO address
  app_addr    <= '0' & app_cmd_addr;

  -- MIG write command = "000";
  -- MIG read  command = "001";
  app_en      <= mst_cmd_q(MST_CMD_WIDTH-1) and mst_cmd_adv;
  app_cmd     <= "00" & app_cmd_rd;
  app_sz      <= app_cmd_bl8;
  app_wdf_end <= mst_cmd_q(MST_CMD_WIDTH-5) and mst_cmd_adv;


  -- Read burst tracking FIFO
  -- Stores master cmd FIFO outputs app_cmd_bl8 and app_cmd_tag on mst_cmd_adv during read command
  -- Advance read when we have written all resp data for current command to slv_rdata FIFO
  -- rbt_not_full used to enable read of master cmd FIFO
  -- rbt FIFO will become full if the number of MIG read commands in flight is the same as its size
  -- rbt FIFO is half the size of slave rdata FIFO so in the worst case where all commands are BL8
  -- reads (2 resp data words expected) the slave rdata FIFO will be large enough

  rbt_wr  <= mst_cmd_adv and app_cmd_rd;
  rbt_d   <= app_cmd_bl8 & app_cmd_tag;
  rbt_adv <= rbt_valid and rbt_bl_comp and slv_rdata_wr;

  rbt_fifo : adb3_ocp_new_fifo
  generic map(
    ORDER     => RBT_DEPTH,
    WIDTH     => RBT_WIDTH)
  port map(
    rst      => mig_rst,
    clk      => mig_clk,
    wr       => rbt_wr,
    d        => rbt_d,
    adv      => rbt_adv,
    full     => open,
    not_full => rbt_not_full,
    valid    => rbt_valid,
    q        => rbt_q);

  rbt_q_bl8 <= rbt_q(RBT_WIDTH-1);
  rbt_q_tag <= rbt_q(ADB3_OCP_TAG_WIDTH-1 downto 0);

  -- Error checking
  rbt_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (rbt_wr = '1')and(rbt_not_full = '0') then
        fifo_err(8) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to write to rbt_fifo when full"
          severity error;
      end if;

      if (rbt_valid = '0')and(rbt_bl_comp = '1')and(slv_rdata_wr = '1') then
        fifo_err(9) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to read from rbt_fifo when empty"
          severity error;
      end if;

      if (slv_rdata_wr = '1')and(rbt_valid = '0') then
        fifo_err(10) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Resp data valid when rbt_fifo is empty"
          severity error;
      end if;
    end if;
  end process;

  -- Count read resp data until command is complete
  -- Compare rbt_bl_cnt with BL8 flag from rbt FIFO
  rbt_bl_cnt_p : process(mig_rst,mig_clk)
  begin
    if (mig_rst = '1') then
      rbt_bl_cnt <= '0';
    elsif rising_edge(mig_clk) then
      if (rbt_adv = '1') then
        rbt_bl_cnt <= '0';
      elsif (slv_rdata_wr = '1') then
        rbt_bl_cnt <= not(rbt_bl_cnt);
      end if;
    end if;
  end process;

  rbt_bl_comp <= '1' when (rbt_bl_cnt = rbt_q_bl8) else '0';


  ---------------------------------------------------------------------------
  -- Write data path
  ---------------------------------------------------------------------------

  -- Slave write data FIFO
  -- Advance read pointer if:
  --   slave write data FIFO is not empty
  --   command translation FSM indicares we are writing master write data
  slv_wdata_wr  <= slv_wdata_not_full and ocp_m2s.DataValid;
  slv_wdata_d   <= ocp_m2s.DataByteEn & ocp_m2s.Data;
  slv_wdata_adv <= slv_wdata_valid and n_mcx_mst_wr and n_mcx_cmd_wr;

  slv_wdata_fifo_i : afifo
    generic map(
      width  => SLV_WDATA_WIDTH,
      order  => SLV_WDATA_DEPTH,
      wpfl   => 2,
      wpfl2  => 0,
      rpel   => 1, -- Need to know if 2 write data words are present during BL8 write 1st cycle
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => ocp_clk,
      wd     => slv_wdata_d,
      wadv   => slv_wdata_wr,
      we     => open,
      wpf    => slv_wdata_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => mig_clk,
      rq     => slv_wdata_q,
      radv   => slv_wdata_adv,
      rflush => '0',
      re     => slv_wdata_empty,
      rpe    => slv_wdata_nempty,
      rf     => open,
      rudf   => open,
      rlevel => open);

  slv_wdata_not_full <= not(slv_wdata_nfull);
  ocp_s2m.DataAccept <= slv_wdata_not_full;
  slv_wdata_valid    <= not(slv_wdata_empty);
  slv_wdata_nvalid   <= not(slv_wdata_nempty);

  slv_wdata <= (DataByteEn => slv_wdata_q(SLV_WDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH),
                Data       => slv_wdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0));

  -- Error checking
  slv_wdata_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (slv_wdata_valid = '0')and(n_mcx_mst_wr = '1')and(n_mcx_cmd_wr = '1') then
        fifo_err(12) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to read from slv_wdata_fifo when empty"
          severity error;
      end if;
    end if;
  end process;


  -- Master wdata FIFO
  -- Advance read pointer if:
  --   master wdata FIFO is not empty
  --   rbt/rft FIFOs are not full
  --   we are doing a write command
  --   MIG is ready to receive command (app_rdy) and data (app_wdf_rdy)
  mst_wdata_wr  <= n_mcx_mst_wr and n_mcx_cmd_wr;
  mst_wdata_d   <= not(slv_wdata.DataByteEn) & slv_wdata.Data;
  mst_wdata_adv <= mst_wdata_valid and app_cmd_wr and rbt_not_full and rft_not_full and app_wdf_rdy and app_rdy;

  mst_wdata_fifo : adb3_ocp_new_fifo
  generic map(
    ORDER     => MST_WDATA_DEPTH,
    WIDTH     => MST_WDATA_WIDTH)
  port map(
    rst      => mig_rst,
    clk      => mig_clk,
    wr       => mst_wdata_wr,
    d        => mst_wdata_d,
    adv      => mst_wdata_adv,
    full     => open,
    not_full => mst_wdata_not_full,
    valid    => mst_wdata_valid,
    q        => mst_wdata_q);

  -- Error checking
  mst_wdata_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (mst_wdata_wr = '1')and(mst_wdata_not_full = '0') then
        fifo_err(16) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to write to mst_wdata_fifo when nearly full"
          severity error;
      end if;

      if (mst_wdata_valid = '0')and(mst_cmd_adv = '1')and(app_cmd_wr = '1') then
        fifo_err(17) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to issue MIG write command when mst_wdata_fifo when empty"
          severity error;
      end if;
    end if;
  end process;

  app_wdf_wren <= mst_wdata_adv;
  app_wdf_mask <= mst_wdata_q(MST_WDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  app_wdf_data <= mst_wdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0);


  ---------------------------------------------------------------------------
  -- Command Translation FSM
  ---------------------------------------------------------------------------

  gen_mcx_state : process(mig_rst, mig_clk)
  begin
    if mig_rst = '1' then
      mcx_state   <= mcx_s_idle;
      mcx_align   <= '-';
      mcx_cmd_wr  <= '-';
      mcx_cmd_rd  <= '-';
      mcx_blen    <= (others => '-');
      mcx_tag     <= (others => '-');
      mcx_blast   <= '-';
      mcx_mst_wr  <= '0';
      mcx_en      <= '-';
      mcx_sz      <= '-';
      mcx_wr_end  <= '-';
      mcx_addr    <= (others => '-');
    elsif rising_edge(mig_clk) then
      mcx_state   <= n_mcx_state;
      mcx_align   <= n_mcx_align;
      mcx_cmd_wr  <= n_mcx_cmd_wr;
      mcx_cmd_rd  <= n_mcx_cmd_rd;
      mcx_blen    <= n_mcx_blen;
      mcx_tag     <= n_mcx_tag;
      mcx_blast   <= n_mcx_blast;
      mcx_mst_wr  <= n_mcx_mst_wr;
      mcx_en      <= n_mcx_en;
      mcx_sz      <= n_mcx_sz;
      mcx_wr_end  <= n_mcx_wr_end;
      mcx_addr    <= n_mcx_addr;
    end if;
  end process;


  gen_n_mcx_state : process(
    mcx_state,
    mcx_align,
    mcx_cmd_wr,
    mcx_cmd_rd,
    mcx_blen,
    mcx_tag,
    mcx_blast,
    mcx_en,
    mcx_sz,
    mcx_wr_end,
    mcx_addr,
    slv_cmd_valid,
    slv_cmd,
    slv_cmd_wr_bl8,
    slv_cmd_wr_bc4,
    slv_cmd_rd_bl8,
    slv_wdata_valid,
    slv_wdata_nvalid,
    mst_cmd_not_full,
    mst_wdata_not_full)
  begin
    case mcx_state is
      when mcx_s_idle =>
        -- First data burst in first command, there are the following cases:
        --     32-byte aligned, burst count: ..,6,5,4,3,2: BL8 write (cycle 1 of 2) => not aligned
        --                                   ..,6,5,4,3,2: BL8 Read cycle           => aligned/last burst
        --     32-byte aligned, burst count:            1: BC4 write/read cycle     => end of burst
        -- Not 32-byte aligned, burst count: ..,5,4,3,2,1: BC4 write/read cycle     => aligned/last burst
        -- or return to mcx_s_idle
        if (slv_cmd_valid = '1')and(slv_cmd_wr_bl8 = '1')and(slv_wdata_nvalid = '1')and
           (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
          -- BL8 write (cycle 1 of 2) with 2 data words available
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '1';
          n_mcx_blen   <= slv_cmd.BurstLength;
          n_mcx_tag    <= slv_cmd.Tag;
          n_mcx_align  <= not(slv_cmd.Addr(4));
          n_mcx_cmd_wr <= '1';
          n_mcx_cmd_rd <= '0';
          n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
          n_mcx_mst_wr <= '1';
          n_mcx_en     <= '1';
          n_mcx_sz     <= '1';
          n_mcx_wr_end <= '0';
          n_mcx_blast  <= '0'; -- BL8 write (cycle 1 of 2), cannot be last burst of write

        elsif (slv_cmd_valid = '1')and(slv_cmd_wr_bc4 = '1')and(slv_wdata_valid = '1')and
              (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
          -- BC4 write cycle with 1 data word available
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '1';
          n_mcx_blen   <= slv_cmd.BurstLength;
          n_mcx_tag    <= slv_cmd.Tag;
          n_mcx_align  <= not(slv_cmd.Addr(4));
          n_mcx_cmd_wr <= '1';
          n_mcx_cmd_rd <= '0';
          n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
          n_mcx_mst_wr <= '1';
          n_mcx_en     <= '1';
          n_mcx_sz     <= '0';
          n_mcx_wr_end <= '1';
          if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            n_mcx_blast <= '1';
          else
            n_mcx_blast <= '0';
          end if;

        elsif (slv_cmd_valid = '1')and(slv_cmd_wr_bl8 = '0')and(slv_cmd_wr_bc4 = '0')and
              (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
          -- BL8 read cycle/BC4 read cycle
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '1';
          n_mcx_blen   <= slv_cmd.BurstLength;
          n_mcx_tag    <= slv_cmd.Tag;
          n_mcx_align  <= not(slv_cmd.Addr(4));
          n_mcx_cmd_wr <= '0';
          n_mcx_cmd_rd <= '1';
          n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
          n_mcx_mst_wr <= '1';
          n_mcx_en     <= '1';
          n_mcx_wr_end <= '0';
          if (slv_cmd_rd_bl8 = '1') then -- BL8 read cycle
            n_mcx_sz <= '1';
            if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          else -- BC4 read cycle
            n_mcx_sz <= '0';
            if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          end if;

        else
          -- Return to mcx_s_idle state
          n_mcx_state  <= mcx_s_idle;
          n_mcx_bstart <= '0';
          n_mcx_blen   <= (others => '-');
          n_mcx_tag    <= (others => '-');
          n_mcx_align  <= '-';
          n_mcx_cmd_wr <= '-';
          n_mcx_cmd_rd <= '-';
          n_mcx_addr   <= (others => '-');
          n_mcx_mst_wr <= '0';
          n_mcx_en     <= '-';
          n_mcx_sz     <= '-';
          n_mcx_wr_end <= '-';
          n_mcx_blast  <= '-';
        end if;

      -- Process burst of active command in command FIFO
      when mcx_s_burst =>
        -- First data burst in next command or next/last data burst in current command
        if (mcx_blast = '1') then
          -- First data burst in next command, there are the following cases:
          --     32-byte aligned, burst count: ..,6,5,4,3,2: BL8 write (cycle 1 of 2) => not aligned
          --                                   ..,6,5,4,3,2: BL8 Read cycle           => aligned/last burst
          --     32-byte aligned, burst count:            1: BC4 write/read cycle     => end of burst
          -- Not 32-byte aligned, burst count: ..,5,4,3,2,1: BC4 write/read cycle     => aligned/last burst
          -- or return to mcx_s_idle
          if (slv_cmd_valid = '1')and(slv_cmd_wr_bl8 = '1')and(slv_wdata_nvalid = '1')and
             (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- BL8 write (cycle 1 of 2) with 2 data words available
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '1';
            n_mcx_blen   <= slv_cmd.BurstLength;
            n_mcx_tag    <= slv_cmd.Tag;
            n_mcx_align  <= not(slv_cmd.Addr(4));
            n_mcx_cmd_wr <= '1';
            n_mcx_cmd_rd <= '0';
            n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
            n_mcx_mst_wr <= '1';
            n_mcx_en     <= '1';
            n_mcx_sz     <= '1';
            n_mcx_wr_end <= '0';
            n_mcx_blast  <= '0'; -- BL8 write (cycle 1 of 2), cannot be last burst of write

          elsif (slv_cmd_valid = '1')and(slv_cmd_wr_bc4 = '1')and(slv_wdata_valid = '1')and
                (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- BC4 write cycle with 1 data word available
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '1';
            n_mcx_blen   <= slv_cmd.BurstLength;
            n_mcx_tag    <= slv_cmd.Tag;
            n_mcx_align  <= not(slv_cmd.Addr(4));
            n_mcx_cmd_wr <= '1';
            n_mcx_cmd_rd <= '0';
            n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
            n_mcx_mst_wr <= '1';
            n_mcx_en     <= '1';
            n_mcx_sz     <= '0';
            n_mcx_wr_end <= '1';
            if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;

          elsif (slv_cmd_valid = '1')and(slv_cmd_wr_bl8 = '0')and(slv_cmd_wr_bc4 = '0')and
                (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- BL8 read cycle/BC4 read cycle
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '1';
            n_mcx_blen   <= slv_cmd.BurstLength;
            n_mcx_tag    <= slv_cmd.Tag;
            n_mcx_align  <= not(slv_cmd.Addr(4));
            n_mcx_cmd_wr <= '0';
            n_mcx_cmd_rd <= '1';
            n_mcx_addr   <= slv_cmd.Addr(MIG_ADDR_WIDTH+1 downto 2);  -- 4-byte addressing
            n_mcx_mst_wr <= '1';
            n_mcx_en     <= '1';
            n_mcx_wr_end <= '0';
            if (slv_cmd_rd_bl8 = '1') then -- BL8 read cycle
              n_mcx_sz <= '1';
              if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then
                n_mcx_blast <= '1';
              else
                n_mcx_blast <= '0';
              end if;
            else -- BC4 read cycle
              n_mcx_sz <= '0';
              if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
                n_mcx_blast <= '1';
              else
                n_mcx_blast <= '0';
              end if;
            end if;

          else
            -- Return to mcx_s_idle state
            n_mcx_state  <= mcx_s_idle;
            n_mcx_bstart <= '0';
            n_mcx_blen   <= (others => '-');
            n_mcx_tag    <= (others => '-');
            n_mcx_align  <= '-';
            n_mcx_cmd_wr <= '-';
            n_mcx_cmd_rd <= '-';
            n_mcx_addr   <= (others => '-');
            n_mcx_mst_wr <= '0';
            n_mcx_en     <= '-';
            n_mcx_sz     <= '-';
            n_mcx_wr_end <= '-';
            n_mcx_blast  <= '-';
          end if;
        else
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '0';
          n_mcx_cmd_wr <= mcx_cmd_wr;
          n_mcx_cmd_rd <= mcx_cmd_rd;
          n_mcx_tag    <= mcx_tag;

          -- Next/last data burst in current command, there are the following cases:
          -- Last cycle not 32-byte aligned, burst count: ..,6,5,4,3,2: BL8 write (cycle 1 of 2) = aligned
          --                                                         1: BC4 write/read cycle     = aligned/last burst
          --                                              ..,6,5,4,3,2: BL8 read cycle           = aligned
          -- last cycle     32-byte aligned, burst count: ..,5,4,3,2,1: BL8 write (cycle 2 of 2) = unaligned/last burst
          --                                              ..,6,5,4,3,2: BL8 read cycle           = aligned
          --
          -- or wait until conditions are satisfied
          if (mcx_align = '0')and(mcx_cmd_wr = '1')and(mcx_blen /= std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH)))and
             (slv_wdata_nvalid = '1')and(mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- Not aligned BL8 write (cycle 1 of 2) with 2 data words available
            n_mcx_en     <= '1';
            n_mcx_align  <= '1';
            n_mcx_addr   <= mcx_addr + 4;  -- 4-byte addressing
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_mst_wr <= '1';
            n_mcx_sz     <= '1';
            n_mcx_wr_end <= '0';
            n_mcx_blast  <= '0'; -- BL8 write (cycle 1 of 2), cannot be last burst of write. BC4/BL8a

          elsif (mcx_align = '0')and(mcx_cmd_wr = '1')and(mcx_blen = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH)))and
                (slv_wdata_valid = '1')and(mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- Not aligned BC4 write cycle with 1 data word available
            n_mcx_en     <= '1';
            n_mcx_align  <= '1';
            n_mcx_addr   <= mcx_addr + 4;  -- 4-byte addressing
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_mst_wr <= '1';
            n_mcx_sz     <= '0';
            n_mcx_wr_end <= '1';
            n_mcx_blast  <= '1'; -- BC4 write, last burst. Last: BC4/BL8b,BC4 = 2

          elsif (mcx_align = '0')and(mcx_cmd_wr = '0')and
                (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- Not aligned BL8 read cycle/BC4 read cycle
            n_mcx_en     <= '1';
            n_mcx_align  <= '1';
            n_mcx_addr   <= mcx_addr + 4;  -- 4-byte addressing
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_mst_wr <= '1';
            n_mcx_wr_end <= '0';

            if (mcx_blen = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then -- BC4 read
              n_mcx_sz    <= '0';
              n_mcx_blast <= '1'; -- BC4 read, last burst. Last: BC4,BC4 = 2
            else -- BL8 read
              n_mcx_sz <= '1';
              if (mcx_blen = std_logic_vector(to_unsigned(3,ADB3_OCP_BURST_WIDTH))) then
                n_mcx_blast <= '1'; -- BL8 read, last burst. Last: BC4,BL8 = 3
              else
                n_mcx_blast <= '0';
              end if;
            end if;

          elsif (mcx_align = '1')and(mcx_cmd_wr = '1')and(slv_wdata_valid = '1')and
                (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- Aligned BL8 write (cycle 2 of 2) with 1 data word available
            n_mcx_en     <= '0';
            n_mcx_align  <= '0';
            n_mcx_addr   <= mcx_addr + 4;  -- 4-byte addressing
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_mst_wr <= '1';
            n_mcx_sz     <= '0';
            n_mcx_wr_end <= '1';

            if (mcx_blen = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then -- Last: BL8a,BL8b = 2
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;

          elsif (mcx_align = '1')and(mcx_cmd_rd = '1')and
                (mst_cmd_not_full = '1')and(mst_wdata_not_full = '1') then
            -- Aligned BL8 read cycle/BC4 read cycle
            n_mcx_en     <= '1';
            n_mcx_addr   <= mcx_addr + 8;  -- 4-byte addressing
            n_mcx_blen   <= mcx_blen-2;
            n_mcx_mst_wr <= '1';
            n_mcx_wr_end <= '0';

            if (mcx_blen = std_logic_vector(to_unsigned(3,ADB3_OCP_BURST_WIDTH))) then -- BC4 read. Last: BL8,BC4 = 3
              n_mcx_align <= '0';
              n_mcx_sz    <= '0';
              n_mcx_blast <= '1'; -- BC4 read, last burst
            else -- BL8 read cycle
              n_mcx_align <= '1';
              n_mcx_sz    <= '1';

              if (mcx_blen = std_logic_vector(to_unsigned(4,ADB3_OCP_BURST_WIDTH))) then -- Last: BL8,BL8 = 4
                n_mcx_blast <= '1';
              else
                n_mcx_blast <= '0';
              end if;
            end if;

          else
            -- Wait until conditions are satisfied
            n_mcx_blen    <= mcx_blen;
            n_mcx_align   <= mcx_align;
            n_mcx_addr    <= mcx_addr;
            n_mcx_mst_wr  <= '0';
            n_mcx_en      <= mcx_en;
            n_mcx_sz      <= mcx_sz;
            n_mcx_wr_end  <= mcx_wr_end;
            n_mcx_blast   <= mcx_blast;
          end if;
        end if;
    end case;
  end process;


  ---------------------------------------------------------------------------
  -- Read resp data path
  ---------------------------------------------------------------------------

  -- OCP read resp FIFO
  -- Must accept read resp data if available
  -- Advance read pointer if:
  --   read resp FIFO is not empty
  --   slave s2m OCP is ready to receive resp data
  slv_rdata_wr  <= app_rd_data_valid;
  slv_rdata_d   <= app_rd_data & rbt_q_tag;
  slv_rdata_adv <= slv_rdata_valid and ocp_m2s.RespAccept;

  slv_rdata_fifo_i : afifo
    generic map(
      width  => SLV_RDATA_WIDTH,
      order  => SLV_RDATA_DEPTH,
      wpfl   => 2,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => mig_clk,
      wd     => slv_rdata_d,
      wadv   => slv_rdata_wr,
      we     => open,
      wpf    => slv_rdata_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => slv_rdata_wlevel,  -- *******
      rk     => ocp_clk,
      rq     => slv_rdata_q,
      radv   => slv_rdata_adv,
      rflush => '0',
      re     => slv_rdata_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

--  slv_rdata_not_full <= not(slv_rdata_nfull);
  slv_rdata_valid    <= not(slv_rdata_empty);

  ocp_s2m.Resp <= OCP_RESP_VALID when (slv_rdata_valid = '1') else OCP_RESP_NONE;
  ocp_s2m.Data <= slv_rdata_q(SLV_RDATA_WIDTH-1 downto ADB3_OCP_TAG_WIDTH);
  ocp_s2m.Tag  <= slv_rdata_q(ADB3_OCP_TAG_WIDTH-1 downto 0);

  -- Error checking
  slv_rdata_err_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (slv_rdata_wr = '1')and(slv_rdata_nfull = '1') then
        fifo_err(20) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to write to slv_rdata_fifo when nearly full"
          severity error;
      end if;
    end if;
  end process;

  -- Read full tracking FIFO
  -- Stores master cmd FIFO output app_cmd_bl8 on mst_cmd_adv during read command
  -- Advance read when we have read all resp data for current command from slv_rdata FIFO
  -- rft_not_full used to enable read of master cmd FIFO
  -- rft FIFO will become full if the number of MIG read commands in the rft FIFO + the number of
  -- MIG read commands in flight is the same as its size
  -- rft FIFO is half the size of slave rdata FIFO so in the worst case where all commands are BL8
  -- reads (2 resp data words expected) the slave rdata FIFO will be large enough
  rft_wr   <= mst_cmd_adv and app_cmd_rd;
  rft_d(0) <= app_cmd_bl8;
  rft_adv  <= rft_valid and rft_bl_comp and slv_rdata_adv;

  rft_fifo_i : afifo
    generic map(
      width  => RFT_WIDTH,
      order  => RFT_DEPTH,
      wpfl   => 2,
      wpfl2  => 0,
      rpel   => 0,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map(
      rst    => ocp_rst,   -- Asynchronous to wr and rd clocks
      wk     => mig_clk,
      wd     => rft_d,
      wadv   => rft_wr,
      we     => open,
      wpf    => rft_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => rft_wlevel,  -- *******
      rk     => ocp_clk,
      rq     => rft_q,
      radv   => rft_adv,
      rflush => '0',
      re     => rft_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  rft_not_full <= not(rft_nfull);
  rft_valid    <= not(rft_empty);

  rft_q_bl8 <= rft_q(RFT_WIDTH-1);

  -- Error checking
  rft_err1_p : process(mig_clk)
  begin
    if rising_edge(mig_clk) then
      if (rft_wr = '1')and(rft_nfull = '1') then
        fifo_err(24) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to write to rft_fifo_i when nearly full"
          severity error;
      end if;
    end if;
  end process;

  rft_err2_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (slv_rdata_adv = '1')and(rft_valid = '0') then
        fifo_err(25) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): OCP Resp valid when rft_fifo is empty"
          severity error;
      end if;

      if (rft_valid = '0')and(rft_bl_comp = '1')and(slv_rdata_adv = '1') then
        fifo_err(26) <= '1';
        assert false
          report "(adb3_ocp_ocp2ddr3): Attempt to read from rft_fifo when empty"
          severity error;
      end if;
    end if;
  end process;

  -- Count read resp data until command is complete
  -- Compare rft_bl_cnt with BL8 flag from rft FIFO
  rft_bl_cnt_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      rft_bl_cnt <= '0';
    elsif rising_edge(ocp_clk) then
      if (rft_adv = '1') then
        rft_bl_cnt <= '0';
      elsif (slv_rdata_adv = '1') then
        rft_bl_cnt <= not(rft_bl_cnt);
      end if;
    end if;
  end process;

  rft_bl_comp <= '1' when (rft_bl_cnt = rft_q_bl8) else '0';





  -- Debug signals
  debug_info <= fifo_err; --(others => '0');
end;