--==============================================================================
--
--     File Name: adb3_ocp_split_b.vhd
--       Version: 0.2
--          Date: 09 Feb 2011
--         Model: synth
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP address space de-mux (blocking)
--
--   Description: Component giving single OCP slave port control over
--                multiple OCP master ports
--                De-muxes OCP address space using address range table
--                generic input
--
--   Limitations:
--
--         Notes: blocking component
--                Re-time of de-mux input selectable using generic retime_slave
--                Re-time of de-mux outputs selectable using generic retime_masters
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 02 Dec 2010: Initial design
--                0.2, TW,  09 Feb 2011: Now references adb3_ocp_comp package
--                  instead of adb3_target_pkg.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library Work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;

entity adb3_ocp_split_b is
generic (
  addr_range_table : addr_range_table_t;
  retime_slave     : boolean := false;
  retime_masters   : boolean := false);
port(
  ocp_rst     : in    std_logic;
  ocp_clk     : in    std_logic;
  -- OCP primary port
  slave_m2s   : in    adb3_ocp_m2sT;
  slave_s2m   : out   adb3_ocp_s2mT;
  -- OCP secondary ports
  masters_m2s : out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1);
  masters_s2m : in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1);
  -- Debug
  debug_info  : out   std_logic_vector(31 downto 0));
end;


architecture synth of adb3_ocp_split_b is

  constant DEMUX_OUTPUTS : natural := addr_range_table'length;
  constant TOTAL_OUTPUTS : natural := DEMUX_OUTPUTS + 1;

  subtype act_port_s  is std_logic_vector(DEMUX_OUTPUTS-1 downto 0);
  subtype all_port_s  is std_logic_vector(TOTAL_OUTPUTS-1 downto 0);

  constant ACT_PORT_M2S_DEF : all_port_s := std_logic_vector(to_unsigned(1,TOTAL_OUTPUTS));

  signal slave_m2s_ret : adb3_ocp_m2sT;
  signal slave_s2m_ret : adb3_ocp_s2mT;

  signal masters_m2s_int : adb3_ocp_m2sT_array(0 to TOTAL_OUTPUTS-1);
  signal masters_s2m_int : adb3_ocp_s2mT_array(0 to TOTAL_OUTPUTS-1);
  signal masters_m2s_out : adb3_ocp_m2sT_array(0 to DEMUX_OUTPUTS-1);
  signal masters_s2m_out : adb3_ocp_s2mT_array(0 to DEMUX_OUTPUTS-1);

  signal act_port_m2s   : all_port_s := ACT_PORT_M2S_DEF;
  signal switched_m2s   : std_logic := '0';

  signal master_m2s_err : adb3_ocp_m2sT;
  signal master_s2m_err : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  signal err_wr    : std_logic := '0';
  signal err_rd    : std_logic := '0';
  signal err_data  : std_logic := '0';
  signal err_resp  : std_logic := '0';
  signal err_bl    : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');

  signal writing_m2s     : std_logic := '0';
  signal reading_m2s     : std_logic := '0';
  signal wburst_m2s      : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal rburst_m2s      : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal wdata_last_m2s  : std_logic := '0';
  signal rdata_last_m2s  : std_logic := '0';
  signal wburst_comp_m2s : std_logic := '0';
  signal rburst_comp_m2s : std_logic := '0';

  signal en_bit  : std_logic;
  signal en_cmd  : ocp_CmdT;
  signal en_resp : ocp_RespT;

  signal act_port_m2s_err : std_logic;
  signal act_port_m2s_cnt : std_logic_vector(7 downto 0) := (others => '0');

begin

  -- Address range table checking
  -- Check that address ranges do not overlap
  -- Check that mask address is aligned with base address
  gen_addr_range_check1_g: for i in 0 to DEMUX_OUTPUTS-1 generate
  begin
    gen_addr_range_check2_g: for j in 0 to DEMUX_OUTPUTS-1 generate
      constant RANGE_I : adb3_ocp_addr_s := addr_range_table(i).addr_base and not(addr_range_table(i).addr_mask);
      constant RANGE_J : adb3_ocp_addr_s := addr_range_table(j).addr_base and not(addr_range_table(j).addr_mask);
    begin
      assert not((RANGE_I = RANGE_J)and(i /= j))
        report "(adb3_ocp_split_b): Address range entries " & integer'image(j) &
               " and " & integer'image(i) & " overlap in addr_range_table generic."
        severity failure;
    end generate;

    assert not(OR_reduce(addr_range_table(i).addr_base and addr_range_table(i).addr_mask) = '1')
      report "(adb3_ocp_split_b): Mask/Base address alignment problem for entry " & integer'image(i) &
             " in addr_range_table generic."
      severity failure;
  end generate;


  retime_slave_g : if (retime_slave) generate
    -- Re-time OCP slave port
    adb3_ocp_retime_nb_i : adb3_ocp_retime_nb
    port map(
      ocp_rst    => ocp_rst,        -- in    std_logic
      ocp_clk    => ocp_clk,        -- in    std_logic
      -- OCP slave port
      slave_m2s  => slave_m2s,      -- in    adb3_ocp_m2sT
      slave_s2m  => slave_s2m,      -- out   adb3_ocp_s2mT
      -- OCP master port
      master_m2s => slave_m2s_ret,  -- out   adb3_ocp_m2sT
      master_s2m => slave_s2m_ret); -- in    adb3_ocp_s2mT)
  end generate;

  route_slave_g : if not(retime_slave) generate
    -- Route OCP slave port
    slave_m2s_ret <= slave_m2s;
    slave_s2m <= slave_s2m_ret;
  end generate;


  routing_m2s_g : for n in 0 to DEMUX_OUTPUTS generate
    -- De-Mux command group m2s single slave OCP port to m2s master OCP ports using act_port_m2s
    masters_m2s_int(n).Cmd         <= slave_m2s_ret.Cmd and en_cmd when act_port_m2s(n) = '1' else ADB3_OCP_DEFAULT_M2S.Cmd;
    masters_m2s_int(n).Addr        <= slave_m2s_ret.Addr;
    masters_m2s_int(n).BurstLength <= slave_m2s_ret.BurstLength;
    masters_m2s_int(n).Tag         <= slave_m2s_ret.Tag;

    -- Mux command accept s2m master OCP ports to s2m single slave OCP port using act_port_m2s
    slave_s2m_ret.CmdAccept <= masters_s2m_int(n).CmdAccept and en_bit when act_port_m2s(n) = '1' else ADB3_OCP_Z_S2M.CmdAccept;

    -- De-Mux data write group m2s single slave OCP port to m2s master OCP ports using act_port_m2s
    masters_m2s_int(n).Data       <= slave_m2s_ret.Data;
    masters_m2s_int(n).DataByteEn <= slave_m2s_ret.DataByteEn;
    masters_m2s_int(n).DataValid  <= slave_m2s_ret.DataValid and writing_m2s when act_port_m2s(n) = '1' else ADB3_OCP_DEFAULT_M2S.DataValid;

    -- Mux data accept s2m master OCP ports to s2m single slave OCP port using act_port_m2s
    slave_s2m_ret.DataAccept <= masters_s2m_int(n).DataAccept and writing_m2s when act_port_m2s(n) = '1' else ADB3_OCP_Z_S2M.DataAccept;

    -- Mux data read group s2m master OCP ports to s2m single slave OCP port using act_port_m2s
    slave_s2m_ret.Data <= masters_s2m_int(n).Data             when act_port_m2s(n) = '1' else ADB3_OCP_Z_S2M.Data;
    slave_s2m_ret.Tag  <= masters_s2m_int(n).Tag              when act_port_m2s(n) = '1' else ADB3_OCP_Z_S2M.Tag;
    slave_s2m_ret.Resp <= masters_s2m_int(n).Resp and en_resp when act_port_m2s(n) = '1' else ADB3_OCP_Z_S2M.Resp;

    -- De-Mux data accept m2s single slave OCP port to m2s master OCP ports using act_port_m2s
    masters_m2s_int(n).RespAccept <= slave_m2s_ret.RespAccept when act_port_m2s(n) = '1' else ADB3_OCP_DEFAULT_M2S.RespAccept;
  end generate;

  -- Connect masters_m2s_int to master ports and master error port
  masters_m2s_out <= masters_m2s_int(0 to DEMUX_OUTPUTS-1);
  master_m2s_err  <= masters_m2s_int(DEMUX_OUTPUTS);

  -- Commect master ports and master error port to masters_s2m_int
  masters_s2m_int(0 to DEMUX_OUTPUTS-1) <= masters_s2m_out;
  masters_s2m_int(DEMUX_OUTPUTS)        <= master_s2m_err;


  -- OCP cmd/write data end-point processing block
  -- Controls command/write data acceptance to re-generate their correct relative order
  process(ocp_clk)
    variable switched_m2s_var : std_logic := '0';
    variable act_port_m2s_var : act_port_s := std_logic_vector(to_unsigned(1,DEMUX_OUTPUTS));
  begin
    if rising_edge(ocp_clk) then
      -- Look for next active m2s port
      if (switched_m2s = '0') then
        -- Set to 0 so that if there is no activity on this cycle, we keep looking
        switched_m2s_var := '0';
        -- Only set switched on valid command
        if (slave_m2s_ret.Cmd = ocp_Cmd_WRITE)or(slave_m2s_ret.Cmd = ocp_Cmd_READ) then
          -- Loop round all address ranges to find active one
          act_port_l : for n in 0 to DEMUX_OUTPUTS-1 loop
            -- Test for address match in table entry n
            if (slave_m2s_ret.Addr and not(addr_range_table(n).addr_mask)) = (addr_range_table(n).addr_base and not(addr_range_table(n).addr_mask)) then
              act_port_m2s_var(n) := '1';
            else
              act_port_m2s_var(n) := '0';
            end if;
          end loop;
          switched_m2s_var := '1';
        end if;

        -- Set MSB of vectors on invalid address
        if (OR_reduce(act_port_m2s_var) = '1') then
          act_port_m2s <= '0' & act_port_m2s_var;
        else
          act_port_m2s <= '1' & act_port_m2s_var;
        end if;

        switched_m2s <= switched_m2s_var;
      end if;

      -- Assign default values
      wburst_comp_m2s <= '0';
      rburst_comp_m2s <= '0';

      -- Use start of OCP write command to set writing and last write data to clear
      -- CmdAccept qualified by switched and not writing/reading
      -- DataAccept qualified by writing
      if (slave_m2s_ret.Cmd = OCP_CMD_WRITE)and(slave_s2m_ret.CmdAccept = '1') then
        writing_m2s <= '1';
      elsif ((wdata_last_m2s = '1')and(slave_m2s_ret.DataValid = '1')and(slave_s2m_ret.DataAccept = '1'))or(wburst_comp_m2s = '1') then
        writing_m2s <= '0';
        switched_m2s <= '0';
      end if;

      -- Process m2s write data
      -- Count each burst of OCP write data
      -- CmdAccept qualified by switched and not writing/reading
      -- DataValid qualified by writing
      if (slave_m2s_ret.Cmd = OCP_CMD_WRITE)and(slave_s2m_ret.CmdAccept = '1') then
        if (slave_m2s_ret.DataValid = '1')and(slave_s2m_ret.DataAccept = '1') then
          wburst_m2s <= slave_m2s_ret.BurstLength-1;
          if (slave_m2s_ret.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            wburst_comp_m2s <= '1';
          end if;
        else
          wburst_m2s <= slave_m2s_ret.BurstLength;
        end if;
      else
        if (slave_m2s_ret.DataValid = '1')and(slave_s2m_ret.DataAccept = '1') then
          wburst_m2s <= wburst_m2s-1;
        end if;
      end if;

      -- Use start of OCP read command to set reading and last read data to clear
      -- Cmd qualified by switched and not writing/reading
      if (slave_m2s_ret.Cmd = OCP_CMD_READ)and(slave_s2m_ret.CmdAccept = '1') then
        reading_m2s <= '1';
      elsif ((rdata_last_m2s = '1')and(slave_m2s_ret.RespAccept = '1')and(slave_s2m_ret.Resp = OCP_RESP_VALID))or(rburst_comp_m2s = '1') then
        reading_m2s <= '0';
        switched_m2s <= '0';
      end if;

      -- Process m2s read data
      -- Count each burst of OCP read data
      -- CmdAccept qualified by switched and not writing/reading
      -- Resp qualified by reading
      if (slave_m2s_ret.Cmd = OCP_CMD_READ)and(slave_s2m_ret.CmdAccept = '1') then
        if (slave_m2s_ret.RespAccept = '1')and(slave_s2m_ret.Resp = OCP_RESP_VALID) then
          rburst_m2s <= slave_m2s_ret.BurstLength-1;
          if (slave_m2s_ret.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            rburst_comp_m2s <= '1';
          end if;
        else
          rburst_m2s <= slave_m2s_ret.BurstLength;
        end if;
      else
        if (slave_m2s_ret.RespAccept = '1')and(slave_s2m_ret.Resp = OCP_RESP_VALID) then
          rburst_m2s <= rburst_m2s-1;
        end if;
      end if;
    end if;
  end process;

  en_bit  <= switched_m2s and not(writing_m2s or reading_m2s);
  en_cmd  <= en_bit & en_bit & en_bit;
  en_resp <= reading_m2s & reading_m2s;

  wdata_last_m2s <= '1' when wburst_m2s = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH)) else '0';
  rdata_last_m2s <= '1' when rburst_m2s = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH)) else '0';


  -- Connect Error OCP port for out of range transactions
  master_s2m_err_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      -- Command Accept
      if ((master_m2s_err.Cmd = OCP_CMD_WRITE)or(master_m2s_err.Cmd = OCP_CMD_READ))and(master_s2m_err.CmdAccept = '0') then
        master_s2m_err.CmdAccept <= '1';
        if (master_m2s_err.Cmd = OCP_CMD_WRITE) then
          err_bl <= master_m2s_err.BurstLength;
          err_wr <= '1';
          err_rd <= '0';
        end if;
        if (master_m2s_err.Cmd = OCP_CMD_READ) then
          master_s2m_err.Data <= X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE";
          master_s2m_err.Tag  <= master_m2s_err.Tag;
          err_bl <= master_m2s_err.BurstLength;
          err_rd <= '1';
          err_wr <= '0';
        end if;
      else
        master_s2m_err.CmdAccept <= '0';
      end if;

      -- Data
      if (err_wr = '1')and(master_s2m_err.CmdAccept = '1')  then
        err_data <= '1';
      end if;

      if (err_data = '1')and(master_m2s_err.DataValid = '1') then
        master_s2m_err.DataAccept <= '1';
        if (err_bl = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
          err_data <= '0';
          err_wr <= '0';
        end if;
        err_bl <= err_bl - 1;
      else
        master_s2m_err.DataAccept <= '0';
      end if;

      -- Resp
      if (err_rd = '1')and(master_s2m_err.CmdAccept = '1')  then
        err_resp <= '1';
      end if;

      if (err_resp = '1')and(master_m2s_err.RespAccept = '1') then
        master_s2m_err.Resp <= OCP_RESP_VALID;
        if (err_bl = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
          err_resp <= '0';
          err_rd <= '0';
        end if;
        err_bl <= err_bl - 1;
      else
        master_s2m_err.Resp <= OCP_RESP_NONE;
      end if;
    end if;
  end process;


  retime_masters_g : if (retime_masters) generate
    -- Re-time OCP master ports
    retime_ports_g : for n in 0 to DEMUX_OUTPUTS-1 generate
      adb3_ocp_retime_nb_i : adb3_ocp_retime_nb
      port map(
        ocp_rst    => ocp_rst,            -- in    std_logic
        ocp_clk    => ocp_clk,            -- in    std_logic
        -- OCP slave port
        slave_m2s  => masters_m2s_out(n), -- in    adb3_ocp_m2sT
        slave_s2m  => masters_s2m_out(n), -- out   adb3_ocp_s2mT
        -- OCP master port
        master_m2s => masters_m2s(n),     -- out   adb3_ocp_m2sT
        master_s2m => masters_s2m(n));    -- in    adb3_ocp_s2mT)
    end generate;
  end generate;

  route_masters_g : if not(retime_masters) generate
    -- Route OCP master ports
    route_ports_g : for n in 0 to DEMUX_OUTPUTS-1 generate
      masters_m2s(n) <= masters_m2s_out(n);
      masters_s2m_out(n) <= masters_s2m(n);
    end generate;
  end generate;


  -- Error tracking
  -- Report if more that one bit of active port is active
  act_port_check_p : process(ocp_clk)
    variable act_port_m2s_var : natural;
  begin
    if rising_edge(ocp_clk) then
      act_port_m2s_var := 0;

      gen_act_port_check_l: for i in 0 to TOTAL_OUTPUTS-1 loop
        if (act_port_m2s(i) = '1') then
          act_port_m2s_var := act_port_m2s_var + 1;
        end if;
      end loop;

      if (act_port_m2s_var > 1) then
        act_port_m2s_err <= '1';
      else
        act_port_m2s_err <= '0';
      end if;
    end if;
  end process;

  -- Count number of cycles of error flags
  count_errors_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (act_port_m2s_err = '1') then
        act_port_m2s_cnt <= act_port_m2s_cnt + 1;
        assert false
          report "(adb3_ocp_split_b): Multiple OCP ports decoded as active in act_port_m2s"
          severity error;
      end if;
    end if;
  end process;

  debug_info <= X"000000" & act_port_m2s_cnt;
end;