--==============================================================================
--
--     File Name: mem_if_pkg.vhd
--       Version: 0.2
--          Date: 09 May 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package mem_if_pkg
--
--   Description: Package of constants used by memory interfaces
--
--   Limitations: Currently supports MIG V3.6 only
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 21 Sep 2010: Initial design
--                0.2, KDC  09 May 2011:
--                      - Modified MIG_V3_6_COMMON so that selection between
--                        1Gib and 2Gib parts is automatic
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;

package mem_if_pkg is

  -- DDR3 memory interface types
  type ddr3_addr_out_t is record
    a : std_logic_vector(DDR3_BANK_ROW_WIDTH-1 downto 0);
  end record;

  type ddr3_ctrl_out_t is record
    ras_l   : std_logic;
    cas_l   : std_logic;
    we_l    : std_logic;
    cs_l    : std_logic_vector(0 downto 0);
    dm      : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1  downto 0);
    ba      : std_logic_vector(2 downto 0);
    reset_l : std_logic;
    odt     : std_logic_vector(0 downto 0);
    cke     : std_logic_vector(0 downto 0);
  end record;

  type ddr3_data_inout_t is record
    dq    : std_logic_vector(DDR3_BANK_DATA_WIDTH-1 downto 0);
    dqs_p : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
    dqs_n : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
  end record;

  type ddr3_clk_out_t is record
    clk_p : std_logic_vector(0 downto 0);
    clk_n : std_logic_vector(0 downto 0);
  end record;

  type ddr3_addr_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_addr_out_t;
  type ddr3_ctrl_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_ctrl_out_t;
  type ddr3_data_inout_array_t is array (0 to DDR3_BANKS-1) of ddr3_data_inout_t;
  type ddr3_clk_out_array_t    is array (0 to DDR3_BANKS-1) of ddr3_clk_out_t;

  type mem_addr_out_t is record
    ddr3_addr_out : ddr3_addr_out_array_t;
  end record;

  type mem_ctrl_out_t is record
    ddr3_ctrl_out : ddr3_ctrl_out_array_t;
  end record;

  type mem_data_inout_t is record
    ddr3_data_inout : ddr3_data_inout_array_t;
  end record;

  type mem_clk_out_t is record
    ddr3_clk_out : ddr3_clk_out_array_t;
  end record;


  -- Functions used to resolve correct settings for simulation/synthesis
  function conv_sim_bypass_init_cal(
    constant val: in target_use_t)
  return string;

  function conv_sim_init_option(
    constant val: in target_use_t)
  return string;

  function conv_sim_cal_option(
    constant val: in target_use_t)
  return string;


  --------------------------------------------
  -- DDR3 MIG core version 3.6 declarations --
  --------------------------------------------

  -- Define clocking generics
  type mig_v3_6_clocks_t is record
    REFCLK_FREQ     : real;
    IODELAY_GRP     : string(1 to 11);
    CLKFBOUT_MULT_F : integer;
    DIVCLK_DIVIDE   : integer;
    CLKOUT_DIVIDE   : integer;
    nCK_PER_CLK     : integer;
    tCK             : integer;
    RST_ACT_LOW     : integer;
    INPUT_CLK_TYPE  : string(1 to 12);
    STARVE_LIMIT    : integer;
  end record;

  constant MIG_V3_6_CLOCKS : mig_v3_6_clocks_t :=
    (REFCLK_FREQ      => 200.0,
     IODELAY_GRP      => "IODELAY_MIG",
     CLKFBOUT_MULT_F  => 6,
     DIVCLK_DIVIDE    => 2,
     CLKOUT_DIVIDE    => 3,
     nCK_PER_CLK      => 2,
     tCK              => 2500,
     RST_ACT_LOW      => 0,  -- Changed from 1 (active low)
     INPUT_CLK_TYPE   => "SINGLE_ENDED",
     STARVE_LIMIT     => 2);


  -- Define common generics
  type mig_v3_6_common_t is record
    DEBUG_PORT      : string(1 to 3);
    nCS_PER_RANK    : integer;
    DQS_CNT_WIDTH   : integer;
    RANK_WIDTH      : integer;
    BANK_WIDTH      : integer;
    CK_WIDTH        : integer;
    CKE_WIDTH       : integer;
    COL_WIDTH       : integer;
    CS_WIDTH        : integer;
    DM_WIDTH        : integer;
    DQ_WIDTH        : integer;
    DQS_WIDTH       : integer;
    ROW_WIDTH       : integer;
    BURST_MODE      : string(1 to 3);
    BM_CNT_WIDTH    : integer;
    ADDR_CMD_MODE   : string(1 to 5);
    ORDERING        : string(1 to 4);
    WRLVL           : string(1 to 2);
    PHASE_DETECT    : string(1 to 2);
    RTT_NOM         : string(1 to 2);
    RTT_WR          : string(1 to 3);
    OUTPUT_DRV      : string(1 to 4);
    REG_CTRL        : string(1 to 3);
    tPRDI           : integer;
    tREFI           : integer;
    tZQI            : integer;
    tRFC            : integer;
    ADDR_WIDTH      : integer;
    ECC_TEST        : string(1 to 3);
    TCQ             : integer;
    DATA_WIDTH      : integer;
    PAYLOAD_WIDTH   : integer;
  end record;

  constant MIG_V3_6_COMMON : mig_v3_6_common_t :=
    (DEBUG_PORT       => "OFF",
     nCS_PER_RANK     => 1,
     DQS_CNT_WIDTH    => 2,
     RANK_WIDTH       => 1,
     BANK_WIDTH       => 3,
     CK_WIDTH         => 1,
     CKE_WIDTH        => 1,
     COL_WIDTH        => 10,
     CS_WIDTH         => 1,
     DM_WIDTH         => 4,
     DQ_WIDTH         => 32,
     DQS_WIDTH        => 4,
     ROW_WIDTH        => DDR3_BANK_ROW_WIDTH,  -- 13 for 1Gib devices, 14 for 2Gib devices
     BURST_MODE       => "OTF",
     BM_CNT_WIDTH     => 2,
     ADDR_CMD_MODE    => "UNBUF",
     ORDERING         => "NORM",
     WRLVL            => "ON",
     PHASE_DETECT     => "ON",
     RTT_NOM          => "60",
     RTT_WR           => "OFF",
     OUTPUT_DRV       => "HIGH",
     REG_CTRL         => "OFF",
     tPRDI            => 1000000,
     tREFI            => 7800000,
     tZQI             => 128000000,
     tRFC             => 160000,  -- >160ns for 2Gb devices (>110ns for 1Gb devices)
     ADDR_WIDTH       => DDR3_BYTE_ADDR_WIDTH-1, -- 27 for 1Gib devices, 28 for 2Gib devices
     ECC_TEST         => "OFF",
     TCQ              => 100,
     DATA_WIDTH       => 32,
     PAYLOAD_WIDTH    => 32);


  -- Define column generics
  type mig_v3_6_bank01_t is record
    nDQS_COL0       : integer;
    nDQS_COL1       : integer;
    nDQS_COL2       : integer;
    nDQS_COL3       : integer;
    DQS_LOC_COL0    : std_logic_vector(31 downto 0);
    DQS_LOC_COL1    : std_logic_vector(0 downto 0);
    DQS_LOC_COL2    : std_logic_vector(0 downto 0);
    DQS_LOC_COL3    : std_logic_vector(0 downto 0);
  end record;

  constant MIG_V3_6_BANK01 : mig_v3_6_bank01_t :=
    (nDQS_COL0        => 4,
     nDQS_COL1        => 0,
     nDQS_COL2        => 0,
     nDQS_COL3        => 0,
     DQS_LOC_COL0     => X"03020100",
     DQS_LOC_COL1     => "0",
     DQS_LOC_COL2     => "0",
     DQS_LOC_COL3     => "0");

  type mig_v3_6_bank2_t is record
    nDQS_COL0       : integer;
    nDQS_COL1       : integer;
    nDQS_COL2       : integer;
    nDQS_COL3       : integer;
    DQS_LOC_COL0    : std_logic_vector(0 downto 0);
    DQS_LOC_COL1    : std_logic_vector(0 downto 0);
    DQS_LOC_COL2    : std_logic_vector(31 downto 0);
    DQS_LOC_COL3    : std_logic_vector(0 downto 0);
  end record;

  constant MIG_V3_6_BANK2 : mig_v3_6_bank2_t :=
    (nDQS_COL0        => 0,
     nDQS_COL1        => 0,
     nDQS_COL2        => 4,
     nDQS_COL3        => 0,
     DQS_LOC_COL0     => "0",
     DQS_LOC_COL1     => "0",
     DQS_LOC_COL2     => X"03020100",
     DQS_LOC_COL3     => "0");

  type mig_v3_6_bank3_t is record
    nDQS_COL0       : integer;
    nDQS_COL1       : integer;
    nDQS_COL2       : integer;
    nDQS_COL3       : integer;
    DQS_LOC_COL0    : std_logic_vector(15 downto 0);
    DQS_LOC_COL1    : std_logic_vector(15 downto 0);
    DQS_LOC_COL2    : std_logic_vector(0 downto 0);
    DQS_LOC_COL3    : std_logic_vector(0 downto 0);
  end record;

  constant MIG_V3_6_BANK3 : mig_v3_6_bank3_t :=
    (nDQS_COL0        => 2,
     nDQS_COL1        => 2,
     nDQS_COL2        => 0,
     nDQS_COL3        => 0,
     DQS_LOC_COL0     => X"0100",
     DQS_LOC_COL1     => X"0302",
     DQS_LOC_COL2     => "0",
     DQS_LOC_COL3     => "0");

  component ddr3_if_bank_v3_6
  generic(
    bank : integer range 0 to DDR3_BANKS-1 := 0);
  port(
    -- OCP interface
    ocp_rst               : in    std_logic;
    ocp_clk               : in    std_logic;
    ocp_m2s               : in    adb3_ocp_m2sT;
    ocp_s2m               : out   adb3_ocp_s2mT;
    -- Memory interface
    ddr3_rst              : in    std_logic;
    ddr3_ref_clk          : in    std_logic;
    ddr3_clk              : in    std_logic;
    ddr3_iodelay_ctrl_rdy : in    std_logic;
    -- Memory status
    ddr3_if_rdy           : out   std_logic;
    ddr3_if_stat          : out   std_logic_vector(3 downto 0);
    ddr3_if_err           : out   std_logic_vector(3 downto 0);
    -- Physical memory interface
    ddr3_addr_out         : out   ddr3_addr_out_t;
    ddr3_ctrl_out         : out   ddr3_ctrl_out_t;
    ddr3_data_inout       : inout ddr3_data_inout_t;
    ddr3_clk_out          : out   ddr3_clk_out_t;
    -- Debug info
    ddr3_if_debug         : out   std_logic_vector(31 downto 0));
  end component;

end;


package body mem_if_pkg is

  function conv_sim_bypass_init_cal(
    constant val: in target_use_t)
  return string is
  begin
    case val is
      when SIM_OCP  => return "FAST";
      when SIM_MPTL => return "FAST";
      when SYN_NGC  => return "OFF";
    end case;
  end;

  function conv_sim_init_option(
    constant val: in target_use_t)
  return string is
  begin
    case val is
      when SIM_OCP  => return "SKIP_PU_DLY";
      when SIM_MPTL => return "SKIP_PU_DLY";
      when SYN_NGC  => return "NONE";
    end case;
  end;

  function conv_sim_cal_option(
    constant val: in target_use_t)
  return string is
  begin
    case val is
      when SIM_OCP  => return "FAST_CAL";
      when SIM_MPTL => return "FAST_CAL";
      when SYN_NGC  => return "NONE";
    end case;
  end;

end;