--==============================================================================
--
--     File Name: ddr3_sdram_init_fsm.vhd
--       Version: 0.4
--          Date: 28 Dec 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description: State Machine for initialisation sequence
--
--   Limitations: Precharge Power Down and Active Power Down not supported
--                OCD Calibration not supported
--                Reads interrupted by reads not supported
--                Writes interrupted by writes not supported
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 28 Dec 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.ddr3_sdram_pkg.all;
use work.ddr3_sdram_int_pkg.all;


entity ddr3_sdram_init_fsm is
  generic(
    part             : part_t);
  port(
    ck               : in    std_logic;
    reset_l          : in    std_logic;
    ck_stopped_error : in    boolean;
    cke              : in    std_logic;
    in_command       : in    command_t;
    in_addr          : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
    short_init_dly   : in    boolean;
    t_xpr_cyc        : in    natural;
    init_complete    : out   boolean;
    delay_500us_over : out   boolean;
    timing_msg_en    : out   boolean);
  end ddr3_sdram_init_fsm;


architecture behavioural of ddr3_sdram_init_fsm is

  -- State declaration
  type state_t is (
    PWR_DN,
    RES_OFF,
    LD1_MR2,
    LD1_MR3,
    LD1_MR1,
    LD1_MR0,
    ZQ_CL,
    ZQ_WAIT,
    INIT_END);

  -- State variables
  signal curr_st : state_t := PWR_DN;
  signal next_st : state_t;

  -- FSM outputs
  signal dll_reset : boolean;
  signal zq_init   : boolean;
  signal xpr_init  : boolean;

  -- Delay generation
  signal i_delay_200us_over  : boolean := false;
  signal i_delay_500us_over  : boolean := false;
  signal i_delay_10ns_over   : boolean := false;
  signal i_delay_dll_over    : boolean := false;
  signal i_delay_xpr_over    : boolean := false;
  signal i_delay_zqinit_over : boolean := false;
  signal i_reset_0_event     : time    := 0 ns;
  signal i_reset_1_event     : time    := 0 ns;
  signal i_cke_0_event       : time    := 0 ns;
  signal i_reset_valid       : boolean := false;

--  signal dll_tc : boolean;

  shared variable expect_idle : boolean;
  shared variable expect_cmd  : command_t;
  shared variable allow_aref  : boolean;

begin
  -- Detect valid reset pulse of 100ns
  gen_reset : process(reset_l,i_reset_0_event)
    variable reset_l_0_del : time;
  begin
    reset_l_0_del := now - i_reset_0_event;

    if (reset_l_0_del > DELAY100NS)and(rising_edge(reset_l)) then
      i_reset_valid <= true;
    else
      i_reset_valid <= false;
    end if;

    if (reset_l'event)and(reset_l = '0') then
      i_reset_0_event <= now;
    end if;
  end process;


  -- Generate 200 us initialisation delay
  gen_200us_delay : process(ck_stopped_error,short_init_dly,reset_l,i_reset_0_event,i_cke_0_event)
    variable reset_l_0_del : time;
    variable cke_0_del : time;
  begin
    reset_l_0_del := now - i_reset_0_event;
    cke_0_del     := now - i_cke_0_event;

    if (ck_stopped_error) then
      i_delay_200us_over <= false;
    elsif (((short_init_dly)and(reset_l_0_del > DELAY200SHRT))or
            (not(short_init_dly)and(reset_l_0_del > DELAY200FULL)))and
            (rising_edge(reset_l)) then
      i_delay_200us_over <= true;
    end if;

    if (ck_stopped_error) then
      i_delay_10ns_over <= false;
    elsif (cke_0_del > DELAY10NS)and(rising_edge(reset_l)) then
      i_delay_10ns_over <= true;
    end if;

    if (reset_l'event)and(reset_l = '1') then
      i_reset_1_event <= now;
    end if;
  end process;

  -- Generate 500 us initialisation delay
  gen_500us_delay : process(ck_stopped_error,short_init_dly,reset_l,cke,i_reset_1_event)
    variable reset_l_1_del : time;
  begin
    reset_l_1_del := now - i_reset_1_event;

    if (ck_stopped_error) then
      i_delay_500us_over <= false;
    elsif (((short_init_dly)and(reset_l_1_del > DELAY500SHRT)and(reset_l_1_del>DELAY500SHRT))or
       (not(short_init_dly)and(reset_l_1_del > DELAY500FULL)and(reset_l_1_del>DELAY500SHRT)))and
       (reset_l = '1')and(cke = '1') then
      i_delay_500us_over <= true;
    end if;

    if (cke'event)and(cke = '0') then
      i_cke_0_event <= now;
    end if;
  end process;


  -- Update memory state on rising edge of clock ck
  update_state : process(ck,ck_stopped_error)
  begin
    if ck_stopped_error then
      curr_st <= PWR_DN;
    elsif (ck'event and ck = '1') then
      curr_st <= next_st;

      if (curr_st /= PWR_DN)and(curr_st /= RES_OFF)and(curr_st /= INIT_END) then
        if expect_idle then
          assert not((in_command /= NOOP)and(in_command /= DDES))
            report "DDR3 SDRAM Init FSM (1) : Deviation from recommended initialisation sequence: " & command_t'image(in_command) & " when expecting NOOP or DDES"
            severity warning;
        else
          if (in_command /= NOOP)and(in_command /= DDES)and not((in_command = AREF)and(allow_aref)) then
            assert not(in_command /= expect_cmd)
              report "DDR3 SDRAM Init FSM (2) : Deviation from recommended initialisation sequence: " & command_t'image(in_command) & " when expecting " & command_t'image(expect_cmd)
              severity warning;
          end if;
        end if;
      end if;

      case curr_st is
        when PWR_DN =>
          if (reset_l = '1') then
            assert i_delay_200us_over
              report "DDR3 SDRAM Init FSM (3) : Deviation from recommended initialisation sequence: violation of 200us delay before RESET_L de-assertion"
              severity warning;
            assert i_delay_10ns_over
              report "DDR3 SDRAM Init FSM (4) : Deviation from recommended initialisation sequence: violation of 10ns CKE delay before RESET_L de-assertion"
              severity warning;
          end if;
        when RES_OFF =>
          if (cke = '1') then
            assert i_delay_500us_over
              report "DDR3 SDRAM Init FSM (5) : Deviation from recommended initialisation sequence: violation of 500us delay before CKE assertion"
              severity warning;
          end if;

        when LD1_MR2 =>
          if (in_command = MRS2) then
            assert i_delay_xpr_over
              report "DDR3 SDRAM Init FSM (6) : Deviation from recommended initialisation sequence: violation of t_xpr delay before MR2 set command"
              severity warning;
          end if;

        when LD1_MR1 =>
          if (in_command = MRS1)and(in_addr(0) /= '0') then
            assert FALSE
              report "DDR3 SDRAM Init FSM (7) : Deviation from recommended initialisation sequence: DLL not enabled in 1st MR1 set command"
              severity warning;
          end if;

        when LD1_MR0 =>
          if (in_command = MRS0)and(in_addr(8) /= '1') then
            assert FALSE
              report "DDR3 SDRAM Init FSM (8) : Deviation from recommended initialisation sequence: DLL reset not asserted in MR0 set command"
              severity warning;
          end if;

        when others =>
          null; -- Do nothing
      end case;

    end if;
  end process;

  delay_500us_over <= i_delay_500us_over;

  -- Generate next state from command
  gen_next_state : process(
    curr_st,
    reset_l,
    cke,
    i_delay_dll_over,
    i_delay_zqinit_over,
    in_command)
  begin
    -- Assign default values for outputs/next state in states
    -- where they are not explicitly defined or in illegal states
    next_st       <= curr_st;
    dll_reset     <= false;
    zq_init       <= false;
    xpr_init      <= false;
    init_complete <= false;
    timing_msg_en <= false;
    expect_idle   := true;
    allow_aref    := false;

    case curr_st is
      when PWR_DN   => if (reset_l = '1') then
                         next_st <= RES_OFF;
                       end if;

      when RES_OFF  => if (reset_l = '1')and(cke = '1') then
                         next_st <= LD1_MR2;
                         xpr_init <= true;
                       end if;

      when LD1_MR2  => expect_idle := false;
                       expect_cmd := MRS2;
                       if in_command = MRS2 then
                         next_st <= LD1_MR3;
                       end if;

      when LD1_MR3  => expect_idle := false;
                       expect_cmd := MRS3;
                       if in_command = MRS3 then
                         next_st <= LD1_MR1;
                       end if;

      when LD1_MR1  => expect_idle := false;
                       expect_cmd := MRS1;
                       if in_command = MRS1 then
                         next_st <= LD1_MR0;
                       end if;

      when LD1_MR0  => expect_idle := false;
                       expect_cmd := MRS0;
                       if in_command = MRS0 then
                         next_st <= ZQ_CL;
                         dll_reset <= true;
                       end if;

      when ZQ_CL    => expect_idle := false;
                       expect_cmd := ZQCL;
                       if in_command = ZQCL then
                         next_st <= ZQ_WAIT;
                         zq_init <= true;
                       end if;

      when ZQ_WAIT  => expect_idle := true;
                       if (i_delay_dll_over)and(i_delay_zqinit_over) then
                         next_st <= INIT_END;
                         timing_msg_en <= true;
                       end if;

      when INIT_END => init_complete <= true; -- Do nothing
    end case;
  end process;

  -- Initialisation t_xpr (t_xpr_cyc cycle) delay counter
  -- Load with minimum Delay time in cycles
  t_xpr_cnt : process(ck)
    variable xpr_cnt : natural;
  begin
    if (ck'event and ck = '1') then
      if xpr_init then
        xpr_cnt := t_xpr_cyc - 1;
        i_delay_xpr_over  <= false;
      else
        if (xpr_cnt > 0) then
          xpr_cnt := xpr_cnt - 1;
          i_delay_xpr_over <= false;
        else
          xpr_cnt := 0;
          i_delay_xpr_over <= true;
        end if;
      end if;
    end if;
  end process;

  -- Initialisation DLL reset (part.speed_grade.t_dllk_min cycle) delay counter
  -- Load with minimum Delay time in cycles
  dll_lock_cnt : process(ck)
    variable dll_cnt : natural;
  begin
    if (ck'event and ck = '1') then
      if dll_reset then
        dll_cnt := part.speed_grade.t_dllk_min - 1;
        i_delay_dll_over  <= false;
      else
        if (dll_cnt > 0) then
          dll_cnt := dll_cnt - 1;
          i_delay_dll_over <= false;
        else
          dll_cnt := 0;
          i_delay_dll_over <= true;
        end if;
      end if;
    end if;
  end process;

  -- Initialisation ZQCL delay (part.speed_grade.t_zqinit_min cycle) delay counter
  -- Load with minimum Delay time in cycles
  zq_init_cnt : process(ck)
    variable zq_cnt : natural;
  begin
    if (ck'event and ck = '1') then
      if zq_init then
        zq_cnt := part.speed_grade.t_zqinit_min - 1;
        i_delay_zqinit_over  <= false;
      else
        if (zq_cnt > 0) then
          zq_cnt := zq_cnt - 1;
          i_delay_zqinit_over <= false;
        else
          zq_cnt := 0;
          i_delay_zqinit_over <= true;
        end if;
      end if;
    end if;
  end process;

end behavioural;