--==============================================================================
--
--     File Name: ddr3_sdram_main_fsm.vhd
--       Version: 0.1
--          Date: 14 Feb 2011
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description: State Machine for main operation
--
--   Limitations: OCD Calibration not supported
--                Reads interrupted by reads not supported
--                Writes interrupted by writes not supported
--
--         Notes: When a timing violation is detected, for example a SPRE
--                command occurring before T_RAS(min) is satisfied, the model
--                behaves as if the command were legal, and its bank state
--                changes to BANK_IDLE accordingly. It does, however, flag this
--                as a timing violation.
--
--                Making a state transition even though the command that caused
--                the transition is clearly in violation of timing parameters
--                is intended to cut down the number of error messages emitted
--                by the model. In effect, the model attempts to match the state
--                that the DDR-II SDRAM controller expects it to have (even if
--                the controller is violating the model's timing parameters).
--
--                If the model did not behave this way, then when a timing
--                parameter is violated, the respective states of the model and
--                the controller would very quickly diverge, and the result
--                would be a very large number of potentially confusing error
--                messages emitted by the model.
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 28 Dec 2009: Initial design
--                0.2, KDC, 14 Feb 2011:
--                  - Corrected read/write auto precharge holdoff until t_ras
--                    complete.
--                0.3, KDC, 22 Apr 2011:
--                  - Use rtp_delay(i) instead of t_rd_pre_cyc in generation of
--                    rd_pre_delayed
--                  - Use wtp_delay(i) instead of t_wr_pre_cyc in generation of
--                    wr_pre_delayed
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;

library work;
use work.ddr3_sdram_int_pkg.all;


entity ddr3_sdram_main_fsm is
  generic(
    message_in : boolean;
    num_banks  : natural);
  port(
    ck               : in    std_logic;      -- Clock
    reset_l          : in    std_logic;      -- Reset (active low)
    cke              : in    std_logic;      -- Current CKE level on pins
    cke_d1           : in    std_logic;      -- As 'cke', but delayed by one cycle
    odt              : in    std_logic;      -- Current ODT pin level
    odt_d1           : in    std_logic;      -- As 'odt', but delayed by one cycle
    cs_l             : in    std_logic;      -- Current CS# pin level
    in_command       : in    command_t;      -- Decoded command currently on pins
    in_bank_num      : in    natural;        -- Bank index currently on pins
    delay_500us_over : in    boolean;        -- TRUE => initial 500 us delay is over
    pd_fast_exit     : in    boolean;        -- TRUE => fast exit power-down exit, from MREG[12]
    al_cyc           : in    natural;        -- Additive Latency from EMR1
    bl_cyc           : in    natural;        -- Burst Length from MR
    cl_cyc           : in    natural;        -- CAS latency from MR
    wl_cyc           : in    natural;        -- Write Length = RL - 1
    wr_cyc           : in    natural;        -- Write recovery from MR
    t_aofd           : in    time;           -- ODT turn-off delay
    t_mod            : in    time;           -- EMR1 to ODT turn-on delay
    t_wtr_cyc        : in    natural;        -- Write to read delay
    t_rfc_cyc        : in    natural;        -- Auto-refresh command period
    t_xs_cyc         : in    natural;        -- Exit self-refresh to ? command delay
    t_xsdll_cyc      : in    natural;        -- Exit self-refresh to ? command delay
    t_xpdll_cyc      : in    natural;        -- Exit power-down (slow) to command delay
    t_xp_cyc         : in    natural;        -- Exit power-down (fast) to command delay
    t_cke_cyc        : in    natural;        -- Clock enable low minimum clock cycles
    t_mrd_cyc        : in    natural;        -- MREG/EMR1/EMR2/EMR3 set command period
    t_ccd_cyc        : in    natural;        -- CAS-to-CAS delay between any two banks
    t_ras_cyc        : in    natural;        -- Minimum row open period
    t_rp_cyc         : in    natural;        -- Precharge single bank command period
    t_rpa_cyc        : in    natural;        -- Precharge all banks command period
    t_rc_cyc         : in    natural;        -- Activate-to-activate delay, same bank
    t_rrd_cyc        : in    natural;        -- Activate-to-activate delay between any two banks
    t_rcd_cyc        : in    natural;        -- RAS-to-CAS delay for a particular bank
    t_rtp_cyc        : in    natural;        -- Read to precharge delay for a particular bank
    t_wr_cyc         : in    natural;        -- Write recovery time, in cycles
    t_zqoper_cyc     : in    natural;        -- ZQ Calibration Long delay
    t_zqcs_cyc       : in    natural;        -- ZQ Calibration Short delay
--    t_anpd_cyc       : in    natural;        -- ODP high/low to power down entry period
--    t_axpd_cyc       : in    natural;        -- Power down exit to ODP high/low period
    t_ck_min         : in    time;           -- Clock period (minimum)
    dll_delay_tc     : in    boolean;        -- DLL reset to first read delay counter terminal count
    ck_stopped_error : out   boolean;        -- TRUE => Clock has gone out of spec (not in self refresh)
    load_mr0         : out   boolean;        -- TRUE => current command is MR0 set
    load_mr1         : out   boolean;        -- TRUE => current command is MR1 set
    load_mr2         : out   boolean;        -- TRUE => current command is MR2 set
    load_mr3         : out   boolean;        -- TRUE => current command is MR3 set
    auto_refresh     : out   boolean;        -- TRUE => current command is auto-refresh
    self_refresh     : out   boolean;        -- TRUE => currently in self-refresh mode
    bank_act_any     : out   boolean;        -- TRUE => bank activate (any bank)
    bank_read_any    : out   boolean;        -- TRUE => bank read (any bank)
    bank_write_any   : out   boolean;        -- TRUE => bank write (any bank)
    bank_pre_any     : out   boolean;        -- TRUE => bank pre-charge (any bank)
    bank_pre_all     : out   boolean;        -- TRUE => bank pre-charge all
    -- These outputs indicate when certain DDR3 SDRAM inputs must be valid
    cke_setup        : out   boolean;        -- TRUE => CKE must be valid
    a_setup          : out   boolean;        -- TRUE => A except A[10] must be valid
    a10_setup        : out   boolean;        -- TRUE => A[10] must be valid
    ba_setup         : out   boolean;        -- TRUE => BA must be valid
    ras_l_setup      : out   boolean;        -- TRUE => RAS# must be valid
    cas_l_setup      : out   boolean;        -- TRUE => CAS# must be valid
    cs_l_setup       : out   boolean;        -- TRUE => CS# must be valid
    we_l_setup       : out   boolean;        -- TRUE => WE# must be valid
    odt_setup        : out   boolean);       -- TRUE => ODT must be valid
  end ddr3_sdram_main_fsm;


architecture behavioural of ddr3_sdram_main_fsm is

  -- Device state declaration
  type state_t is (
    PWRDOWN,
    IDLE,
    SREFRESH);

  -- Device state variables
  signal curr_st : state_t := PWRDOWN;
  signal next_st : state_t;

  -- Bank state declaration
  type bank_state_t is (
    BANK_IDLE,
    BANK_OPEN);
  type bank_state_vector_t is array(0 to num_banks-1) of bank_state_t;
  signal all_banks_idle : boolean;

  -- Bank state variables
  signal curr_bank_st : bank_state_vector_t := (others => BANK_OPEN); -- Expecting initial precharge all commands
  signal next_bank_st : bank_state_vector_t;

  -- Refresh command period counter (device)
  signal rfc_delay     : natural;
  signal rfc_delay_go  : boolean;
  signal rfc_delay_tc  : boolean := true;

  -- MREG set command period counter (device)
  signal mrs_delay     : natural;
  signal mrs_delay_go  : boolean;
  signal mrs_delay_tc  : boolean := true;

  -- Activate to activate (for different banks) delay counter (device)
  signal rrd_delay     : natural;
  signal rrd_delay_go  : boolean;
  signal rrd_delay_tc  : boolean := true;

  -- CAS to CAS delay counter (device)
  signal ccd_delay     : natural;
  signal ccd_delay_go  : boolean;
  signal ccd_delay_tc  : boolean := true;

  -- Read to readdelay counter (device)
  signal rtr_delay     : natural;
  signal rtr_delay_go  : boolean;
  signal rtr_delay_tc  : boolean := true;

  -- Read to Write delay counter (device)
  signal rtw_delay     : natural;
  signal rtw_delay_go  : boolean;
  signal rtw_delay_tc  : boolean := true;

  -- Write to write delay counter (device)
  signal wtw_delay     : natural;
  signal wtw_delay_go  : boolean;
  signal wtw_delay_tc  : boolean := true;

  -- Write to Read delay counter (device)
  signal wtr_delay     : natural;
  signal wtr_delay_go  : boolean;
  signal wtr_delay_tc  : boolean := true;

  -- Self-Refresh Exit to ? delay counter (device)
  signal xs_delay     : natural;
  signal xs_delay_go  : boolean;
  signal xs_delay_tc  : boolean := true;

  -- Self-Refresh Exit to ? delay counter (device)
  signal xsdll_delay     : natural;
  signal xsdll_delay_go  : boolean;
  signal xsdll_delay_tc  : boolean := true;

  -- Read to Power-Down Entry delay counter (device)
  signal rtpden_delay    : natural;
  signal rtpden_delay_go : boolean;
  signal rtpden_delay_tc : boolean := true;

  -- Write to Power-Down Entry delay counter (device)
  signal wrte_pden_delay    : natural;
  signal wrte_pden_delay_go : boolean;
  signal wrte_pden_delay_tc : boolean := true;

  -- Write with auto-precharge to Power-Down Entry delay counter (device)
  signal wrap_pden_delay    : natural;
  signal wrap_pden_delay_go : boolean;
  signal wrap_pden_delay_tc : boolean := true;

  -- CKE legal transition delay counter (device)
  signal cke_stable_delay    : natural;
  signal cke_stable_delay_go : boolean;
  signal cke_stable_delay_tc : boolean := true;

  -- Power-down exit to read delay counter (device)
  signal xard_delay    : natural;
  signal xard_delay_go : boolean;
  signal xard_delay_tc : boolean := true;

  -- Power-down exit to non-read delay counter (device)
  signal xp_delay     : natural;
  signal xp_delay_go  : boolean;
  signal xp_delay_tc  : boolean := true;

  -- odt to Power-down exit to delay counter (device)
  signal anpd_delay     : natural;
  signal anpd_delay_go  : boolean;
  signal anpd_delay_tc  : boolean := true;

  -- Power-down exit to odt delay counter (device)
  signal axpd_delay     : natural;
  signal axpd_delay_go  : boolean;
  signal axpd_delay_tc  : boolean := true;

  -- ZQ Calibration Short delay counter (device)
  signal zqcs_delay     : natural;
  signal zqcs_delay_go  : boolean;
  signal zqcs_delay_tc  : boolean := true;

  -- ZQ Calibration Long delay counter (device)
  signal zqcl_delay     : natural;
  signal zqcl_delay_go  : boolean;
  signal zqcl_delay_tc  : boolean := true;

  -- Precharge command period (t_rp) counters (one per bank)
  signal rp_delay      : natural_array_t(0 to num_banks-1);
  signal rp_delay_go   : boolean_array_t(0 to num_banks-1);
  signal rp_delay_tc   : boolean_array_t(0 to num_banks-1);

  -- RAS-to-CAS command period (t_rcd) counters (one per bank)
  signal rcd_delay     : natural_array_t(0 to num_banks-1);
  signal rcd_delay_go  : boolean_array_t(0 to num_banks-1);
  signal rcd_delay_tc  : boolean_array_t(0 to num_banks-1);

  -- RAS minimum period (t_ras) counters (one per bank)
  signal rasmin_delay     : natural_array_t(0 to num_banks-1);
  signal rasmin_delay_go  : boolean_array_t(0 to num_banks-1);
  signal rasmin_delay_tc  : boolean_array_t(0 to num_banks-1);

  -- Read to precharge delay (t_rtp) counters (one per bank)
  signal rtp_delay     : natural_array_t(0 to num_banks-1);
  signal rtp_delay_go  : boolean_array_t(0 to num_banks-1);
  signal rtp_delay_tc  : boolean_array_t(0 to num_banks-1);

  -- Write to precharge delay (t_wtp) counters (one per bank)
  signal wtp_delay     : natural_array_t(0 to num_banks-1);
  signal wtp_delay_go  : boolean_array_t(0 to num_banks-1);
  signal wtp_delay_tc  : boolean_array_t(0 to num_banks-1);

  -- Activate to activate (for same bank) delay counter (device)
  signal rc_delay     : natural_array_t(0 to num_banks-1);
  signal rc_delay_go  : boolean_array_t(0 to num_banks-1);
  signal rc_delay_tc  : boolean_array_t(0 to num_banks-1);


  --
  -- t_rd_pre delay pipes (one per bank).
  --
  -- Maximum pipe length is AL_MAX + BL_MAX + max(t_rtp_cyc) - 2.
  -- Assume max(t_rtp_cyc) = 8 (which would be ridiculous in a real device).
  --
  constant T_RD_PRE_CYC_MAX : natural := (AL_MAX + BL_MAX + 8 - 2);
  type rd_pre_pipe_vector_t is array (0 to num_banks-1) of boolean_array_t(1 to T_RD_PRE_CYC_MAX);
  signal rd_pre_pipe        : rd_pre_pipe_vector_t;
  signal rdap_command       : boolean_array_t(0 to num_banks-1);
  signal rd_pre_delayed     : boolean_array_t(0 to num_banks-1);
  signal rdap_delayed       : boolean_array_t(0 to num_banks-1);
  signal ext_rd_pre         : boolean_array_t(0 to num_banks-1);
--  signal in_bank_num_rdap_pipe    : natural_array_t(1 to T_RTP_CYC_MAX);
--  signal in_bank_num_rdap_delayed : natural range 0 to num_banks;

  --
  -- t_wr_pre delay pipes (one per bank).
  --
  -- Maximum pipe length is WL_MAX + BL_MAX + max(t_wr_cyc).
  -- Assume max(t_wr_cyc) = 8 (which would be ridiculous in a real device).
  --
  constant T_WR_PRE_CYC_MAX : natural := (WL_MAX + BL_MAX + 8);
  type wr_pre_pipe_vector_t is array (0 to num_banks-1) of boolean_array_t(1 to T_WR_PRE_CYC_MAX);
  signal wr_pre_pipe        : wr_pre_pipe_vector_t;
  signal wrap_command       : boolean_array_t(0 to num_banks-1);
  signal wr_pre_delayed     : boolean_array_t(0 to num_banks-1);
  signal wrap_delayed       : boolean_array_t(0 to num_banks-1);
  signal ext_wr_pre         : boolean_array_t(0 to num_banks-1);
--  signal in_bank_num_wrap_pipe    : natural_array_t(1 to T_WR_CYC_MAX);
--  signal in_bank_num_wrap_delayed : natural range 0 to num_banks;

  --
  -- t_ras delay pipes (one per bank).
  --
  -- Maximum pipe length is max(t_ras_cyc).
  -- Assume max(t_ras_cyc) = 20 (which would be ridiculous in a real device).
  --
  constant T_RAS_CYC_MAX : natural := 20;
  type ras_pipe_vector_t is array (0 to num_banks-1) of boolean_array_t(1 to T_RAS_CYC_MAX);
  signal ras_pipe        : ras_pipe_vector_t;
  signal bact_command    : boolean_array_t(0 to num_banks-1);
  signal ras_delayed     : boolean_array_t(0 to num_banks-1);
--  signal in_bank_num_ras_pipe    : natural_array_t(1 to T_RAS_CYC_MAX);
--  signal in_bank_num_ras_delayed : natural range 0 to num_banks;

  -- Dynamically calculated delay values, measured in cycles
  signal t_rd_rd_cyc     : natural;
  signal t_rd_wr_cyc     : natural;
  signal t_wr_wr_cyc     : natural;
  signal t_wr_rd_cyc     : natural;
  signal t_rcd_al_cyc    : natural;
  signal t_rd_pre_cyc    : natural;
  signal t_wr_pre_cyc    : natural;
  signal t_rd_pden_cyc   : natural;
  signal t_wrte_pden_cyc : natural;
  signal t_wrap_pden_cyc : natural;
  signal t_pdex_rd_cyc   : natural;

  signal i_cke_setup   : boolean;
  signal i_cs_l_setup  : boolean;

  signal power_down    : boolean;
  signal ck_stopped    : boolean := true;
  signal mr1_odt_hold  : boolean;

  signal bl_cyc_rdwr : natural;

begin

  --
  -- Generate setup / hold / validity indicators for device inputs
  --

  i_cke_setup  <= not ck_stopped;
  odt_setup    <= i_cke_setup;
  i_cs_l_setup <= true when (i_cke_setup and (curr_st = IDLE)) else
                  true when (i_cke_setup and (curr_st = PWRDOWN) and (cke = '1')) else
                  true when (i_cke_setup and (curr_st = SREFRESH) and (cke = '1')) else
                  false;
  ras_l_setup  <= i_cs_l_setup and cs_l = '0';
  cas_l_setup  <= i_cs_l_setup and cs_l = '0';
  we_l_setup   <= i_cs_l_setup and cs_l = '0';

  cke_setup    <= i_cke_setup;
  cs_l_setup   <= i_cs_l_setup;

  -- Clock stopped detection
  -- Detect clock stopped or too slow after 10 clock periods (minimum)
  gen_ck_stopped : process(ck)
    variable stopped_time : time;
  begin
    if (ck'event and ck = '1') then
      if (t_ck_min > 0 ns) then
        stopped_time := 10 * t_ck_min;
      else
        stopped_time := 20 ns;
      end if;
      ck_stopped <= inertial false, true after stopped_time;
    end if;
  end process;

  --
  -- Calculate dynamic delays
  --

  -- Calculate cycles between read and read commands to any bank
  t_rd_rd_cyc <= (bl_cyc_rdwr);

  -- Calculate cycles between read and write commands to any bank
  t_rd_wr_cyc <= (bl_cyc_rdwr + 2);

  -- Calculate cycles between write and write commands to any bank
  t_wr_wr_cyc <= (bl_cyc_rdwr);

  -- Calculate cycles between write and read commands to any bank
  -- Use maximum of t_wtr_cyc or 2 cycles
  t_wr_rd_cyc <= (cl_cyc - 1 + bl_cyc_rdwr + 2) when (t_wtr_cyc < 2) else (cl_cyc - 1 + bl_cyc_rdwr + t_wtr_cyc);

  -- Calculate t_rcd_cyc adjusted for additive latency al_cyc
  t_rcd_al_cyc <= (t_rcd_cyc - al_cyc) when (al_cyc < t_rcd_cyc) else 0;

  -- Calculate Read to Precharge time = (AL + BL/2 - 2 + t_rtp_cyc) when t_rtp_cyc > 2
  -- Calculate Read to Precharge time = (AL + BL/2 - 2 + 2)         otherwise
  t_rd_pre_cyc <= (al_cyc + bl_cyc_rdwr + (t_rtp_cyc - 2)) when (t_rtp_cyc > 2) else (al_cyc + bl_cyc_rdwr);
  -- Minimum value is AL_MIN + BL_MIN/2 + 2 = 0 + 4/2 + 2 = 4

  -- Calculate Write to Precharge time = (WL + BL/2 + t_wr_cyc)
  t_wr_pre_cyc <= (wl_cyc + bl_cyc_rdwr + t_wr_cyc);
  -- Minimum value is WL_MIN + BL_MIN/2 + 1 = 2 + 4/2 + 1 = 5

  -- Calculate READ to PDEN entry delay
  t_rd_pden_cyc <= al_cyc + cl_cyc + bl_cyc_rdwr + 1;

  -- Calculate WRTE to PDEN entry delay
  t_wrte_pden_cyc <= wl_cyc + bl_cyc_rdwr + t_wtr_cyc;

  -- Calculate WRAP to PDEN entry delay
  t_wrap_pden_cyc <= wl_cyc + bl_cyc_rdwr + t_wr_cyc;

  -- Calculate SPEX to read command delay
  t_pdex_rd_cyc <= t_xp_cyc when pd_fast_exit else t_xpdll_cyc;

  -- Update memory and bank state on rising edge of ck
  -- If clock stops while not in self refresh, then enter power down - initialisation required
  update_state : process(ck,ck_stopped,curr_st)
  begin
    if (ck_stopped)and(curr_st /= SREFRESH) then
      curr_st          <= PWRDOWN;
      curr_bank_st     <= (others => BANK_IDLE);
      ck_stopped_error <= true;
    elsif (ck'event and ck = '1') then
      curr_st          <= next_st;
      curr_bank_st     <= next_bank_st;
      ck_stopped_error <= false;
    end if;
  end process;

  gen_all_banks_idle : process(curr_bank_st)
    variable tmp : boolean;
  begin
    tmp := true;
    for i in 0 to num_banks-1 loop
      tmp := tmp and (curr_bank_st(i) = BANK_IDLE);
    end loop;
    all_banks_idle <= tmp;
  end process;

  -- Error checking
  error_check : process(ck,ck_stopped)
  begin
    if ck_stopped then
      if (ck_stopped'event) then
        assert not((curr_st = IDLE)or(curr_st = PWRDOWN))
          report "DDR3 SDRAM Main FSM (01): CK appears to have stopped or is too slow while not in Self Refresh - Memory contents lost - Initialisation required"
          severity error;

        assert not(curr_st = SREFRESH)
          report "DDR3 SDRAM Main FSM (02): CK appears to have stopped or is too slow in Self Refresh"
          severity warning;
      end if;
    elsif (ck'event and ck = '1') then
      case curr_st is
        when PWRDOWN =>
          for i in 0 to num_banks-1 loop
            if rdap_delayed(i) or wrap_delayed(i) then
              assert false
                report "DDR3 SDRAM Main FSM (03): Auto-precharge occurred while powered-down; implies last PDEN command was illegal"
                severity error;
            end if;
          end loop;

          if (cke = '1') then
            if (in_command /= SPEX) then
--              if not delay_500us_over then
--                -- Delay from power-on / reset should be at least 500 us
--                assert false
--                  report "DDR3 SDRAM Main FSM (04): Attempting to exit power-down mode before 500 us has elapsed from power-on / reset"
--                  severity error;
--              end if;
--            else
              -- Only DDES/NOOP are valid patterns at power-down exit
              assert false
                report "DDR3 SDRAM Main FSM (04): Illegal command exiting power-down mode: " & command_t'image(in_command)
                severity error;
            end if;
          end if;

        when SREFRESH =>
          for i in 0 to num_banks-1 loop
            if rdap_delayed(i) or wrap_delayed(i) then
              assert false
                report "DDR3 SDRAM Main FSM (05): Auto-precharge occurred while in self-refresh mode; implies last SREN command was illegal"
                severity error;
            end if;
          end loop;

          if cke = '1' then
            if in_command = SPEX then
              if (odt /= '0') then -- ODT must not be enabled at self-refresh exit
                assert false
                  report "DDR3 SDRAM Main FSM (06): ODT not de-asserted at self-refresh exit"
                  severity error;
              end if;
            else
              -- Only DDES/NOOP are valid patterns at self-refresh exit
              assert false
                report "DDR3 SDRAM Main FSM (07): Illegal command exiting self-refresh mode: " & command_t'image(in_command)
                severity error;
            end if;
          end if;

        when IDLE =>
          if not mrs_delay_tc and (in_command /= NOOP and in_command /= DDES) then
            -- The only legal commands during MREG/EMR1/EMR2/EMR3 command period are NOOP & DDES
            assert false
              report "DDR3 SDRAM Main FSM (08): T_MRS period violated by " & command_t'image(in_command)
              severity error;
          end if;

          if not rfc_delay_tc and (in_command /= NOOP and in_command /= DDES) then
            -- The only legal commands during auto-refresh command period are NOOP & DDES
            assert false
              report "DDR3 SDRAM Main FSM (09): T_RFC period violated by " & command_t'image(in_command)
              severity error;
          end if;

--          if not(xsnr_delay_tc) and (in_command /= NOOP and in_command /= DDES) then
--            case in_command is
--              when NOOP | DDES =>
--                null; -- OK
--
--              when RDAP | READ =>
--                null; -- t_xsnr doesn't apply to reads; t_xsrd does;
--
--              when others =>
--                assert false
--                  report "DDR3 SDRAM Main FSM (11): T_XSNR after self-refresh exit violated by " & command_t'image(in_command)
--                  severity error;
--            end case;
--          end if;

          if not xp_delay_tc and (in_command /= NOOP and in_command /= DDES) then
            case in_command is
              when NOOP | DDES =>
                null; -- OK

              when RDAP | READ =>
                null; -- T_XP doesn't apply to reads; T_XARD does.

              when others =>
                assert false
                  report "DDR3 SDRAM Main FSM (10): T_XP after power-down exit violated by " & command_t'image(in_command)
                  severity error;
            end case;
          end if;

          if not zqcs_delay_tc and (in_command /= NOOP and in_command /= DDES) then
            -- The only legal commands during ZQCS command period are NOOP & DDES
            assert false
              report "DDR3 SDRAM Main FSM (11): T_ZQCS period violated by " & command_t'image(in_command)
              severity error;
          end if;

          if not zqcl_delay_tc and (in_command /= NOOP and in_command /= DDES) then
            -- The only legal commands during ZQCL command period are NOOP & DDES
            assert false
              report "DDR3 SDRAM Main FSM (12): T_ZQCL period violated by " & command_t'image(in_command)
              severity error;
          end if;

          for i in 0 to num_banks-1 loop
            if wrap_delayed(i) then
              if curr_bank_st(i) = BANK_OPEN and not rasmin_delay_tc(i) then
                assert false
                  report "DDR3 SDRAM Main FSM (13): T_RAS(min) delay violated following write with auto-precharge"
                  severity error;
              end if;

              if curr_bank_st(i) = BANK_OPEN and not wtp_delay_tc(i) then
                assert false
                  report "DDR3 SDRAM Main FSM (14): T_WTP delay violated following write with auto-precharge"
                  severity error;
              end if;
            end if;

            if rdap_delayed(i) then
              if curr_bank_st(i) = BANK_OPEN and not rasmin_delay_tc(i) then
                assert false
                  report "DDR3 SDRAM Main FSM (15): T_RAS(min) delay violated following read with auto-precharge"
                  severity error;
              end if;

              if curr_bank_st(i) = BANK_OPEN and not rtp_delay_tc(i) then
                assert false
                  report "DDR3 SDRAM Main FSM (16): T_RTP delay violated following read with auto-precharge"
                  severity error;
              end if;
            end if;
          end loop;

          case in_command is
            when AREF =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (17): Performing Auto-Refresh operation"
                severity note;

              if not(all_banks_idle) then
                -- Attempt to auto-refresh when not all banks are idle
                assert false
                  report "DDR3 SDRAM Main FSM (18): Auto-refresh command when not all banks idle"
                  severity error;
              else
                for i in 0 to num_banks-1 loop
                  if curr_bank_st(i) = BANK_IDLE then
                    if not(rp_delay_tc(i)) then
                      -- t_rp violation on AREF command
                      assert false
                        report "DDR3 SDRAM Main FSM (19): T_RP for bank " & natural'image(i) & " violated by auto-refresh command"
                        severity error;
                    end if;
                  end if;
                end loop;
              end if;

            when SREN =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (20): Performing Self-Refresh Entry operation"
                severity note;

              if (odt'last_event < t_aofd) then
                -- t_aofd violation entering self-refresh
                assert false
                  report "DDR3 SDRAM Main FSM (21): T_AOFD violated by self-refresh entry command"
                  severity error;
              end if;

              if (odt /= '0') then
                -- odt high entering self-refresh
                assert false
                  report "DDR3 SDRAM Main FSM (22): ODT not low during self-refresh entry command"
                  severity error;
              end if;

              if not(all_banks_idle) then
                -- Attempt to enter self-refresh when not all banks are idle
                assert false
                  report "DDR3 SDRAM Main FSM (23): Self-refresh entry command when not all banks idle"
                  severity error;
              else
                for i in 0 to num_banks-1 loop
                  if (curr_bank_st(i) = BANK_IDLE) then
                    if not(rp_delay_tc(i)) then
                      -- t_rp violation on AREF command
                      assert false
                        report "DDR3 SDRAM Main FSM (24): T_RP for bank " & natural'image(i) & " violated by self-refresh entry command"
                        severity error;
                    end if;
                  end if;
                end loop;
              end if;

            when BACT =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (25): Performing Bank Activate operation"
                severity note;

              if curr_bank_st(in_bank_num) = BANK_OPEN then
                -- Attempt to activate a bank when already open
                assert false
                  report "DDR3 SDRAM Main FSM (26): Attempt to activate bank " & natural'image(in_bank_num) & " while open"
                  severity error;
              end if;
              if not rp_delay_tc(in_bank_num) then
                -- Attempt to activate a bank while precharging
                assert false
                  report "DDR3 SDRAM Main FSM (27): T_RP for bank " & natural'image(in_bank_num) & " violated by bank activate command"
                  severity error;
              end if;
              if not rrd_delay_tc then
                -- Attempt to activate another bank too soon, violates t_rrd
                assert false
                  report "DDR3 SDRAM Main FSM (28): T_RRD violated by bank activate command"
                  severity error;
              end if;
              if not rc_delay_tc(in_bank_num) then
                -- Attempt to activate same bank too soon, violates t_rc
                assert false
                  report "DDR3 SDRAM Main FSM (29): T_RC for bank " & natural'image(in_bank_num) & " violated by bank activate command"
                  severity error;
              end if;
              if wrap_delayed(in_bank_num) or rdap_delayed(in_bank_num) then
                -- Current command conflicts with auto-precharge from a previous RDAP/WRAP command
                assert false
                  report "DDR3 SDRAM Main FSM (30): Bank activate conflicts with auto-precharge from a previous read/write command. The auto-precharge will be ignored."
                  severity error;
              end if;

            when READ | RDAP =>
              if not ccd_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (31): T_CCD violated by read/read with auto-precharge command"
                  severity error;
              end if;
              if not rtr_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (32): T_RTR violated by read/read with auto-precharge command"
                  severity error;
              end if;
              if not wtr_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (33): T_WTR violated by read/read with auto-precharge command"
                  severity error;
              end if;
--              if not xsrd_delay_tc then
--                assert false
--                  report "DDR3 SDRAM Main FSM (31): T_XSRD violated by read/read with auto-precharge command"
--                  severity error;
--              end if;
              if not xard_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (34): T_XARD violated by read/read with auto-precharge command"
                  severity error;
              end if;
              if not dll_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (35): DLL reset to first read delay violated by read/read with auto-precharge command"
                  severity error;
              end if;
              if wrap_delayed(in_bank_num) or rdap_delayed(in_bank_num) then
                -- Current read command conflicts with auto-precharge from a previous RDAP/WRAP command
                assert false
                  report "DDR3 SDRAM Main FSM (36): Read/read with auto-precharge command conflicts with auto-precharge from a previous command. The read will be ignored."
                  severity error;
              end if;

            when WRTE | WRAP =>
              if not ccd_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (37): T_CCD violated by write/write with auto-precharge command"
                  severity error;
              end if;
              if not rtw_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (38): T_RTW violated by write/write with auto-precharge command"
                  severity error;
              end if;
              if not wtw_delay_tc then
                assert false
                  report "DDR3 SDRAM Main FSM (39): T_WTW violated by write/write with auto-precharge command"
                  severity error;
              end if;
              if wrap_delayed(in_bank_num) or rdap_delayed(in_bank_num) then
                -- Current write command conflicts with auto-precharge from a previous RDAP/WRAP command
                assert false
                  report "DDR3 SDRAM Main FSM (40): Write/write with auto-precharge command conflicts with auto-precharge from a previous command. The write will be ignored."
                  severity error;
              end if;

            when PDEN =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (41): Performing Power-Down Entry operation"
                severity note;

              if not rtpden_delay_tc then
                -- PDEN command violates read to power down delay
                assert false
                  report "DDR3 SDRAM Main FSM (42): Power-down entry command violates read to power down entry delay"
                  severity error;
              end if;
              if not wrte_pden_delay_tc then
                -- PDEN command violates write to power down delay
                assert false
                  report "DDR3 SDRAM Main FSM (43): Power-down entry command violates write to power down entry delay"
                  severity error;
              end if;
              if not wrap_pden_delay_tc then
                -- PDEN command violates write with auto-precharge to power down delay
                assert false
                  report "DDR3 SDRAM Main FSM (44): Power-down entry command violates write with auto-precharge to power down entry delay"
                  severity error;
              end if;
              if not mrs_delay_tc then
                -- PDEN command violates MR0/MR1/MR2/MR3 to power down delay
                assert false
                  report "DDR3 SDRAM Main FSM (45): Power-down entry command violates T_MRS (set register to power down entry) delay"
                  severity error;
              end if;
              if not anpd_delay_tc then
                -- PDEN command violates odt high/low to power down delay
                assert false
                  report "DDR3 SDRAM Main FSM (46): Power-down entry command violates T_ANPD (odt change to power down entry) delay"
                  severity error;
              end if;

            when APRE =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (47): Performing All Bank Pre-charge operation"
                severity note;

              for i in 0 to num_banks-1 loop
                if curr_bank_st(i) = BANK_OPEN and not rasmin_delay_tc(i) then
                  -- Precharge all violates t_ras for bank
                  assert false
                    report "DDR3 SDRAM Main FSM (48): T_RAS(min) for bank " & natural'image(i) & " violated by precharge all command"
                    severity error;
                end if;
                if curr_bank_st(i) = BANK_OPEN and not rtp_delay_tc(i) then
                  -- Precharge all violates t_rtp for bank
                  assert false
                    report "DDR3 SDRAM Main FSM (49): T_RTP for bank " & natural'image(i) & " violated by precharge all command"
                    severity error;
                end if;
                if curr_bank_st(i) = BANK_OPEN and not wtp_delay_tc(i) then
                  -- Precharge all violates t_wtp for bank
                  assert false
                    report "DDR3 SDRAM Main FSM (50): T_WTP for bank " & natural'image(i) & " violated by precharge all command"
                    severity error;
                end if;
              end loop;

            when SPRE =>
              assert not(message_in)
                report "DDR3 SDRAM Main FSM (51): Performing Single Bank Pre-charge operation"
                severity note;

              if curr_bank_st(in_bank_num) = BANK_OPEN and not rasmin_delay_tc(in_bank_num) then
                -- Precharge bank violates t_ras for bank
                assert false
                  report "DDR3 SDRAM Main FSM (52): T_RAS(min) for bank " & natural'image(in_bank_num) & " violated by precharge bank command"
                  severity error;
              end if;
              if curr_bank_st(in_bank_num) = BANK_OPEN and not rtp_delay_tc(in_bank_num) then
                -- Precharge bank violates t_rtp for bank
                assert false
                  report "DDR3 SDRAM Main FSM (53): T_RTP for bank " & natural'image(in_bank_num) & " violated by precharge bank command"
                  severity error;
              end if;
              if curr_bank_st(in_bank_num) = BANK_OPEN and not wtp_delay_tc(in_bank_num) then
                -- Precharge bank violates t_wtp for bank
                assert false
                  report "DDR3 SDRAM Main FSM (54): T_WTP for bank " & natural'image(in_bank_num) & " violated by precharge bank command"
                  severity error;
              end if;

            when NOOP | DDES | MRS0 | MRS2 | MRS3 =>
              null; -- Do nothing

            when ZQCL =>
              if not(all_banks_idle) then
                -- Attempt to enter self-refresh when not all banks are idle
                assert false
                  report "DDR3 SDRAM Main FSM (55): ZQ Calibration Long command when not all banks idle"
                  severity error;
              else
                for i in 0 to num_banks-1 loop
                  if (curr_bank_st(i) = BANK_IDLE) then
                    if not(rp_delay_tc(i)) then
                      -- t_rp violation on ZQCL command
                      assert false
                        report "DDR3 SDRAM Main FSM (56): T_RP for bank " & natural'image(i) & " violated by ZQ Calibration Long command"
                        severity error;
                    end if;
                  end if;
                end loop;

                assert not(message_in)
                  report "DDR3 SDRAM Main FSM (57): Performing ZQ Calibration Long operation"
                  severity note;
              end if;

            when ZQCS =>
              if not(all_banks_idle) then
                -- Attempt ZQ Calibration when not all banks are idle
                assert false
                  report "DDR3 SDRAM Main FSM (58): ZQ Calibration Short command when not all banks idle"
                  severity error;
              else
                for i in 0 to num_banks-1 loop
                  if (curr_bank_st(i) = BANK_IDLE) then
                    if not(rp_delay_tc(i)) then
                      -- t_rp violation on ZQCL command
                      assert false
                        report "DDR3 SDRAM Main FSM (59): T_RP for bank " & natural'image(i) & " violated by ZQ Calibration Short command"
                        severity error;
                    end if;
                  end if;
                end loop;

                assert not(message_in)
                  report "DDR3 SDRAM Main FSM (60): Performing ZQ Calibration Short operation"
                  severity note;
              end if;

            when MRS1 =>
              mr1_odt_hold <= transport true, false after t_mod;

              if (odt'last_event < t_aofd) then
                -- t_aofd violation entering self-refresh
                assert false
                  report "DDR3 SDRAM Main FSM (61): T_AOFD violated by MRS MR1 command"
                  severity error;
              end if;

              if (odt /= '0') then
                -- odt high entering self-refresh
                assert false
                  report "DDR3 SDRAM Main FSM (62): ODT not low during MRS MR1 command"
                  severity error;
              end if;

            when others =>
              report "DDR3 SDRAM Main FSM (63): Illegal command detected"
              severity error;
          end case;
      end case;

      if (cke_d1 /= cke) and not(cke_stable_delay_tc) then
        if (cke = '0') then
          assert false
            report "DDR3 SDRAM Main FSM (64): CKE low before " & integer'image(cke_stable_delay) & " consecutive cycles of CKE high"
            severity error;
        end if;
        if (cke = '1') then
          assert false
            report "DDR3 SDRAM Main FSM (65): CKE high before " & integer'image(cke_stable_delay) & "consecutive cycles of CKE low"
            severity error;
        end if;
      end if;
    end if;
  end process;


  -- Check ODT low hold time referenced to EMR1 command
  check_mr1_odt_hold : process(odt)
  begin
    assert not(mr1_odt_hold)
      report "DDR3 SDRAM Main FSM (66): T_MOD (MR1 to ODT high time) violated by ODT changing to " & std_logic'image(odt)
      severity error;
  end process;

  -- Check power down exit to ODT high/low
  check_pdex_odt : process(ck)
  begin
    if rising_edge(ck) then
      if (odt_d1 /= odt) then
        assert not((axpd_delay_tc = false)and((odt = '0')or(odt = '1')))
          report "DDR3 SDRAM Main FSM (67): T_AXPD (Power down exit to ODT high/low time) violated by ODT changing to " & std_logic'image(odt)
          severity error;
      end if;
    end if;
  end process;

  -- Start odt delay counters for entry to power down mode
--  start_odt_delays : process(odt,ck)
--  begin
--    if (odt'event)and((odt = '1')or(odt = '0')) then
--      anpd_delay <= t_anpd_cyc; anpd_delay_go <= true;
--    elsif rising_edge(ck) then
--      anpd_delay_go <= false;
--    end if;
--  end process;


  -- Generate next state from command
  gen_next_state : process(
    curr_st,
    curr_bank_st,
    odt,
    cke,
--    delay_500us_over,
    in_command,
    in_bank_num,
    bl_cyc,
    t_pdex_rd_cyc,
    t_xp_cyc,
    t_cke_cyc,
    t_rp_cyc,
    t_rfc_cyc,
    t_rcd_al_cyc,
    t_ras_cyc,
    t_mrd_cyc,
    t_mod,
--    t_axpd_cyc,
    t_ccd_cyc,
    t_rrd_cyc,
    t_rc_cyc,
    t_rd_rd_cyc,
    t_rd_wr_cyc,
    t_wr_wr_cyc,
    t_rd_pden_cyc,
    t_rd_pre_cyc,
    t_wr_rd_cyc,
    t_wrte_pden_cyc,
    t_wrap_pden_cyc,
    t_wr_pre_cyc,
    t_xs_cyc,
    t_xsdll_cyc,
    t_zqcs_cyc,
    t_zqoper_cyc,
    wrap_delayed,
    rdap_delayed)
  begin

    next_st       <= curr_st;
    next_bank_st  <= curr_bank_st;

    -- Assign default values for outputs/next state in states
    -- where they are not explicitly defined or in illegal states
    mrs_delay_go        <= false;
    rfc_delay_go        <= false;
    rrd_delay_go        <= false;
    ccd_delay_go        <= false;
    rtr_delay_go        <= false;
    rtw_delay_go        <= false;
    wtr_delay_go        <= false;
    wtw_delay_go        <= false;
    xs_delay_go         <= false;
    xsdll_delay_go      <= false;
    xp_delay_go         <= false;
    xard_delay_go       <= false;
    rtpden_delay_go     <= false;
    wrte_pden_delay_go  <= false;
    wrap_pden_delay_go  <= false;
    cke_stable_delay_go <= false;
    axpd_delay_go       <= false;
    rp_delay_go         <= (others => false);
    rcd_delay_go        <= (others => false);
    rasmin_delay_go     <= (others => false);
    rtp_delay_go        <= (others => false);
    wtp_delay_go        <= (others => false);
    rc_delay_go         <= (others => false);
    zqcs_delay_go       <= false;
    zqcl_delay_go       <= false;

    load_mr0       <= false;
    load_mr1       <= false;
    load_mr2       <= false;
    load_mr3       <= false;
    auto_refresh   <= false;
    self_refresh   <= false;
    power_down     <= false;
    bank_act_any   <= false;
    bank_read_any  <= false;
    bank_write_any <= false;
    bank_pre_any   <= false;
    bank_pre_all   <= false;
    rdap_command   <= (others => false);
    wrap_command   <= (others => false);
    bact_command   <= (others => false);

    a_setup   <= false;
    a10_setup <= false;
    ba_setup  <= false;

    case curr_st is
      when PWRDOWN  => if (cke = '0') then
                         power_down <= true;
                       else
                         if (in_command = SPEX) then
                           next_st <= IDLE;
                           xard_delay <= t_pdex_rd_cyc; xard_delay_go <= true;
                           xp_delay <= t_xp_cyc; xp_delay_go <= true;
                           cke_stable_delay <= t_cke_cyc; cke_stable_delay_go <= true;
--                           axpd_delay <= t_axpd_cyc; axpd_delay_go <= true;
                         else
                           power_down <= true;
                         end if;
                       end if;

      when SREFRESH => if (cke = '0') then
                         self_refresh <= true;
                       else
                         if (in_command = SPEX) then
                           next_st <= IDLE;
                           xs_delay <= t_xs_cyc; xs_delay_go <= true;  -- t_xs delay to ? command
                           xsdll_delay <= t_xsdll_cyc; xsdll_delay_go <= true;  -- t_xsdll delay to ? command
                           cke_stable_delay <= t_cke_cyc; cke_stable_delay_go <= true;
                         else
                           self_refresh <= true;
                         end if;
                       end if;

      when IDLE     => -- Handle delayed auto-precharge from a previous RDAP / WRAP command
                       for i in 0 to num_banks-1 loop
                         if wrap_delayed(i) then
                           next_bank_st(i) <= BANK_IDLE;
                           rp_delay(i) <= t_rp_cyc; rp_delay_go(i) <= true;
                         end if;

                         if rdap_delayed(i) then
                           next_bank_st(i) <= BANK_IDLE;
                           rp_delay(i) <= t_rp_cyc; rp_delay_go(i) <= true;
                         end if;
                       end loop;

                       case in_command is
                         when AREF      => rfc_delay <= t_rfc_cyc; rfc_delay_go <= true;
                                           auto_refresh <= true;

                         when NOOP|DDES => null; -- Nothing to do

                         when BACT      => bact_command(in_bank_num) <= true;
                                           next_bank_st(in_bank_num) <= BANK_OPEN;
                                           bank_act_any <= true;
                                           ba_setup <= true;
                                           a10_setup <= true;
                                           a_setup <= true;

                                           rc_delay(in_bank_num) <= t_rc_cyc; rc_delay_go(in_bank_num) <= true;
                                           rrd_delay <= t_rrd_cyc; rrd_delay_go <= true;
                                           rcd_delay(in_bank_num) <= t_rcd_al_cyc; rcd_delay_go(in_bank_num) <= true;
                                           rasmin_delay(in_bank_num) <= t_ras_cyc; rasmin_delay_go(in_bank_num) <= true;

                         when MRS0      => load_mr0 <= true;
                                           mrs_delay <= t_mrd_cyc; mrs_delay_go <= true;

                         when MRS1      => load_mr1 <= true;
                                           mrs_delay <= t_mrd_cyc; mrs_delay_go <= true;

                         when MRS2      => load_mr2 <= true;
                                           mrs_delay <= t_mrd_cyc; mrs_delay_go <= true;

                         when MRS3      => load_mr3 <= true;
                                           mrs_delay <= t_mrd_cyc; mrs_delay_go <= true;

                         when READ|RDAP => if in_command = RDAP then
                                             rdap_command(in_bank_num) <= true;
                                           end if;
                                           if wrap_delayed(in_bank_num) or rdap_delayed(in_bank_num) then
                                             -- Current command conflicts with auto-precharge from a previous RDAP/WRAP command
                                             null; -- ignore this read
                                           else
                                             bank_read_any <= true;
                                             bl_cyc_rdwr <= bl_cyc;

                                             if (in_command = RDAP) then
                                               ba_setup <= true;
                                             end if;
                                             a10_setup <= true;
                                             a_setup <= true;

                                             ccd_delay <= t_ccd_cyc; ccd_delay_go <= true;
                                             rtw_delay <= t_rd_wr_cyc; rtw_delay_go <= true;
                                             rtr_delay <= t_rd_rd_cyc; rtr_delay_go <= true;
                                             rtpden_delay <= t_rd_pden_cyc; rtpden_delay_go <= true;
                                             if in_command = RDAP then
                                               rtp_delay(in_bank_num) <= t_rd_pre_cyc;  rtp_delay_go(in_bank_num) <= true;
                                             end if;
                                           end if;

                         when WRTE|WRAP => if in_command = WRAP then
                                             wrap_command(in_bank_num) <= true;
                                           end if;
                                           if wrap_delayed(in_bank_num) or rdap_delayed(in_bank_num) then
                                             -- Current command conflicts with auto-precharge from a previous RDAP/WRAP command
                                             null; -- ignore this write
                                           else
                                             bank_write_any <= true;
                                             bl_cyc_rdwr <= bl_cyc;

                                             if (in_command = WRTE) then
                                               wrte_pden_delay <= t_wrte_pden_cyc; wrte_pden_delay_go <= true;
                                             else
                                               ba_setup <= true;
                                               wrap_pden_delay <= t_wrap_pden_cyc; wrap_pden_delay_go <= true;
                                             end if;
                                             a10_setup <= true;
                                             a_setup <= true;

                                             ccd_delay <= t_ccd_cyc;   ccd_delay_go <= true;
                                             wtr_delay <= t_wr_rd_cyc; wtr_delay_go <= true;
                                             wtw_delay <= t_wr_wr_cyc; wtw_delay_go <= true;
                                             if in_command = WRAP then
                                               wtp_delay(in_bank_num) <= t_wr_pre_cyc; wtp_delay_go(in_bank_num) <= true;
                                             end if;
                                           end if;

                         when SREN      => next_st <= SREFRESH;
                                           self_refresh <= true;
                                           cke_stable_delay <= t_cke_cyc; cke_stable_delay_go <= true;

                         when PDEN      => next_st <= PWRDOWN;
                                           power_down <= true;
                                           cke_stable_delay <= t_cke_cyc; cke_stable_delay_go <= true;

                         when APRE      => next_bank_st <= (others => BANK_IDLE);
                                           bank_pre_any <= true;
                                           bank_pre_all <= true;
                                           a10_setup <= true;
                                           rp_delay <= (others => t_rp_cyc); rp_delay_go <= (others => true);

                         when SPRE      => next_bank_st(in_bank_num) <= BANK_IDLE;
                                           bank_pre_any <= true;
                                           a10_setup <= true;
                                           ba_setup <= true;
                                           rp_delay(in_bank_num) <= t_rp_cyc; rp_delay_go(in_bank_num) <= true;

                         when ZQCS      => zqcs_delay <= t_zqcs_cyc; zqcs_delay_go <= true;

                         when ZQCL      => zqcl_delay <= t_zqoper_cyc; zqcl_delay_go <= true;

                         when others    => null; -- Do nothing
                       end case;
    end case;
  end process;

  --
  -- Delay counters that are global to the device
  --

  -- Refresh command period counter
  -- Smallest delay generated is 1 cycle
  t_rfc_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => rfc_delay_go,
      delay    => rfc_delay,
      delay_tc => rfc_delay_tc);

  -- Activate-to-activate for different banks delay counter
  -- Smallest delay generated is 1 cycle
  t_rrd_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => rrd_delay_go,
      delay    => rrd_delay,
      delay_tc => rrd_delay_tc);

  -- CAS-to-CAS delay counter
  -- Smallest delay generated is 1 cycle
  t_ccd_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => ccd_delay_go,
      delay    => ccd_delay,
      delay_tc => ccd_delay_tc);

  -- Read to read delay counter
  -- Smallest delay generated is 1 cycle
  t_rtr_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => rtr_delay_go,
      delay    => rtr_delay,
      delay_tc => rtr_delay_tc);

  -- Read to write delay counter
  -- Smallest delay generated is 1 cycle
  t_rtw_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => rtw_delay_go,
      delay    => rtw_delay,
      delay_tc => rtw_delay_tc);

  -- Write to write delay counter
  -- Smallest delay generated is 1 cycle
  t_wtw_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => wtw_delay_go,
      delay    => wtw_delay,
      delay_tc => wtw_delay_tc);

  -- Write to read delay counter
  -- Smallest delay generated is 1 cycle
  t_wtr_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => wtr_delay_go,
      delay    => wtr_delay,
      delay_tc => wtr_delay_tc);

  -- Self-refresh exit to read delay counter
  -- Smallest delay generated is 1 cycle
  t_xsrd_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => xs_delay_go,
      delay    => xs_delay,
      delay_tc => xs_delay_tc);

  -- Self-refresh exit to non-read delay counter
  -- Smallest delay generated is 1 cycle
  t_xsnr_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => xsdll_delay_go,
      delay    => xsdll_delay,
      delay_tc => xsdll_delay_tc);

  -- MR0/MR1/MR2/MR3 set command period counter
  -- Smallest delay generated is 1 cycle
  t_mrs_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => mrs_delay_go,
      delay    => mrs_delay,
      delay_tc => mrs_delay_tc);

  -- Read to power-down entry delay counter
  -- Smallest delay generated is 1 cycle
  t_rtpden_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => rtpden_delay_go,
      delay    => rtpden_delay,
      delay_tc => rtpden_delay_tc);

  -- Write to power-down entry delay counter
  -- Smallest delay generated is 1 cycle
  t_wrte_pden_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => wrte_pden_delay_go,
      delay    => wrte_pden_delay,
      delay_tc => wrte_pden_delay_tc);

  -- Write with auto-precharge to power-down entry delay counter
  -- Smallest delay generated is 1 cycle
  t_wrap_pden_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => wrap_pden_delay_go,
      delay    => wrap_pden_delay,
      delay_tc => wrap_pden_delay_tc);

  -- CKE stable delay counter
  -- Smallest delay generated is 1 cycle
  cke_stable_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => cke_stable_delay_go,
      delay    => cke_stable_delay,
      delay_tc => cke_stable_delay_tc);

  -- Power-down exit to read delay counter
  -- Smallest delay generated is 1 cycle
  t_xard_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => xard_delay_go,
      delay    => xard_delay,
      delay_tc => xard_delay_tc);

  -- Power-down exit to non-read delay counter (any bank)
  -- Smallest delay generated is 1 cycle
  t_xp_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => xp_delay_go,
      delay    => xp_delay,
      delay_tc => xp_delay_tc);

  -- odt high/low to Power-down entry
  -- Smallest delay generated is 1 cycle
  t_anpd_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => anpd_delay_go,
      delay    => anpd_delay,
      delay_tc => anpd_delay_tc);

  -- Power-down entry to odt high/low
  -- Smallest delay generated is 1 cycle
  t_axpd_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => axpd_delay_go,
      delay    => axpd_delay,
      delay_tc => axpd_delay_tc);

  -- ZQ Calibration Short delay
  -- Smallest delay generated is 1 cycle
  t_zqoper_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => zqcs_delay_go,
      delay    => zqcs_delay,
      delay_tc => zqcs_delay_tc);

  -- ZQ Calibration Long delay
  -- Smallest delay generated is 1 cycle
  t_zqinit_count : ddr3_sdram_dly_cnt
    port map(
      ck       => ck,
      delay_go => zqcl_delay_go,
      delay    => zqcl_delay,
      delay_tc => zqcl_delay_tc);

  --
  -- Per-bank delay counters
  --

  bank_counters : for i in 0 to num_banks-1 generate

    -- Precharge command period counter
    t_rp_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => rp_delay_go(i),
        delay    => rp_delay(i),
        delay_tc => rp_delay_tc(i));

    -- RAS-to-CAS delay counter
    t_rcd_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => rcd_delay_go(i),
        delay    => rcd_delay(i),
        delay_tc => rcd_delay_tc(i));

    -- Read-to-precharge delay counter
    t_rtp_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => rtp_delay_go(i),
        delay    => rtp_delay(i),
        delay_tc => rtp_delay_tc(i));

    -- Write-to-precharge delay counter
    t_wtp_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => wtp_delay_go(i),
        delay    => wtp_delay(i),
        delay_tc => wtp_delay_tc(i));

    -- RAS period (minimum) delay counter
    t_rasmin_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => rasmin_delay_go(i),
        delay    => rasmin_delay(i),
        delay_tc => rasmin_delay_tc(i));

    -- Activate-to-activate for same bank delay counter
    -- Smallest delay generated is 1 cycle
    t_rc_count : ddr3_sdram_dly_cnt
      port map(
        ck       => ck,
        delay_go => rc_delay_go(i),
        delay    => rc_delay(i),
        delay_tc => rc_delay_tc(i));
  end generate;

  -- Per-bank delay pipelines
  bank_pipes : for i in 0 to num_banks-1 generate
    -- t_rd_pre_cyc minimum value is 4
    -- t_wr_pre_cyc minimum value is 5
    -- t_ras_cyc minimum value is 15
    pipe_p : process(ck)
    begin
      if rising_edge(ck) then
        if (cke = '1') then
          if (rtp_delay(i) > 0) then
            rd_pre_pipe(i)(1)                 <= rdap_command(i);
            rd_pre_pipe(i)(2 to rtp_delay(i)) <= rd_pre_pipe(i)(1 to rtp_delay(i) - 1);
          end if;
          if (wtp_delay(i) > 0) then
            wr_pre_pipe(i)(1)                 <= wrap_command(i);
            wr_pre_pipe(i)(2 to wtp_delay(i)) <= wr_pre_pipe(i)(1 to wtp_delay(i) - 1);
          end if;
          if (t_ras_cyc > 0) then
            ras_pipe(i)(1)              <= bact_command(i);
            ras_pipe(i)(2 to t_ras_cyc) <= ras_pipe(i)(1 to t_ras_cyc - 1);
          end if;
        end if;
      end if;
    end process;

    -- Device does a precharge internally 't_rd_pre_cyc' cycles after RDAP command AS LONG AS t_ras(min) is satisfied.
    -- Device does a precharge internally 't_wr_pre_cyc' cycles after WRAP command AS LONG AS t_ras(min) is satisfied.
    -- If not satisfied, device delays the internal precharge until t_ras(min) is satisfied.
    rd_pre_delayed(i) <= rd_pre_pipe(i)(rtp_delay(i)) when (rtp_delay(i) > 0) else false;
    wr_pre_delayed(i) <= wr_pre_pipe(i)(wtp_delay(i)) when (wtp_delay(i) > 0) else false;
    ras_delayed(i)    <= ras_pipe(i)(t_ras_cyc) when (t_ras_cyc > 0) else false;

    extend_p : process(ck)
    begin
      if rising_edge(ck) then
        if (cke = '1') then
          if rd_pre_delayed(i) and not(rasmin_delay_tc(i)) then
            ext_rd_pre(i) <= true;
          elsif rdap_delayed(i) then
            ext_rd_pre(i) <= false;
          end if;
          if wr_pre_delayed(i) and not(rasmin_delay_tc(i)) then
            ext_wr_pre(i) <= true;
          elsif wrap_delayed(i) then
            ext_wr_pre(i) <= false;
          end if;
        end if;
      end if;
    end process;

    rdap_delayed(i) <= (rd_pre_delayed(i) and rasmin_delay_tc(i)) or (ras_delayed(i) and ext_rd_pre(i));
    wrap_delayed(i) <= (wr_pre_delayed(i) and rasmin_delay_tc(i)) or (ras_delayed(i) and ext_wr_pre(i));
  end generate;

end behavioural;