--==============================================================================
--
--     File Name: simple_l.vhd
--       Version: 0.2
--          Date: 21 Feb 2011
--         Model: syn (Synthesisable)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 Target Design
--
--   Description: Illustrates:
--                  OCP Direct Slave channel:
--                    32-bit local bus emulation register interface (simple)
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Mar 10: Initial design
--                0.2, TW, 21 Feb 2011: Removed ref_clk_bufg to remove
--                  uncertainty about name of usr_clk/ref_clk_bufg.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity simple is
port(
  -- Non-MGT Clock Interface
  ref_clk                  : in    std_logic_dbl_t;
  -- MGT Clock Interface
  mptl_clk                 : in    std_logic_dbl_t;
  -- MPTL Interface
  mptl_b2t                 : in    mptl_pins_t;
  mptl_t2b                 : out   mptl_pins_t;
  -- MPTL Initialisation
  mptl_sb_b2t              : in    mptl_sb_b2t_t;
  mptl_sb_t2b              : out   mptl_sb_t2b_t);
end simple;


architecture synth of simple is

  -- Width of data registers in bits, must be 8/16/32/64/128 bytes
  constant DS_DATA_WIDTH : natural range 8 to 128 := 32;

  ------------------------
  -- User clock signals --
  ------------------------
  signal ref_clk_ibuf   : std_logic;
  signal usr_clk        : std_logic;

  ----------------------------
  -- MPTL interface signals --
  ----------------------------
  signal direct_slave_m2s   : adb3_ocp_m2sT;
  signal direct_slave_s2m   : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;
  signal dma_channels_m2s   : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  signal dma_channels_s2m   : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);
  signal gpio_b2t           : std_logic_vector(15 downto 0);
  signal gpio_t2b           : std_logic_vector(15 downto 0) := (others => '0');

  --------------------------------------------------------------
  -- Direct Slave OCP interface signals ('local bus' drop-in) --
  --------------------------------------------------------------
  signal usr_rst  : std_logic := '1';
  signal ld_o     : std_logic_vector(DS_DATA_WIDTH-1 downto 0);
  signal ld_i     : std_logic_vector(DS_DATA_WIDTH-1 downto 0);
  signal lbe_i    : std_logic_vector((DS_DATA_WIDTH/8)-1 downto 0);
  signal ds_write : std_logic;
  signal ds_read  : std_logic;
  signal la_q     : std_logic_vector(DS_ADDR_WIDTH-1 downto 0);

  --------------------------------
  -- Register interface signals --
  --------------------------------
  signal wdata_reg   : std_logic_vector(DS_DATA_WIDTH-1 downto 0);
  signal rdata_reg   : std_logic_vector(DS_DATA_WIDTH-1 downto 0);

begin

  --------------------------
  -- User clock interface --
  --------------------------

  -- Use reference clock (200MHz oscillator) as design reference clock
  ref_clk_dif_i : IBUFDS
  port map(
    I  => ref_clk.p,
    IB => ref_clk.n,
    O  => ref_clk_ibuf);

  ref_clk_bufg_i : BUFG
  port map(
    I => ref_clk_ibuf,
    O => usr_clk);

  -- There is no reset output present in the ADB3 OCP interface
  -- Direct slave register reset may be user generated if required
  usr_rst <= '0';


  -------------------------------------
  -- Direct Slave register interface --
  -------------------------------------

  --
  -- If the current cycle is a write, update the registers
  --

  update_regs : process(usr_rst,usr_clk)
  begin
    if usr_rst = '1' then
      wdata_reg <= (others => '0');
    elsif rising_edge(usr_clk) then
      if ds_write = '1' then
        if la_q(2) = '0' then
          for i in 0 to 3 loop
            if lbe_i(i) = '1' then
              wdata_reg(8*(i+1)-1 downto 8*i) <= ld_i(8*(i+1)-1 downto 8*i);
            end if;
          end loop;
        end if;
      end if;
    end if;
  end process;

  --
  -- Generate the 'ld_o' signal, outbound data for 'local bus'
  --

  generate_ld_o : process(usr_rst,usr_clk)
  begin
    if usr_rst = '1' then
      ld_o <= (others => '0');
    elsif rising_edge(usr_clk) then
      if la_q(2) = '0' then
        ld_o <= rdata_reg;
      else
        ld_o <= (others => '-');
      end if;
    end if;
  end process;


  -----------------------------------------------------
  -- simple: example of direct slave register access --
  -----------------------------------------------------

  -- Uses direct slave registers:
  --   wdata_reg
  --   rdata_reg

  rdata_reg <= wdata_reg(3  downto 0)  & wdata_reg(7  downto 4)  &
               wdata_reg(11 downto 8)  & wdata_reg(15 downto 12) &
               wdata_reg(19 downto 16) & wdata_reg(23 downto 20) &
               wdata_reg(27 downto 24) & wdata_reg(31 downto 28);


  -----------------------------------------------------------
  -- Direct Slave register interface ('local bus' drop-in) --
  -----------------------------------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width => DS_ADDR_WIDTH, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width => DS_DATA_WIDTH) -- natural range 8 to ADB3_OCP_DS_DATA_WIDTH
  port map(
    -- OCP Interface
    ocp_rst => usr_rst,           -- in    std_logic
    ocp_clk => usr_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ld_i,              -- out   std_logic_vector(DS_DATA_WIDTH-1 downto 0)
    q       => ld_o,              -- in    std_logic_vector(DS_DATA_WIDTH-1 downto 0)
    a       => la_q,              -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_write,          -- out   std_logic
    r       => ds_read,           -- out   std_logic
    we      => lbe_i);            -- out   std_logic_vector((DS_DATA_WIDTH/8)-1 downto 0))


  ---------------------
  -- MPTL interface --
  ---------------------

  mptl_if_target_wrap_i : mptl_if_target_wrap
  port map(
    -- OCP Interface
    ocp_clk                  => usr_clk,          -- in    std_logic
    direct_slave_m2s         => direct_slave_m2s, -- out   adb3_ocp_m2sT
    direct_slave_s2m         => direct_slave_s2m, -- in    adb3_ocp_s2mT
    dma_channels_m2s         => dma_channels_m2s, -- out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m         => dma_channels_s2m, -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    -- MPTL Interface
    mptl_t2b                 => mptl_t2b,         -- out   mptl_pins_t;
    mptl_b2t                 => mptl_b2t,         -- in    mptl_pins_t;
    mptl_clk                 => mptl_clk,         -- in    std_logic_dbl_t;
    -- MPTL Initialisation
    ocp_ready                => '1',              -- in    std_logic;
    mptl_sb_b2t              => mptl_sb_b2t,      -- in    mptl_sb_b2t_t;
    mptl_sb_t2b              => mptl_sb_t2b,      -- out   mptl_sb_t2b_t;
    -- MPTL Sideband
    gpio_t2b                 => gpio_t2b,         -- in    std_logic_vector(15 downto 0)
    gpio_b2t                 => gpio_b2t);        -- out   std_logic_vector(15 downto 0)

end;