--==============================================================================
--
--     File Name: test_simple.vhd
--       Version: 0.1
--          Date: 23 Mar 2010
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: simple_l
--
--   Description: Illustrates with ADB3:
--                  Register access using direct slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Mar 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.adb3_probe_pkg.all;

entity test_simple is
end test_simple;

architecture test of test_simple is

  -- Fixed Clock periods
  constant CLK_250_PERIOD : time :=  4.00 ns;    -- 250 MHz
  constant CLK_200_PERIOD : time :=  5.00 ns;    -- 200 MHz
  constant CLK_125_PERIOD : time :=  8.00 ns;    -- 125 MHz
  constant CLK_100_PERIOD : time := 10.00 ns;    -- 100 MHz
  constant CLK_80_PERIOD  : time := 12.50 ns;    --  80 MHz

  -- Clock frequencies in MHz
  constant CLK_250_FREQ   : natural := 1 us / CLK_250_PERIOD;
  constant CLK_200_FREQ   : natural := 1 us / CLK_200_PERIOD;
  constant CLK_125_FREQ   : natural := 1 us / CLK_125_PERIOD;
  constant CLK_100_FREQ   : natural := 1 us / CLK_100_PERIOD;
  constant CLK_80_FREQ    : natural := 1 us / CLK_80_PERIOD;

  -- Constant to control OCP transaction probe file logging
  constant PROBE_LOG      : boolean := false;

  -- Test clocks
  signal clk_250 : std_logic_dbl_t := ('1','0');
  signal clk_200 : std_logic_dbl_t := ('1','0');
  signal clk_125 : std_logic_dbl_t := ('1','0');
  signal clk_100 : std_logic_dbl_t := ('1','0');
  signal clk_80  : std_logic_dbl_t := ('1','0');

  -- Test results
  signal simple_complete : boolean;
  signal simple_passed   : boolean;
  signal test_complete   : boolean := false;
  signal test_passed     : boolean := true;

  -- Bridge clocks
  signal ocp_clk_full : std_logic;
  signal ocp_clk      : std_logic;
  signal mptl_clk     : std_logic_dbl_t;

  -- Bridge MPTL Sideband signals
  signal b_gpio_b2t,b_gpio_t2b : std_logic_vector(15 downto 0);
  signal b_mptl_status         : std_logic_vector(63 downto 0);
  signal b_mptl_config         : std_logic_vector(31 downto 0);

  --------------------
  -- Target signals --
  --------------------

  -- clocks
  signal ref_clk : std_logic_dbl_t;

  -- Configuration/online signals
  signal mptl_online         : std_logic;
  signal mptl_online_cnt     : std_logic_vector(7 downto 0);
  signal mptl_online_long    : std_logic;
  signal mptl_online_long_d1 : std_logic;

  -- MPTL signals
  signal mptl_b2t : mptl_pins_t;
  signal mptl_t2b : mptl_pins_t;

  -- MPTL sideband signals
  signal mptl_sb_b2t : mptl_sb_b2t_t;
  signal mptl_sb_t2b : mptl_sb_t2b_t;

  -- OCP signals
  signal direct_slave_m2s   : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal direct_slave_s2m   : adb3_ocp_s2mT;
  signal dma_channels_m2s   : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_M2S);
  signal dma_channels_s2m   : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  signal unused_s2m         : adb3_ocp_s2mT_array(0 to DM_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);


  component simple
  port(
    -- Clock Interface
    ref_clk                  : in    std_logic_dbl_t;
    mptl_clk                 : in    std_logic_dbl_t;
    -- MPTL Interface
    mptl_b2t                 : in    mptl_pins_t;
    mptl_t2b                 : out   mptl_pins_t;
    -- MPTL Sideband Signals
    mptl_sb_b2t              : in    mptl_sb_b2t_t;
    mptl_sb_t2b              : out   mptl_sb_t2b_t);
  end component;

begin

  -- Generate unique reference clock frequencies
  -- 250 MHz clock
  clk_250.p <= not clk_250.p after CLK_250_PERIOD / 2;
  clk_250.n <= not clk_250.n after CLK_250_PERIOD / 2;

  -- 200 MHz clock
  clk_200.p <= not clk_200.p after CLK_200_PERIOD / 2;
  clk_200.n <= not clk_200.n after CLK_200_PERIOD / 2;

  -- 125 MHz clock
  clk_125.p <= not clk_125.p after CLK_125_PERIOD / 2;
  clk_125.n <= not clk_125.n after CLK_125_PERIOD / 2;

  -- 100 MHz clock
  clk_100.p <= not clk_100.p after CLK_100_PERIOD / 2;
  clk_100.n <= not clk_100.n after CLK_100_PERIOD / 2;

  --  80 MHz clock
  clk_80.p  <= not clk_80.p  after CLK_80_PERIOD / 2;
  clk_80.n  <= not clk_80.n  after CLK_80_PERIOD / 2;

  -- Report design constants
  report_setup_p : process
  begin
    report "Board Type : " & board_type_t'image(BOARD_TYPE);  -- adb3_target_inc_pkg
    report "Target Use : " & target_use_t'image(TARGET_USE);  -- adb3_target_inc_pkg
    report "Waiting for MPTL online....";
    wait;
  end process;


  ---------------------------------
  -- MPTL Direct Slave interface --
  ---------------------------------

  direct_slave_p : process
    variable reg_data_wr    : byte_vector_t(0 to 127);
    variable reg_data_rd    : byte_vector_t(0 to 127);
    variable reg_be         : byte_enable_t(0 to 127);
    variable reg_addr       : adb3_ocp_addr_s;
    variable reg_xfer       : natural;

  begin
    wait until (mptl_online_long = '1');  -- OCP commands ignored until MPTL online
    simple_complete <= false;
    simple_passed <= true;

    --------------------------
    -- Simple Register Test --
    --------------------------
    reg_be(0 to 3) := conv_byte_enable("1111");

    -- Write a pattern to register
    reg_addr := (others => '0');
    reg_data_wr(0 to 3) := conv_byte_vector(X"cafeface");

    adb3_target_sim_write_reg32(reg_addr,10,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,direct_slave_m2s,direct_slave_s2m);
    report "Wrote simple DATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));


    -- Read back nibble reversed pattern from register
    reg_addr := (others => '0');
    adb3_target_sim_read_reg32(reg_addr,11,reg_data_rd(0 to 3),reg_xfer,ocp_clk,direct_slave_m2s,direct_slave_s2m);
    report "Read simple DATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    simple_complete <= true;
    if (reg_data_rd(0 to 3) = conv_byte_vector(X"ecafefac")) then
      assert false  report "Test Simple completed: PASSED." severity note;
    else
      assert false  report "Test Simple completed: FAILED." severity error;
      simple_passed <= false;
    end if;

    wait;
  end process;


  -------------------------
  -- Output test results --
  -------------------------

  gen_results_p : process(simple_passed,simple_complete)
  begin
      test_passed <= simple_passed;

      test_complete <= simple_complete;
  end process;

  test_results_p : process(test_complete,test_passed)
  begin
    if (test_complete) then
      if (test_passed) then
        assert false report "Test of design SIMPLE completed: PASSED." severity failure;
      else
        assert false report "Test of design SIMPLE completed: FAILED." severity failure;
      end if;
    end if;
  end process;


  -------------------------------------
  -- Target Component Instantiations --
  -------------------------------------

  -- Connect clock inputs
  ref_clk_g : if (REF_CLK_VALID = '1') generate
    ref_clk <= clk_200; -- Osc. Reference clock 200 MHz (double ended)
  end generate;

  mgt_clks_6tl_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- MPTL Parallel interface: Connect 250MHz PCIe input (mgt114 clk0) to mptl_clk (double ended)
    mptl_clk <= clk_250;
  end generate;

  mgt_clks_6t1_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    -- MPTL serial interface: Connect 250 MHz Osc input (mgt115 clk0) to mptl_clk (double ended)
    mptl_clk <= clk_250;
  end generate;


  uut : simple
  port map(
    -- Clock Interface
    ref_clk                  => ref_clk,       -- in    std_logic_dbl_t
    mptl_clk                 => mptl_clk,      -- in    std_logic_dbl_t
    -- MPTL Interface
    mptl_b2t                 => mptl_b2t,      -- in    mptl_pins_t
    mptl_t2b                 => mptl_t2b,      -- out   mptl_pins_t
    -- MPTL Initialisation
    mptl_sb_b2t              => mptl_sb_b2t,   -- in    mptl_sb_b2t_t
    mptl_sb_t2b              => mptl_sb_t2b);  -- out   mptl_sb_t2b_t


  -----------------------------------------------
  -- Bridge Component Instantiations (Private) --
  -----------------------------------------------

  -- Connect inputs
  b_gpio_b2t    <= (others => '0');
  b_mptl_config <= (others => '0');

  -- Full OCP-MPTL-OCP simulation
  -- Bridge ocp clock can be different from target ocp clock
  ocp_clk_full <= clk_250.p;


  -- bridge interface
  mptl_if_bridge_wrap_i : mptl_if_bridge_wrap
  port map(
    -- OCP Interface
    ocp_clk_in               => ocp_clk_full,        -- in    std_logic
    direct_slave_m2s         => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m         => direct_slave_s2m,    -- out   adb3_ocp_s2mT
    dma_channels_m2s         => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array
    dma_channels_s2m         => dma_channels_s2m,    -- out   adb3_ocp_s2mT_array
    direct_masters_m2s       => open,                -- out   adb3_ocp_m2sT
    direct_masters_s2m       => unused_s2m,          -- in    adb3_ocp_s2mT
    dma_abort                => open,                -- out   std_logic_vector(DMA_CHANNELS-1 downto 0);
    ocp_clk_out              => ocp_clk,             -- out   std_logic
    -- MPTL Interface
    mptl_b2t                 => mptl_b2t,            -- out   mptl_b2t_t;
    mptl_t2b                 => mptl_t2b,            -- in    mptl_t2b_t;
    mptl_clk                 => mptl_clk,            -- in    std_logic_dbl_t;
    -- MPTL Initialisation
    mptl_sb_t2b              => mptl_sb_t2b,         -- in    mptl_sb_t2b_t;
    mptl_sb_b2t              => mptl_sb_b2t,         -- out   mptl_sb_b2t_t;
    -- MPTL Sideband Interface
    gpio_b2t                 => b_gpio_b2t,          -- in    std_logic_vector(15 downto 0)
    gpio_t2b                 => b_gpio_t2b,          -- out   std_logic_vector(15 downto 0)
    mptl_status              => b_mptl_status,       -- out   std_logic_vector(63 downto 0))
    mptl_config              => b_mptl_config);      -- in    std_logic_vector(31 downto 0))


  -- Generate online signal
  mptl_online <= not(mptl_sb_b2t.mptl_bridge_gtp_online_l or mptl_sb_t2b.mptl_target_gtp_online_l);


  -- Count cycles of mptl_online to make sure it is stable
  online_count_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (mptl_online /= '1') then
        mptl_online_cnt <= (others => '1');
      else
        if (or_reduce(mptl_online_cnt) = '1') then
          mptl_online_cnt <= mptl_online_cnt - 1;
        end if;
      end if;
      mptl_online_long <= not(or_reduce(mptl_online_cnt));
    end if;
  end process;

  -- Monitor MPTL link and stop when failure
  monitor_status_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      mptl_online_long_d1 <= mptl_online_long;
    end if;

    -- Detect falling edge on mptl_online
    assert not((mptl_online_long = '0')and(mptl_online_long_d1 = '1')) report "Test of design SIMPLE: MPTL link failure (Undefined data)" severity failure;
  end process;


  ----------------------------
  -- OCP transaction probes --
  ----------------------------

  -- Direct Slave transaction probe
  ds_probe_i : adb3_ocp_transaction_probe
  generic map(
    enable_logging   => PROBE_LOG,
    int_log_filename => "ds_probe",
    addr_align_bits  => 1,
    addr_width_max   => DS_ADDR_WIDTH)
  port map(
    ocp_clk => ocp_clk,
    ocp_m2s => direct_slave_m2s,
    ocp_s2m => direct_slave_s2m,
    status  => open);

end test;