--==============================================================================
--
--     File Name: blk_bram.vhd
--       Version: 0.3
--          Date: 14 Feb 2011
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: BRAM instantiation block
--
--   Description: Implements a 512 kiB block of memory, organised as 2^15
--                words of 16 bytes, with an OCP interface. Bytes are
--                individually writeable.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 12 Mar 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Replicated and registered address signals and increased
--                    read latency in adb3_ocp_simple_bus_if instance to 2, in
--                    order to more easily meet timing.
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                0.3, TW, 14 Feb 2011:
--                  - Now natively OCP, instead of using an instance of
--                    adb3_ocp_simple_bus_if, for higher performance.
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

entity blk_bram_read_fifo is
port(
  rst         : in    std_logic;
  clk         : in    std_logic;
  d           : in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  tag         : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  wr          : in    std_logic;
  adv         : in    std_logic;
  q           : out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  qtag        : out   std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  resp        : out   ocp_RespT;
  nfull       : out   std_logic);
end;

architecture synth of blk_bram_read_fifo is

  constant FIFO_ORDER : natural := 3;
  constant FIFO_DEPTH : natural := 2**FIFO_ORDER;
  constant FIFO_MAX_LEVEL : natural := FIFO_DEPTH - 1;

  constant DATA_LENGTH : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;
  subtype ram_data_t is std_logic_vector(DATA_LENGTH - 1 downto 0);
  type ram_array_t is array(0 to FIFO_DEPTH - 1) of ram_data_t;

  signal wptr, rptr : std_logic_vector(FIFO_ORDER - 1 downto 0);
  signal wn, rn : std_logic_vector(FIFO_ORDER - 1 downto 0);
  signal not_empty : std_logic;

  signal qvalid : std_logic;
  signal qce : std_logic;
  signal rd : std_logic;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

  signal i_resp : ocp_RespT := OCP_RESP_NONE;

begin

  resp <= i_resp;
  nfull <= wn(FIFO_ORDER - 1);

  write_logic : process(rst, clk)
  begin
    if rst = '1' then
      wn(wn'high) <= '0';
      wn(wn'high - 1 downto 0) <= (others => '-');
      wptr <= (others => '0');
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert CONV_INTEGER(wptr - rptr) < FIFO_MAX_LEVEL severity failure;
        wn <= wptr - rptr + 1;
      else
        wn <= wptr - rptr;
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= tag & d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        assert CONV_INTEGER(wptr - rptr) > 0 severity failure;
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
      i_resp <= OCP_RESP_NONE;
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
        if not_empty = '1' then
          i_resp <= OCP_RESP_VALID;
        else
          i_resp <= OCP_RESP_NONE;
        end if;
      end if;
    end if;

    if rising_edge(clk) then
      if qce = '1' then
        q <= sel_ram(ADB3_OCP_DATA_WIDTH - 1 downto 0);
        qtag <= sel_ram(ADB3_OCP_DATA_WIDTH + ADB3_OCP_TAG_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
      end if;
    end if;
  end process;

end;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

entity blk_bram is
port(
  ocp_rst     : in    std_logic;
  ocp_clk     : in    std_logic;
  -- BRAM OCP Interface
  bram_if_m2s : in    adb3_ocp_m2sT;
  bram_if_s2m : out   adb3_ocp_s2mT);
end;

architecture synth of blk_bram is

  component blk_bram_read_fifo
  port(
    rst         : in    std_logic;
    clk         : in    std_logic;
    d           : in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    tag         : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    wr          : in    std_logic;
    adv         : in    std_logic;
    q           : out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    qtag        : out   std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    resp        : out   ocp_RespT;
    nfull       : out   std_logic);
  end component;

  component bram_single_wrap
  generic(
    output_reg : integer := 0);
  port(
    -- Port Interface
    rst  : in    std_logic;
    clk  : in    std_logic;
    addr : in    std_logic_vector(14 downto 0);
    we   : in    std_logic;
    di   : in    std_logic;
    do   : out   std_logic);
  end component;

  constant ORDER : natural := 15;
  constant DEPTH : natural := 2**ORDER;

  attribute equivalent_register_removal : string;
  attribute ram_style : string;

  signal i_bram_if_s2m : adb3_ocp_s2mT;

  signal cmd_accept : std_logic;
  signal write_xfer : std_logic;
  signal read_xfer : std_logic;

  type bram_addr_array_t is array(0 to ADB3_OCP_BE_WIDTH-1) of std_logic_vector(ORDER-1 downto 0);
  signal bram_a  : bram_addr_array_t;
  attribute equivalent_register_removal of bram_a : signal is "no";
  signal bram_d  : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  signal bram_we : std_logic_vector(ADB3_OCP_BE_WIDTH-1 downto 0) := (others => '0');
  signal bram_read : std_logic;
  signal bram_read_q : std_logic;
  signal bram_q  : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  signal bram_q_valid : std_logic;

  signal bs_read : std_logic;

  type bs_state_t is (bss_idle, bss_read, bss_read_drain, bss_writing);
  signal bs_state, n_bs_state : bs_state_t;
  signal bs_idle, n_bs_idle : std_logic;
  signal bs_data_acc, n_bs_data_acc : std_logic;
  signal bs_reading, n_bs_reading : std_logic;
  signal bs_blen : adb3_ocp_burst_s;
  signal bs_blast : std_logic;
  signal bs_rlen : adb3_ocp_burst_s;
  signal bs_rlast : std_logic;
  signal bs_addr : std_logic_vector(14 downto 0);
  signal bs_tag : adb3_ocp_tag_s;

  signal fifo_nfull : std_logic;

begin

  bram_if_s2m <= i_bram_if_s2m;

  i_bram_if_s2m.DataAccept <= bs_data_acc;
  i_bram_if_s2m.CmdAccept <= bs_idle;

  cmd_accept <= '1' when bs_idle = '1' and (bram_if_m2s.Cmd = OCP_CMD_READ or bram_if_m2s.Cmd = OCP_CMD_WRITE) else '0';
  write_xfer <= '1' when bram_if_m2s.DataValid = '1' and i_bram_if_s2m.DataAccept = '1' else '0';
  read_xfer <= '1' when bram_if_m2s.RespAccept = '1' and i_bram_if_s2m.Resp = OCP_RESP_VALID else '0';

  bs_read <= bs_reading and not fifo_nfull;

  gen_bs_state : process(ocp_rst, ocp_clk)
    variable tmp : std_logic_vector(1 downto 0);
  begin
    if ocp_rst = '1' then
      bs_state <= bss_idle;
      bs_idle <= '1';
      bs_data_acc <= '0';
      bs_reading <= '0';
    elsif rising_edge(ocp_clk) then
      bs_state <= n_bs_state;
      bs_idle <= n_bs_idle;
      bs_data_acc <= n_bs_data_acc;
      bs_reading <= n_bs_reading;
    end if;

    if rising_edge(ocp_clk) then
      tmp(0) := cmd_accept;
      tmp(1) := read_xfer or write_xfer;
      case tmp is
        when "01" =>
          bs_blen <= bram_if_m2s.BurstLength;

        when "10" =>
          bs_blen <= bs_blen - 1;

        when "00" =>
          null;

        when others =>
          -- illegal state
          bs_blen <= (others => '-');
          assert false severity failure;
      end case;

      tmp(0) := cmd_accept;
      tmp(1) := bs_read;
      case tmp is
        when "01" =>
          bs_rlen <= bram_if_m2s.BurstLength;

        when "10" =>
          bs_rlen <= bs_rlen - 1;

        when "00" =>
          null;

        when others =>
          -- illegal state
          bs_rlen <= (others => '-');
          assert false severity failure;
      end case;

      tmp(0) := cmd_accept;
      tmp(1) := bs_read or write_xfer;
      case tmp is
        when "01" =>
          bs_addr <= bram_if_m2s.Addr(18 downto 4);

        when "10" =>
          bs_addr <= bs_addr + 1;

        when "00" =>
          null;

        when others =>
          -- illegal state
          bs_addr <= (others => '-');
          assert false severity failure;
      end case;

      if cmd_accept = '1' then
        bs_tag <= bram_if_m2s.Tag;
      end if;
    end if;
  end process;

  bs_blast <= not OR_reduce(bs_blen(bs_blen'high downto 1));
  bs_rlast <= not OR_reduce(bs_rlen(bs_rlen'high downto 1));

  gen_n_bs_state : process(
    bs_state,
    bs_rlast,
    bs_read,
    read_xfer,
    write_xfer,
    bs_blast,
    bram_if_m2s)
  begin
    case bs_state is
      when bss_idle =>
        case bram_if_m2s.Cmd is
          when OCP_CMD_READ =>
            n_bs_state <= bss_read;
            n_bs_idle <= '0';
            n_bs_data_acc <= '0';
            n_bs_reading <= '1';

          when OCP_CMD_WRITE =>
            n_bs_state <= bss_writing;
            n_bs_idle <= '0';
            n_bs_data_acc <= '1';
            n_bs_reading <= '0';

          when others =>
            n_bs_state <= bss_idle;
            n_bs_idle <= '1';
            n_bs_data_acc <= '0';
            n_bs_reading <= '0';
        end case;

      when bss_read =>
        if bs_read = '1' then
          if bs_rlast = '1' then
            n_bs_state <= bss_read_drain;
            n_bs_idle <= '0';
            n_bs_data_acc <= '0';
            n_bs_reading <= '0';
          else
            n_bs_state <= bss_read;
            n_bs_idle <= '0';
            n_bs_data_acc <= '0';
            n_bs_reading <= '1';
          end if;
        else
          n_bs_state <= bss_read;
          n_bs_idle <= '0';
          n_bs_data_acc <= '0';
          n_bs_reading <= '1';
        end if;

      when bss_read_drain =>
        if read_xfer = '1' then
          if bs_blast = '1' then
            n_bs_state <= bss_idle;
            n_bs_idle <= '1';
            n_bs_data_acc <= '0';
            n_bs_reading <= '0';
          else
            n_bs_state <= bss_read_drain;
            n_bs_idle <= '0';
            n_bs_data_acc <= '0';
            n_bs_reading <= '0';
          end if;
        else
          n_bs_state <= bss_read_drain;
          n_bs_idle <= '0';
          n_bs_data_acc <= '0';
          n_bs_reading <= '0';
        end if;

      when bss_writing =>
        if write_xfer = '1' then
          if bs_blast = '1' then
            n_bs_state <= bss_idle;
            n_bs_idle <= '1';
            n_bs_data_acc <= '0';
            n_bs_reading <= '0';
          else
            n_bs_state <= bss_writing;
            n_bs_idle <= '0';
            n_bs_data_acc <= '1';
            n_bs_reading <= '0';
          end if;
        else
          n_bs_state <= bss_writing;
          n_bs_idle <= '0';
          n_bs_data_acc <= '1';
          n_bs_reading <= '0';
        end if;
    end case;
  end process;

  register_data_in : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      bram_d <= bram_if_m2s.Data;
      if write_xfer = '1' then
        bram_we <= bram_if_m2s.DataByteEn;
      else
        bram_we <= (others => '0');
      end if;
      bram_a <= (others => bs_addr);
      bram_read <= bs_read;
      bram_read_q <= bram_read;
      bram_q_valid <= bram_read_q;
    end if;
  end process;

  -- Generate 32k x 128 bits single port BRAM
  bram_bytes_loop_g : for m in 0 to ADB3_OCP_BE_WIDTH-1 generate
    bram_bits_loop_g : for n in 0 to 7 generate
      bram_single_wrap_i : bram_single_wrap
      generic map(
        output_reg => 1) -- integer := 0);
      port map(
        rst  => ocp_rst,                     -- in    std_logic
        clk  => ocp_clk,                     -- in    std_logic
        addr => bram_a(m),                   -- in    std_logic_vector(14 downto 0)
        we   => bram_we(m),                  -- in    std_logic
        di   => bram_d(8*m+n),               -- in    std_logic
        do   => bram_q(8*m+n));              -- out   std_logic)
    end generate;
  end generate;

  -- Read data FIFO
  blk_bram_read_fifo_i : blk_bram_read_fifo
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    d => bram_q,
    tag => bs_tag,
    wr => bram_q_valid,
    adv => bram_if_m2s.RespAccept,
    q => i_bram_if_s2m.Data,
    qtag => i_bram_if_s2m.Tag,
    resp => i_bram_if_s2m.Resp,
    nfull => fifo_nfull);

end;
