--==============================================================================
--
--     File Name: test_uber.vhd
--       Version: 0.1
--          Date: 26 Nov 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: uber
--
--   Description: Illustrates with ADB3:
--                  Register access using direct slave interface
--                  BRAM access using direct slave interface
--                  Multi-channel DMA
--                  Direct master interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 26 Oct 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.mem_if_pkg.all;
use work.uber_pkg.all;
use work.uber_tb_pkg.all;
use work.ddr3_sdram_pkg.all;

entity test_uber is
end test_uber;

architecture test of test_uber is

  -- Constant for MPTL GPIO loopback test
  constant MPTL_GPIO : std_logic_vector(15 downto 0) := X"F1D0";

  -- Ram banks to instantiate
  constant DDR3_ST_BANK     : natural range 0 to MEM_BANKS-1 := 0;
  constant DDR3_MODEL_BANKS : natural range 0 to MEM_BANKS   := DDR3_BANKS;

  -- DDR3 RAM part (information only)
  signal ddr3_part : part_t := TB_DDR3_PART;

  -- Test clocks
  signal clk_250 : std_logic_dbl_t := ('1','0');
  signal clk_200 : std_logic_dbl_t := ('1','0');
  signal clk_125 : std_logic_dbl_t := ('1','0');
  signal clk_100 : std_logic_dbl_t := ('1','0');
  signal clk_80  : std_logic_dbl_t := ('1','0');

  -- Test results
  signal top_comp : top_comp_t;
  signal top_pass : top_pass_t;
  signal ds_comp  : ds_comp_t;
  signal ds_pass  : ds_pass_t;
  signal dma_comp : dma_comp_t;
  signal dma_pass : dma_pass_t;

  signal test_complete : boolean := false;
  signal test_passed   : boolean := true;

  -- Bridge clocks
  signal ocp_clk_full : std_logic;
  signal ocp_clk      : std_logic;
  signal mptl_clk     : std_logic_dbl_t;

  -- Bridge MPTL Sideband signals
  signal b_gpio_b2t,b_gpio_t2b : std_logic_vector(15 downto 0);
  signal b_mptl_status         : std_logic_vector(63 downto 0);
  signal b_mptl_config         : std_logic_vector(31 downto 0);
  signal dma_abort             : std_logic_vector(DMA_CHANNELS-1 downto 0);

  -- Probe errors
  signal probe_ds_err  : std_logic;
  signal probe_dm_err  : std_logic;
  signal probe_dma_err : std_logic;

  -- DDR3 signals
  signal mem_addr_out   : mem_addr_out_t;
  signal mem_ctrl_out   : mem_ctrl_out_t;
  signal mem_data_inout : mem_data_inout_t;
  signal mem_clk_out    : mem_clk_out_t;

  --------------------
  -- Target signals --
  --------------------

  -- Clocks
  signal clks_in     : clks_in_t;
  signal clks_out    : clks_out_t;
  signal clks_mgt_in : clks_mgt_in_t;

  -- Configuration/online signals
  signal linti_l             : std_logic;
  signal mptl_online         : std_logic;
  signal mptl_online_cnt     : std_logic_vector(7 downto 0);
  signal mptl_online_long    : std_logic;
  signal mptl_online_long_d1 : std_logic;

  -- MPTL signals
  signal mptl_b2t : mptl_pins_t;
  signal mptl_t2b : mptl_pins_t;

  -- MPTL sideband signals
  signal mptl_sb_b2t : mptl_sb_b2t_t;
  signal mptl_sb_t2b : mptl_sb_t2b_t;

  -- OCP signals
  signal direct_slave_m2s   : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal direct_slave_s2m   : adb3_ocp_s2mT;
  signal dma_channels_m2s   : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_M2S);
  signal dma_channels_s2m   : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  signal direct_masters_m2s : adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
  signal direct_masters_s2m : adb3_ocp_s2mT_array(0 to DM_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);

  -- GPIO Interface
  signal gpio_inout : gpio_inout_t;


  component test_uber_ds
  port(
    -- OCP Interface
    ocp_clk          : in    std_logic;
    direct_slave_m2s : out   adb3_ocp_m2sT;
    direct_slave_s2m : in    adb3_ocp_s2mT;
    -- Test interface
    mptl_online_long : in    std_logic;
    ds_comp          : out   ds_comp_t;
    ds_pass          : out   ds_pass_t);
  end component;

  component test_uber_dma
  port(
    -- OCP Interface
    ocp_clk          : in    std_logic;
    dma_channels_m2s : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    -- Test interface
    mptl_online_long : in    std_logic;
    dma_comp         : out   dma_comp_t;
    dma_pass         : out   dma_pass_t);
  end component;

  component test_uber_dm
  port(
    -- OCP Interface
    ocp_clk            : in    std_logic;
    direct_masters_m2s : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : out   adb3_ocp_s2mT_array(0 to DM_CHANNELS-1));
  end component;

  component test_uber_probes
  port(
    -- OCP Interface
    ocp_clk            : in    std_logic;
    direct_slave_m2s   : in    adb3_ocp_m2sT;
    direct_slave_s2m   : in    adb3_ocp_s2mT;
    dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m   : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    -- Probe Interface
    probe_ds_err       : out   std_logic;
    probe_dma_err      : out   std_logic;
    probe_dm_err       : out   std_logic);
  end component;

  component uber
  port(
    -- Non-MGT Clock Interface
    clks_in                  : in    clks_in_t;
    clks_out                 : out   clks_out_t;
    -- MGT Clock Interface
    clks_mgt_in              : in    clks_mgt_in_t;
    -- MPTL Interface
    mptl_b2t                 : in    mptl_pins_t;
    mptl_t2b                 : out   mptl_pins_t;
    -- MPTL Sideband Signals
    mptl_sb_b2t              : in    mptl_sb_b2t_t;
    mptl_sb_t2b              : out   mptl_sb_t2b_t;
    -- Interrupt Interface
    finti_l                  : out   std_logic;
    -- GPIO Interface
    gpio_inout               : inout gpio_inout_t;
    -- Memory Interface
    mem_addr_out             : out   mem_addr_out_t;
    mem_ctrl_out             : out   mem_ctrl_out_t;
    mem_data_inout           : inout mem_data_inout_t;
    mem_clk_out              : out   mem_clk_out_t);
  end component;

begin

  -- Generate unique reference clock frequencies
  -- 250 MHz clock
  clk_250.p <= not clk_250.p after CLK_250_PERIOD / 2;
  clk_250.n <= not clk_250.n after CLK_250_PERIOD / 2;

  -- 200 MHz clock
  clk_200.p <= not clk_200.p after CLK_200_PERIOD / 2;
  clk_200.n <= not clk_200.n after CLK_200_PERIOD / 2;

  -- 125 MHz clock
  clk_125.p <= not clk_125.p after CLK_125_PERIOD / 2;
  clk_125.n <= not clk_125.n after CLK_125_PERIOD / 2;

  -- 100 MHz clock
  clk_100.p <= not clk_100.p after CLK_100_PERIOD / 2;
  clk_100.n <= not clk_100.n after CLK_100_PERIOD / 2;

  --  80 MHz clock
  clk_80.p  <= not clk_80.p  after CLK_80_PERIOD / 2;
  clk_80.n  <= not clk_80.n  after CLK_80_PERIOD / 2;

  -- Report design constants
  report_setup_p : process
  begin
    report "Board Type : " & board_type_t'image(BOARD_TYPE);  -- adb3_target_inc_pkg
    report "Target Use : " & target_use_t'image(TARGET_USE);  -- adb3_target_inc_pkg
    report "Master Use : " & master_use_t'image(MASTER_USE);  -- uber_tb_pkg
    report "Waiting for MPTL online....";
    wait;
  end process;


  -----------------------------------
  -- Bridge Direct Slave interface --
  -----------------------------------
  test_uber_ds_i : test_uber_ds
  port map(
    -- OCP Interface
    ocp_clk          => ocp_clk,           -- in    std_logic
    direct_slave_m2s => direct_slave_m2s,  -- out   adb3_ocp_m2sT
    direct_slave_s2m => direct_slave_s2m,  -- in    adb3_ocp_s2mT
    -- Test interface
    mptl_online_long => mptl_online_long,  -- in    std_logic
    ds_comp          => ds_comp,           -- out   ds_comp_r
    ds_pass          => ds_pass);          -- out   ds_pass_r)

  --------------------------
  -- Bridge DMA interface --
  --------------------------
  test_uber_dma_i : test_uber_dma
  port map(
    -- OCP Interface
    ocp_clk          => ocp_clk,           -- in    std_logic
    dma_channels_m2s => dma_channels_m2s,  -- out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m => dma_channels_s2m,  -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    -- Test interface
    mptl_online_long => mptl_online_long,  -- in    std_logic
    dma_comp         => dma_comp,          -- out   dma_comp_t
    dma_pass         => dma_pass);         -- out   dma_pass_t)

  ------------------------------------
  -- Bridge Direct master interface --
  ------------------------------------

  dm32_g : if (MASTER_USE = MASTER_ON) generate
    test_uber_dm_i : test_uber_dm
    port map(
      -- OCP Interface
      ocp_clk            => ocp_clk,             -- in    std_logic
      direct_masters_m2s => direct_masters_m2s,  -- in    adb3_ocp_m2sT_array(0 to MASTER_CHANNELS-1)
      direct_masters_s2m => direct_masters_s2m); -- out   adb3_ocp_s2mT_array(0 to MASTER_CHANNELS-1))
  end generate;


  -- Test MPTL GPIO bus
  mptl_gpio_p : process
  begin
    wait until (mptl_online_long = '1');  -- OCP commands ignored until MPTL online
    adb3_wait_cycles(200,ocp_clk);        -- Wait for MPTL GPIO data to get to bridge

    top_comp.mptl_gpio_complete <= true;
    top_pass.mptl_gpio_passed   <= true;

    if (b_gpio_t2b = MPTL_GPIO) then
      assert false report "Test mptl_gpio completed: PASSED." severity note;
    else
      assert false report "Test mptl_gpio completed: FAILED." severity error;
      top_pass.mptl_gpio_passed <= false;
    end if;
  end process;


  -- Test dma_abort bus
  dma_abort_p : process
  begin
    wait until (mptl_online_long = '1');  -- OCP commands ignored until MPTL online
    adb3_wait_cycles(50,ocp_clk);         -- Wait for data to get to outputs

    top_comp.dma_abort_complete <= true;
    top_pass.dma_abort_passed   <= true;

    loop
      if (OR_reduce(dma_abort(DMA_CHANNELS-1 downto 0)) = '0') then
        --assert false report "Test dma_abort completed: PASSED." severity note;
        null;
      else
        assert false report "Test dma_abort completed: FAILED." severity error;
        top_pass.dma_abort_passed <= false;
        exit;
      end if;

      adb3_wait_cycles(1,ocp_clk);
    end loop;
  end process;


  -- Test clk_out period
  clk_out_p : process
    variable ck_rising : time := 0 ns;
    variable ck_period : time := 0 ns;
  begin
    wait until (mptl_online_long = '1');  -- OCP commands ignored until MPTL online

    top_comp.clk_out_complete <= true;
    top_pass.clk_out_passed   <= true;

    if (XRM_MGTCLK_C2M_VALID = '1') then
      wait until rising_edge(clks_out.xrm_mgtclk_c2m.p);
      ck_rising := now;
      wait until rising_edge(clks_out.xrm_mgtclk_c2m.p);
      ck_period := now - ck_rising;
    end if;

    report "Expected clk_out freq = " & integer'image(CLK_80_FREQ) & "MHz";
    report "Actual   clk_out freq = " & integer'image(1 us / ck_period) & "MHz";

    if (ck_period = CLK_80_PERIOD) then
      assert false report "Test clk_out completed: PASSED." severity note;
    else
      assert false report "Test clk_out completed: FAILED." severity error;
      top_pass.clk_out_passed <= false;
    end if;
  end process;


  -- Monitor interrupt input and detect falling edges
  linti_l_monitor_p : process
  begin
    loop
      -- Wait for a negative edge on 'linti_l'
      wait until falling_edge(linti_l);
      report "Interrupt Monitor: Detected falling edge on linti_l";
      -- Shared variable defined in uber_tb_pkg
      linti_l_edge := '1';
    end loop;
  end process;


  -------------------------
  -- Output test results --
  -------------------------

  gen_results_p : process(top_pass,top_comp,ds_pass,ds_comp,dma_pass,dma_comp)
  begin
    test_passed <= top_pass.mptl_gpio_passed and
                   top_pass.dma_abort_passed and
                   top_pass.clk_out_passed and
                   dma_pass.dma_write_cmd_passed and dma_pass.dma_read_cmd_passed and dma_pass.dma_read_resp_passed and
                   ds_pass.simple_passed and
                   ds_pass.master_passed and
                   ds_pass.clock_passed and
                   ds_pass.frontio_passed and ds_pass.reario_passed and
                   ds_pass.int_passed and
                   ds_pass.info_passed and
                   ds_pass.bram_passed and
                   ds_pass.ram_passed;

    test_complete <= top_comp.mptl_gpio_complete and
                     top_comp.dma_abort_complete and
                     top_comp.clk_out_complete and
                     dma_comp.dma_write_cmd_complete and dma_comp.dma_read_cmd_complete and dma_comp.dma_read_resp_complete and
                     ds_comp.simple_complete and
                     ds_comp.master_complete and
                     ds_comp.clock_complete and
                     ds_comp.frontio_complete and ds_comp.reario_complete and
                     ds_comp.int_complete and
                     ds_comp.info_complete and
                     ds_comp.bram_complete and
                     ds_comp.ram_complete;
  end process;


  test_results_p : process(test_complete,test_passed,probe_ds_err,probe_dm_err,probe_dma_err)
  begin
    if (test_complete) then
      assert not(probe_ds_err  = '1') report "Probe OCP Direct Slave status: ERROR."  severity error;
      assert not(probe_dm_err  = '1') report "Probe OCP Direct Master status: ERROR." severity error;
      assert not(probe_dma_err = '1') report "Probe OCP DMA status: ERROR."           severity error;

      if (test_passed) then
        assert false report "Test of design UBER completed: PASSED." severity failure;
      else
        assert false report "Test of design UBER completed: FAILED." severity failure;
      end if;
    end if;
  end process;


  ------------------------------------
  -- Target Component Instantiation --
  ------------------------------------

  -- Connect clock inputs
  ref_clk_g : if (REF_CLK_VALID = '1') generate
    clks_in.ref_clk <= clk_200; -- Osc. Reference clock 200 MHz (double ended)
  end generate;

  lclk_g : if (LCLK_VALID = '1') generate
    clks_in.lclk <= clk_80.p; -- Bridge programmable clock 40-140 MHz (single ended)
  end generate;

  xrm_clk_g : if (XRM_GCLK_M2C_VALID = '1') generate
    clks_in.xrm_gclk_m2c <= clk_125; -- XRM reference clock (double ended)
  end generate;

  mgt_clks_6tl_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- Connect test clocks to MGT clock inputs
    clks_mgt_in.mgt112_clk0   <= clk_125;     -- Osc. Reference clock 125 MHz
    clks_mgt_in.mgt113_clk0   <= clk_100;     -- PCIe Reference clock 100 MHz
    clks_mgt_in.mgt113_clk1   <= clk_250;     -- PCIe Reference clock 250 MHz
    clks_mgt_in.mgt114_clk0   <= clk_250;     -- PCIe Reference clock 250 MHz
    clks_mgt_in.mgt115_clk0   <= clk_125;     -- Osc. Reference clock 125 MHz
    clks_mgt_in.mgt116_clk0   <= clk_125;     -- Osc. Reference clock 125 MHz
    clks_mgt_in.mgt117_clk0   <= clk_125;     -- XRM  Reference clock
  end generate;

  mgt_clks_6t1_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    -- Connect test clocks to MGT clock inputs
    clks_mgt_in.mgt113_clk0   <= clk_250;     -- Osc. Reference clock 250 MHz
    clks_mgt_in.mgt113_clk1   <= clk_250;     -- PCIe Reference clock 250 MHz
    clks_mgt_in.mgt114_clk0   <= clk_250;     -- PCIe Reference clock 250 MHz
    clks_mgt_in.mgt115_clk0   <= clk_250;     -- Osc. Reference clock 250 MHz
    clks_mgt_in.mgt116_clk0   <= clk_250;     -- Osc. Reference clock 250 MHz
    clks_mgt_in.mgt117_clk0   <= clk_125;     -- XRM  Reference clock
  end generate;


  -- Instantiate target UUT
  uut : uber
  port map(
    -- Clock Interface
    clks_in                  => clks_in,             -- in    clks_in_t
    clks_out                 => clks_out,            -- out   clks_out_t
    -- MGT Clock Interface
    clks_mgt_in              => clks_mgt_in,         -- in    clks_mgt_in_t
    -- MPTL Interface
    mptl_b2t                 => mptl_b2t,            -- in    mptl_pins_t
    mptl_t2b                 => mptl_t2b,            -- out   mptl_pins_t
    -- MPTL Initialisation
    mptl_sb_b2t              => mptl_sb_b2t,         -- in    mptl_sb_b2t_t;
    mptl_sb_t2b              => mptl_sb_t2b,         -- out   mptl_sb_t2b_t;
    -- Interrupt Interface
    finti_l                  => linti_l,             -- out   std_logic
    -- GPIO Interface
    gpio_inout               => gpio_inout,          -- inout gpio_inout_t;
    --Memory Interface
    mem_addr_out             => mem_addr_out,        -- out   mem_addr_out_t
    mem_ctrl_out             => mem_ctrl_out,        -- out   mem_ctrl_out_t
    mem_data_inout           => mem_data_inout,      -- inout mem_data_inout_t
    mem_clk_out              => mem_clk_out);        -- out   mem_clk_out_t)


  -----------------------------------------------
  -- Bridge Component Instantiations (Private) --
  -----------------------------------------------

  -- Connect inputs
  b_gpio_b2t    <= MPTL_GPIO;
  b_mptl_config <= (others => '0');

  -- Full OCP-MPTL-OCP simulation
  -- Bridge ocp clock can be different from target ocp clock
  ocp_clk_full <= clk_250.p;

  mptl_clk_par_g : if (BOARD_TYPE = ADM_XRC_6TL) generate
    -- MPTL Parallel interface: Connect 250MHz PCIe input to bridge mptl_clk
    mptl_clk.p <= clk_250.p;
    mptl_clk.n <= clk_250.n;
  end generate;

  mptl_clk_ser_g : if (BOARD_TYPE = ADM_XRC_6T1) generate
    -- MPTL serial interface: Connect 250 MHz Osc input to bridge mptl_clk
    mptl_clk.p <= clk_250.p;
    mptl_clk.n <= clk_250.n;
  end generate;


  -- bridge interface
  mptl_if_bridge_wrap_i : mptl_if_bridge_wrap
  port map(
    -- OCP Interface
    ocp_clk_in               => ocp_clk_full,        -- in    std_logic
    direct_slave_m2s         => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m         => direct_slave_s2m,    -- out   adb3_ocp_s2mT
    dma_channels_m2s         => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m         => dma_channels_s2m,    -- out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    direct_masters_m2s       => direct_masters_m2s,  -- out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1)
    direct_masters_s2m       => direct_masters_s2m,  -- in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1)
    dma_abort                => dma_abort,           -- out   std_logic_vector(DMA_CHANNELS-1 downto 0)
    ocp_clk_out              => ocp_clk,             -- out   std_logic
    -- MPTL Interface
    mptl_b2t                 => mptl_b2t,            -- in    mptl_pins_t
    mptl_t2b                 => mptl_t2b,            -- out   mptl_pins_t
    mptl_clk                 => mptl_clk,            -- in    std_logic_dbl_t
    -- MPTL Initialisation
    mptl_sb_t2b              => mptl_sb_t2b,         -- in    mptl_sb_t2b_t
    mptl_sb_b2t              => mptl_sb_b2t,         -- out   mptl_sb_b2t_t
    -- MPTL Sideband Interface
    gpio_b2t                 => b_gpio_b2t,          -- in    std_logic_vector(15 downto 0)
    gpio_t2b                 => b_gpio_t2b,          -- out   std_logic_vector(15 downto 0)
    mptl_status              => b_mptl_status,       -- out   std_logic_vector(63 downto 0)
    mptl_config              => b_mptl_config);      -- in    std_logic_vector(31 downto 0))


  -- Generate online signal
  mptl_online <= not(mptl_sb_b2t.mptl_bridge_gtp_online_l or mptl_sb_t2b.mptl_target_gtp_online_l);


  -- Count cycles of mptl_online to make sure it is stable
  online_count_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (mptl_online /= '1') then
        mptl_online_cnt <= (others => '1');
      else
        if (or_reduce(mptl_online_cnt) = '1') then
          mptl_online_cnt <= mptl_online_cnt - 1;
        end if;
      end if;
      mptl_online_long <= not(or_reduce(mptl_online_cnt));
    end if;
  end process;

  -- Monitor MPTL link and stop when failure
  monitor_status_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      mptl_online_long_d1 <= mptl_online_long;
    end if;

    -- Detect falling edge on gtp_online
    assert not((mptl_online_long = '0')and(mptl_online_long_d1 = '1')) report "Test of design UBER: MPTL link failure (Undefined data)" severity failure;
  end process;


  ------------------------
  -- Bridge OCP probes --
  ------------------------

  test_uber_probes_i : test_uber_probes
  port map(
     -- OCP Interface
    ocp_clk            => ocp_clk,             -- in    std_logic
    direct_slave_m2s   => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m   => direct_slave_s2m,    -- in    adb3_ocp_s2mT
    dma_channels_m2s   => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m   => dma_channels_s2m,    -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    direct_masters_m2s => direct_masters_m2s,  -- in    adb3_ocp_m2sT_array(0 to MASTER_CHANNELS-1)
    direct_masters_s2m => direct_masters_s2m,  -- in    adb3_ocp_s2mT_array(0 to MASTER_CHANNELS-1)
    -- Probe Interface
    probe_ds_err       => probe_ds_err,        -- out   std_logic;
    probe_dma_err      => probe_dma_err,       -- out   std_logic;
    probe_dm_err       => probe_dm_err);       -- out   std_logic)


  ----------------------------
  -- DDR3 SDRAM Bank Models --
  ----------------------------

  ddr3_model_g : for n in DDR3_ST_BANK to (DDR3_ST_BANK+DDR3_MODEL_BANKS-1) generate
    -- Instantiate LS data RAM model
    ddr3_sdram_bank_ls_i : ddr3_sdram
      generic map(
        message_level  => 0,            -- natural;
        part           => TB_DDR3_PART, -- part_t;
        short_init_dly => true)         -- boolean);
      port map(
        ck             => mem_clk_out.ddr3_clk_out(n).clk_p(0),                    -- in    std_logic;
        ck_l           => mem_clk_out.ddr3_clk_out(n).clk_n(0),                    -- in    std_logic;
        reset_l        => mem_ctrl_out.ddr3_ctrl_out(n).reset_l,                   -- in    std_logic;
        cke            => mem_ctrl_out.ddr3_ctrl_out(n).cke(0),                    -- in    std_logic;
        cs_l           => mem_ctrl_out.ddr3_ctrl_out(n).cs_l(0),                   -- in    std_logic;
        ras_l          => mem_ctrl_out.ddr3_ctrl_out(n).ras_l,                     -- in    std_logic;
        cas_l          => mem_ctrl_out.ddr3_ctrl_out(n).cas_l,                     -- in    std_logic;
        we_l           => mem_ctrl_out.ddr3_ctrl_out(n).we_l,                      -- in    std_logic;
        odt            => mem_ctrl_out.ddr3_ctrl_out(n).odt(0),                    -- in    std_logic;
        dm             => mem_ctrl_out.ddr3_ctrl_out(n).dm(1 downto 0),            -- in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
        ba             => mem_ctrl_out.ddr3_ctrl_out(n).ba,                        -- in    std_logic_vector(part.part_size.bank_width-1 downto 0);
        a              => mem_addr_out.ddr3_addr_out(n).a(TB_DDR3_ROW-1 downto 0), -- in    std_logic_vector(part.part_size.arow_width-1 downto 0);
        dq             => mem_data_inout.ddr3_data_inout(n).dq(15 downto 0),       -- inout std_logic_vector(part.part_size.data_width-1 downto 0);
        dqs            => mem_data_inout.ddr3_data_inout(n).dqs_p(1 downto 0),     -- inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
        dqs_l          => mem_data_inout.ddr3_data_inout(n).dqs_n(1 downto 0));    -- inout std_logic_vector(part.part_size.data_bytes-1 downto 0);

    -- Instantiate MS data RAM model
    ddr3_sdram_bank_ms_i : ddr3_sdram
      generic map(
        message_level  => 0,            -- natural;
        part           => TB_DDR3_PART, -- part_t;
        short_init_dly => true)         -- boolean);
      port map(
        ck             => mem_clk_out.ddr3_clk_out(n).clk_p(0),                    -- in    std_logic;
        ck_l           => mem_clk_out.ddr3_clk_out(n).clk_n(0),                    -- in    std_logic;
        reset_l        => mem_ctrl_out.ddr3_ctrl_out(n).reset_l,                   -- in    std_logic;
        cke            => mem_ctrl_out.ddr3_ctrl_out(n).cke(0),                    -- in    std_logic;
        cs_l           => mem_ctrl_out.ddr3_ctrl_out(n).cs_l(0),                   -- in    std_logic;
        ras_l          => mem_ctrl_out.ddr3_ctrl_out(n).ras_l,                     -- in    std_logic;
        cas_l          => mem_ctrl_out.ddr3_ctrl_out(n).cas_l,                     -- in    std_logic;
        we_l           => mem_ctrl_out.ddr3_ctrl_out(n).we_l,                      -- in    std_logic;
        odt            => mem_ctrl_out.ddr3_ctrl_out(n).odt(0),                    -- in    std_logic;
        dm             => mem_ctrl_out.ddr3_ctrl_out(n).dm(3 downto 2),            -- in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
        ba             => mem_ctrl_out.ddr3_ctrl_out(n).ba,                        -- in    std_logic_vector(part.part_size.bank_width-1 downto 0);
        a              => mem_addr_out.ddr3_addr_out(n).a(TB_DDR3_ROW-1 downto 0), -- in    std_logic_vector(part.part_size.arow_width-1 downto 0);
        dq             => mem_data_inout.ddr3_data_inout(n).dq(31 downto 16),      -- inout std_logic_vector(part.part_size.data_width-1 downto 0);
        dqs            => mem_data_inout.ddr3_data_inout(n).dqs_p(3 downto 2),     -- inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
        dqs_l          => mem_data_inout.ddr3_data_inout(n).dqs_n(3 downto 2));    -- inout std_logic_vector(part.part_size.data_bytes-1 downto 0);
  end generate;

end test;