--==============================================================================
--
--     File Name: test_uber_dma_1ch.vhd
--       Version: 0.1
--          Date: 17 Mar 2010
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: DMA
--
--   Description: Provides bridge DMA interface (single channel)
--
--   Limitations: Single DMA channel access
--
--         Notes: blocking DMA test
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 17 Mar 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.uber_pkg.all;
use work.uber_tb_pkg.all;

entity test_uber_dma is
port(
  -- OCP Interface
  ocp_clk            : in    std_logic;
  dma_channels_m2s   : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m   : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  -- Test interface
  mptl_online_long   : in    std_logic;
  dma_comp           : out   dma_comp_t;
  dma_pass           : out   dma_pass_t);
end test_uber_dma;

architecture test of test_uber_dma is

  signal dma_m2s : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal dma_s2m : adb3_ocp_s2mT;

begin

  -- Connect dma inputs
  dma_s2m <= dma_channels_s2m(DMA_SINGLE_CHANNEL);


  -- Generate DMA transfers to/from target
  dma_channel_p : process
    variable addr   : ocp_addr_s;
    variable data   : byte_vector_t(0 to 511);
    variable be     : byte_enable_t(0 to 511);
    variable xfer   : natural;
    variable pat32  : std_logic_vector(31 downto 0);
    variable exp32  : std_logic_vector(31 downto 0);
    variable val32  : std_logic_vector(31 downto 0);
    variable tx_cnt : natural;
    variable rx_cnt : natural;
    variable tag    : natural;

    variable remaining : natural := DMA_SIZE;
    variable chunk     : natural;
    variable num_err   : natural;
  begin
    dma_comp.dma_write_cmd_complete <= false;
    dma_pass.dma_write_cmd_passed   <= true;
    dma_comp.dma_read_cmd_complete  <= false;
    dma_pass.dma_read_cmd_passed    <= true;
    dma_comp.dma_read_resp_complete <= false;
    dma_pass.dma_read_resp_passed   <= true;

    wait until (mptl_online_long = '1');  -- OCP commands ignored until ADB3 is online

    report "DMA write process started";

    addr   := DMA_ADDR_WR;
    be     := (others => '1');
    tx_cnt := 0;
    val32  := X"beef0000";
    tag    := 0;

    while (remaining /= 0) loop
      chunk := 96;
      if (chunk > remaining) then
        chunk := remaining;
      end if;

      for i in 0 to ((chunk/4)-1) loop
        data(4*i to 4*(i+1)-1) := conv_byte_vector(val32 + tx_cnt + i);
      end loop;

      adb3_target_sim_write(addr,tag,be(0 to chunk-1),data(0 to chunk-1),xfer,ocp_clk,dma_m2s,dma_s2m);
--      report "Wrote data chunk (Tag: " & integer'image(tag) & ", Chunk: " & integer'image(chunk) &
--             ") Transferred Bytes: " & integer'image(xfer) & " on DMA channel 0";

      remaining  := remaining - xfer;
      tx_cnt     := tx_cnt + (xfer/4);
      addr       := addr + chunk;
      tag        := tag + 1;
    end loop;

    report "DMA write process completed";
    report integer'image(tx_cnt*4) & " bytes transferred.";
    dma_comp.dma_write_cmd_complete <= true;

    if (tx_cnt*4 /= DMA_SIZE) then
      dma_pass.dma_write_cmd_passed <= false;
    end if;


    report "DMA read process started";

    addr   := DMA_ADDR_RD;
    rx_cnt := 0;
    pat32  := X"beef0000";
    tag    := 0;

    num_err := 0;
    remaining := DMA_SIZE;
    while remaining /= 0 loop
      chunk := 512;
      if (chunk > remaining) then
        chunk := remaining;
      end if;

      adb3_target_sim_read(addr,tag,data(0 to chunk-1),xfer,ocp_clk,dma_m2s,dma_s2m);
--      report "Read data chunk (Tag: " & integer'image(tag) & ", Chunk: " & integer'image(chunk) &
--             ") Transferred Bytes: " & integer'image(xfer) & " on DMA channel 1";
      remaining := remaining - xfer;

      for i in 0 to (xfer/4-1) loop
        val32 := data(4*i+3) & data(4*i+2) & data(4*i+1) & data(4*i);
        exp32 := pat32 + rx_cnt + i;
        if (val32 /= exp32) then
          assert false
            report "(stimulate_dma1_p) Verification error: expected = " & conv_string(exp32(15 downto 0)) &
                   " actual = " & conv_string(val32(15 downto 0))
            severity error;
          num_err := num_err + 1;
        end if;
      end loop;

      rx_cnt := rx_cnt + (xfer/4);
      addr   := addr + chunk;
      tag    := tag + 1;
    end loop;

    report "DMA read process completed";
    report integer'image(rx_cnt*4) & " bytes transferred with " & integer'image(num_err) & " data error(s)";
    dma_comp.dma_read_cmd_complete  <= true;
    dma_comp.dma_read_resp_complete <= true;

    if (rx_cnt*4 = DMA_SIZE)and(num_err = 0) then
      assert false report "Test DMA completed: PASSED." severity note;
    else
      assert false report "Test DMA completed: FAILED." severity failure;
      dma_pass.dma_read_cmd_passed  <= false;
      dma_pass.dma_read_resp_passed <= false;
    end if;

    wait;
  end process;

  -- Connect DMA outputs
  dma_channels_m2s(DMA_SINGLE_CHANNEL) <= dma_m2s;

end test;